@@ -1,5 +1,5 @@
-from .lazy import *  # noqa
 from .differentiable import *  # noqa
+from .elementary import *  # noqa
 from .finite_difference import *  # noqa
 from .derivative import *  # noqa
 from .tools import generate_fd_shortcuts  # noqa
