@@ -2,16 +2,19 @@ from collections import ChainMap
 from functools import singledispatch
 
 import sympy
-from sympy.functions.elementary.integers import floor
+from sympy.core.add import _addsort
+from sympy.core.mul import _mulsort
+from sympy.core.decorators import call_highest_priority
 from sympy.core.evalf import evalf_table
 
 from cached_property import cached_property
-from devito.finite_differences.lazy import Evaluable
+from devito.finite_differences.tools import make_shift_x0
 from devito.logger import warning
-from devito.tools import filter_ordered, flatten
+from devito.tools import filter_ordered, flatten, split
+from devito.types.lazy import Evaluable
 from devito.types.utils import DimensionTuple
 
-__all__ = ['Differentiable']
+__all__ = ['Differentiable', 'EvalDerivative']
 
 
 class Differentiable(sympy.Expr, Evaluable):
@@ -49,21 +52,6 @@ class Differentiable(sympy.Expr, Evaluable):
         return min([getattr(i, 'time_order', 100) or 100 for i in self._args_diff],
                    default=100)
 
-    @cached_property
-    def is_TimeDependent(self):
-        # Default False, True if anything is time dependent in the expression
-        return any(getattr(i, 'is_TimeDependent', False) for i in self._args_diff)
-
-    @cached_property
-    def is_VectorValued(self):
-        # Default False, True if is a vector valued expression
-        return any(getattr(i, 'is_VectorValued', False) for i in self._args_diff)
-
-    @cached_property
-    def is_TensorValued(self):
-        # Default False, True if is a tensor valued expression
-        return any(getattr(i, 'is_TensorValued', False) for i in self._args_diff)
-
     @cached_property
     def grid(self):
         grids = {getattr(i, 'grid', None) for i in self._args_diff} - {None}
@@ -72,7 +60,7 @@ class Differentiable(sympy.Expr, Evaluable):
         try:
             return grids.pop()
         except KeyError:
-            raise ValueError("No grid found")
+            return None
 
     @cached_property
     def indices(self):
@@ -102,6 +90,10 @@ class Differentiable(sympy.Expr, Evaluable):
     def is_Staggered(self):
         return any([getattr(i, 'is_Staggered', False) for i in self._args_diff])
 
+    @cached_property
+    def is_TimeDependent(self):
+        return any(i.is_Time for i in self.dimensions)
+
     @cached_property
     def _fd(self):
         return dict(ChainMap(*[getattr(i, '_fd', {}) for i in self._args_diff]))
@@ -157,30 +149,39 @@ class Differentiable(sympy.Expr, Evaluable):
         raise AttributeError("%r object has no attribute %r" % (self.__class__, name))
 
     # Override SymPy arithmetic operators
+    @call_highest_priority('__radd__')
     def __add__(self, other):
         return Add(self, other)
 
+    @call_highest_priority('__add__')
     def __iadd__(self, other):
         return Add(self, other)
 
+    @call_highest_priority('__add__')
     def __radd__(self, other):
         return Add(other, self)
 
+    @call_highest_priority('__rsub__')
     def __sub__(self, other):
         return Add(self, -other)
 
+    @call_highest_priority('__sub__')
     def __isub__(self, other):
         return Add(self, -other)
 
+    @call_highest_priority('__sub__')
     def __rsub__(self, other):
         return Add(other, -self)
 
+    @call_highest_priority('__rmul__')
     def __mul__(self, other):
         return Mul(self, other)
 
+    @call_highest_priority('__mul__')
     def __imul__(self, other):
         return Mul(self, other)
 
+    @call_highest_priority('__mul__')
     def __rmul__(self, other):
         return Mul(other, self)
 
@@ -190,9 +191,11 @@ class Differentiable(sympy.Expr, Evaluable):
     def __rpow__(self, other):
         return Pow(other, self)
 
+    @call_highest_priority('__rdiv__')
     def __div__(self, other):
         return Mul(self, Pow(other, sympy.S.NegativeOne))
 
+    @call_highest_priority('__div__')
     def __rdiv__(self, other):
         return Mul(other, Pow(self, sympy.S.NegativeOne))
 
@@ -200,9 +203,11 @@ class Differentiable(sympy.Expr, Evaluable):
     __rtruediv__ = __rdiv__
 
     def __floordiv__(self, other):
+        from .elementary import floor
         return floor(self / other)
 
     def __rfloordiv__(self, other):
+        from .elementary import floor
         return floor(other / self)
 
     def __mod__(self, other):
@@ -239,16 +244,18 @@ class Differentiable(sympy.Expr, Evaluable):
         derivs = tuple('d%s2' % d.name for d in space_dims)
         return Add(*[getattr(self, d) for d in derivs])
 
-    @property
-    def div(self):
+    def div(self, shift=None):
         space_dims = [d for d in self.dimensions if d.is_Space]
-        derivs = tuple('d%s' % d.name for d in space_dims)
-        return Add(*[getattr(self, d) for d in derivs])
+        shift_x0 = make_shift_x0(shift, (len(space_dims),))
+        return Add(*[getattr(self, 'd%s' % d.name)(x0=shift_x0(shift, d, None, i))
+                     for i, d in enumerate(space_dims)])
 
-    @property
-    def grad(self):
+    def grad(self, shift=None):
         from devito.types.tensor import VectorFunction, VectorTimeFunction
-        comps = [getattr(self, 'd%s' % d.name) for d in self.dimensions if d.is_Space]
+        space_dims = [d for d in self.dimensions if d.is_Space]
+        shift_x0 = make_shift_x0(shift, (len(space_dims),))
+        comps = [getattr(self, 'd%s' % d.name)(x0=shift_x0(shift, d, None, i))
+                 for i, d in enumerate(space_dims)]
         vec_func = VectorTimeFunction if self.is_TimeDependent else VectorFunction
         return vec_func(name='grad_%s' % self.name, time_order=self.time_order,
                         space_order=self.space_order, components=comps, grid=self.grid)
@@ -295,6 +302,11 @@ class DifferentiableOp(Differentiable):
     __sympy_class__ = None
 
     def __new__(cls, *args, **kwargs):
+        # Do not re-evaluate if any of the args is an EvalDerivative,
+        # since the integrity of these objects must be preserved
+        if any(isinstance(i, EvalDerivative) for i in args):
+            kwargs['evaluate'] = False
+
         obj = cls.__base__.__new__(cls, *args, **kwargs)
 
         # Unfortunately SymPy may build new sympy.core objects (e.g., sympy.Add),
@@ -343,14 +355,69 @@ class DifferentiableOp(Differentiable):
         return None
 
 
+class DifferentiableFunction(DifferentiableOp):
+
+    def __new__(cls, *args, **kwargs):
+        return cls.__sympy_class__.__new__(cls, *args, **kwargs)
+
+    @property
+    def evaluate(self):
+        return self.func(*[getattr(a, 'evaluate', a) for a in self.args])
+
+    def _eval_at(self, func):
+        return self
+
+
 class Add(DifferentiableOp, sympy.Add):
     __sympy_class__ = sympy.Add
-    __new__ = DifferentiableOp.__new__
+
+    def __new__(cls, *args, **kwargs):
+        # Here, often we get `evaluate=False` to prevent SymPy evaluation (e.g.,
+        # when `cls==EvalDerivative`), but in all cases we at least apply a small
+        # set of basic simplifications
+
+        # (a+b)+c -> a+b+c (flattening)
+        nested, others = split(args, lambda e: isinstance(e, Add))
+        args = flatten(e.args for e in nested) + list(others)
+
+        # a+0 -> a
+        args = [i for i in args if i != 0]
+
+        # Reorder for homogeneity with pure SymPy types
+        _addsort(args)
+
+        return super().__new__(cls, *args, **kwargs)
 
 
 class Mul(DifferentiableOp, sympy.Mul):
     __sympy_class__ = sympy.Mul
-    __new__ = DifferentiableOp.__new__
+
+    def __new__(cls, *args, **kwargs):
+        # A Mul, being a DifferentiableOp, may not trigger evaluation upon
+        # construction (e.g., when an EvalDerivative is present among its
+        # arguments), so here we apply a small set of basic simplifications
+        # to avoid generating functional, but also ugly, code
+
+        # (a*b)*c -> a*b*c (flattening)
+        nested, others = split(args, lambda e: isinstance(e, Mul))
+        args = flatten(e.args for e in nested) + list(others)
+
+        # a*0 -> 0
+        if any(i == 0 for i in args):
+            return sympy.S.Zero
+
+        # a*1 -> a
+        args = [i for i in args if i != 1]
+
+        # a*-1*-1 -> a
+        nminus = len([i for i in args if i == sympy.S.NegativeOne])
+        if nminus % 2 == 0:
+            args = [i for i in args if i != sympy.S.NegativeOne]
+
+        # Reorder for homogeneity with pure SymPy types
+        _mulsort(args)
+
+        return super().__new__(cls, *args, **kwargs)
 
     @property
     def _gather_for_diff(self):
@@ -375,7 +442,7 @@ class Mul(DifferentiableOp, sympy.Mul):
         for f in self.args:
             if f not in self._args_diff:
                 new_args.append(f)
-            elif f is func_args:
+            elif f is func_args or isinstance(f, DifferentiableFunction):
                 new_args.append(f)
             else:
                 ind_f = f.indices_ref._getters
@@ -393,12 +460,46 @@ class Mul(DifferentiableOp, sympy.Mul):
 class Pow(DifferentiableOp, sympy.Pow):
     _fd_priority = 0
     __sympy_class__ = sympy.Pow
-    __new__ = DifferentiableOp.__new__
 
 
 class Mod(DifferentiableOp, sympy.Mod):
     __sympy_class__ = sympy.Mod
-    __new__ = DifferentiableOp.__new__
+
+
+class EvalDerivative(DifferentiableOp, sympy.Add):
+
+    is_commutative = True
+
+    def __new__(cls, *args, base=None, **kwargs):
+        kwargs['evaluate'] = False
+
+        # a+0 -> a
+        args = [i for i in args if i != 0]
+
+        # Reorder for homogeneity with pure SymPy types
+        _addsort(args)
+
+        obj = super().__new__(cls, *args, **kwargs)
+
+        try:
+            obj.base = base
+        except AttributeError:
+            # This might happen if e.g. one attempts a (re)construction with
+            # one sole argument. The (re)constructed EvalDerivative degenerates
+            # to an object of different type, in classic SymPy style. That's fine
+            assert len(args) <= 1
+            assert not obj.is_Add
+            return obj
+
+        return obj
+
+    @property
+    def func(self):
+        return lambda *a, **kw: EvalDerivative(*a, base=self.base, **kw)
+
+    def _new_rawargs(self, *args, **kwargs):
+        kwargs.pop('is_commutative', None)
+        return self.func(*args, **kwargs)
 
 
 class diffify(object):
@@ -457,6 +558,7 @@ class diffify(object):
     @_cls.register(Mul)
     @_cls.register(Pow)
     @_cls.register(Mod)
+    @_cls.register(EvalDerivative)
     def _(obj):
         return obj.__class__
 
@@ -478,6 +580,9 @@ def diff2sympy(expr):
         except AttributeError:
             # Not of type DifferentiableOp
             pass
+        except TypeError:
+            # Won't lower (e.g., EvalDerivative)
+            pass
         if flag:
             return obj.func(*args, evaluate=False), True
         else:
