@@ -1,19 +1,18 @@
 from devito.ir.clusters.queue import QueueStateful
-from devito.ir.support import (SEQUENTIAL, PARALLEL, PARALLEL_INDEP, PARALLEL_IF_ATOMIC,
-                               AFFINE, ROUNDABLE, TILABLE, Forward)
+from devito.ir.support import (AFFINE, PARALLEL, PARALLEL_INDEP, PARALLEL_IF_ATOMIC,
+                               ROUNDABLE, SEQUENTIAL, Forward)
 from devito.tools import as_tuple, flatten, timed_pass
 
 __all__ = ['analyze']
 
 
 @timed_pass()
-def analyze(clusters):
+def analyze(clusters, options):
     state = QueueStateful.State()
 
     # Collect properties
     clusters = Parallelism(state).process(clusters)
     clusters = Affiness(state).process(clusters)
-    clusters = Tiling(state).process(clusters)
     clusters = Rounding(state).process(clusters)
 
     # Reconstruct Clusters attaching the discovered properties
@@ -97,12 +96,12 @@ class Parallelism(Detector):
                 is_parallel_indep &= (dep.distance_mapper.get(d.root) == 0)
                 continue
 
+            if dep.function in scope.initialized:
+                # False alarm, the dependence is over a locally-defined symbol
+                continue
+
             if dep.is_increment:
-                if dep.function in scope.initialized:
-                    # False alarm, the increment is over a locally-defined symbol
-                    pass
-                else:
-                    is_parallel_atomic = True
+                is_parallel_atomic = True
                 continue
 
             return SEQUENTIAL
@@ -134,7 +133,7 @@ class Rounding(Detector):
         # so that `i`'s trip count can safely be rounded up
         # Note: autopadding guarantees that the padding size along the
         # Fastest Varying Dimension is a multiple of the SIMD vector length
-        functions = [f for f in scope.functions if f.is_Tensor]
+        functions = [f for f in scope.functions if f.is_AbstractFunction]
         if any(not f._honors_autopadding for f in functions):
             return
 
@@ -156,39 +155,3 @@ class Affiness(Detector):
         accesses = [a for a in scope.accesses if not a.is_scalar]
         if all(a.is_regular and a.affine_if_present(d._defines) for a in accesses):
             return AFFINE
-
-
-class Tiling(Detector):
-
-    """
-    Detect the TILABLE Dimensions.
-    """
-
-    def process(self, elements):
-        return self._process_fdta(elements, 1)
-
-    def _callback(self, clusters, d, prefix):
-        # A Dimension is TILABLE only if it's PARALLEL and AFFINE
-        properties = self._fetch_properties(clusters, prefix)
-        if not {PARALLEL, AFFINE} <= properties[d]:
-            return
-
-        # In addition, we use the heuristic that we do not consider
-        # TILABLE a Dimension that is not embedded in at least one
-        # SEQUENTIAL Dimension. This is to rule out tiling when the
-        # computation is not expected to be expensive
-        if not any(SEQUENTIAL in properties[i.dim] for i in prefix[:-1]):
-            return
-
-        # Likewise, it won't be marked TILABLE if there's at least one
-        # local SubDimension in all Clusters
-        if all(any(i.dim.is_Sub and i.dim.local for i in c.itintervals)
-               for c in clusters):
-            return
-
-        # If it induces dynamic bounds, then it's ruled out too
-        scope = self._fetch_scope(clusters)
-        if any(i.is_lex_non_stmt for i in scope.d_all_gen()):
-            return
-
-        return TILABLE
