@@ -1,12 +1,14 @@
 from collections.abc import Iterable
 from operator import attrgetter
 
+from sympy import sympify
+
 from devito.symbolics import (retrieve_functions, retrieve_indexed, split_affine,
                               uxreplace)
 from devito.tools import PartialOrderTuple, filter_sorted, flatten, as_tuple
-from devito.types import Dimension
+from devito.types import Dimension, Eq, IgnoreDimSort
 
-__all__ = ['dimension_sort', 'lower_exprs']
+__all__ = ['dimension_sort', 'generate_implicit_exprs', 'lower_exprs']
 
 
 def dimension_sort(expr):
@@ -34,7 +36,10 @@ def dimension_sort(expr):
                                      if isinstance(d, Dimension)])
         return tuple(relation)
 
-    relations = {handle_indexed(i) for i in retrieve_indexed(expr)}
+    if isinstance(expr.implicit_dims, IgnoreDimSort):
+        relations = set()
+    else:
+        relations = {handle_indexed(i) for i in retrieve_indexed(expr)}
 
     # Add in any implicit dimension (typical of scalar temporaries, or Step)
     relations.add(expr.implicit_dims)
@@ -68,6 +73,43 @@ def dimension_sort(expr):
     return ordering
 
 
+def generate_implicit_exprs(expressions):
+    """
+    Create and add implicit expressions.
+
+    Implicit expressions are those not explicitly defined by the user
+    but instead are requisites of some specified functionality.
+
+    Currently, implicit expressions stem from the following:
+
+        * MultiSubDomains attached to input equations.
+    """
+    found = {}
+    processed = []
+    for e in expressions:
+        if e.subdomain:
+            try:
+                dims = [d.root for d in e.free_symbols if isinstance(d, Dimension)]
+                sub_dims = [d.root for d in e.subdomain.dimensions]
+                sub_dims.extend(e.subdomain.implicit_dimensions)
+                dims = [d for d in dims if d not in frozenset(sub_dims)]
+                dims.extend(e.subdomain.implicit_dimensions)
+                if e.subdomain not in found:
+                    grid = list(retrieve_functions(e, mode='unique'))[0].grid
+                    found[e.subdomain] = [i.func(*i.args, implicit_dims=dims) for i in
+                                          e.subdomain._create_implicit_exprs(grid)]
+                processed.extend(found[e.subdomain])
+                dims.extend(e.subdomain.dimensions)
+                new_e = Eq(e.lhs, e.rhs, subdomain=e.subdomain, implicit_dims=dims)
+                processed.append(new_e)
+            except AttributeError:
+                # Not a MultiSubDomain
+                processed.append(e)
+        else:
+            processed.append(e)
+    return processed
+
+
 def lower_exprs(expressions, **kwargs):
     """
     Lowering an expression consists of the following passes:
@@ -80,6 +122,8 @@ def lower_exprs(expressions, **kwargs):
     --------
     f(x - 2*h_x, y) -> f[xi + 2, yi + 4]  (assuming halo_size=4)
     """
+    # Normalize subs
+    subs = {k: sympify(v) for k, v in kwargs.get('subs', {}).items()}
 
     processed = []
     for expr in as_tuple(expressions):
@@ -90,7 +134,7 @@ def lower_exprs(expressions, **kwargs):
             dimension_map = {}
 
         # Handle Functions (typical case)
-        mapper = {f: f.indexify(lshift=True, subs=dimension_map)
+        mapper = {f: lower_exprs(f.indexify(subs=dimension_map), **kwargs)
                   for f in retrieve_functions(expr)}
 
         # Handle Indexeds (from index notation)
@@ -101,22 +145,23 @@ def lower_exprs(expressions, **kwargs):
             indices = [(lower_exprs(a) + o) for a, o in
                        zip(i.indices, f._size_nodomain.left)]
 
+            # Substitute spacing (spacing only used in own dimension)
+            indices = [i.xreplace({d.spacing: 1, -d.spacing: -1})
+                       for i, d in zip(indices, f.dimensions)]
+
             # Apply substitutions, if necessary
             if dimension_map:
                 indices = [j.xreplace(dimension_map) for j in indices]
 
             mapper[i] = f.indexed[indices]
 
-        subs = kwargs.get('subs')
         # Add dimensions map to the mapper in case dimensions are used
         # as an expression, i.e. Eq(u, x, subdomain=xleft)
         mapper.update(dimension_map)
-        if subs:
-            # Include the user-supplied substitutions, and use
-            # `xreplace` for constant folding
-            processed.append(expr.xreplace({**mapper, **subs}))
-        else:
-            processed.append(uxreplace(expr, mapper))
+        # Add the user-supplied substitutions
+        mapper.update(subs)
+        # Apply mapper to expression
+        processed.append(uxreplace(expr, mapper))
 
     if isinstance(expressions, Iterable):
         return processed
