@@ -2,28 +2,30 @@
 
 import abc
 import inspect
-import numbers
 from cached_property import cached_property
 from collections import OrderedDict, namedtuple
 from collections.abc import Iterable
 
 import cgen as c
+from sympy import IndexedBase
 
 from devito.data import FULL
-from devito.ir.equations import ClusterizedEq
-from devito.ir.support import (SEQUENTIAL, PARALLEL, PARALLEL_IF_ATOMIC, VECTORIZED,
-                               AFFINE, Property, Forward, detect_io)
-from devito.symbolics import ListInitializer, FunctionFromPointer, as_symbol, ccode
-from devito.tools import (Signer, as_tuple, filter_ordered, filter_sorted, flatten,
-                          validate_type, dtype_to_cstr)
-from devito.types import Symbol, Indexed
-from devito.types.basic import AbstractFunction
+from devito.ir.equations import DummyEq
+from devito.ir.support import (SEQUENTIAL, PARALLEL, PARALLEL_IF_ATOMIC,
+                               PARALLEL_IF_PVT, VECTORIZED, AFFINE, COLLAPSED,
+                               Property, Forward, detect_io)
+from devito.symbolics import ListInitializer, CallFromPointer, as_symbol, ccode
+from devito.tools import Signer, as_tuple, filter_ordered, filter_sorted, flatten
+from devito.types.basic import (AbstractFunction, AbstractSymbol, Indexed,
+                                LocalObject, Symbol)
 
-__all__ = ['Node', 'Block', 'Expression', 'Element', 'Callable', 'Call', 'Conditional',
-           'Iteration', 'List', 'LocalExpression', 'Section', 'TimedList', 'Prodder',
-           'MetaCall', 'PointerCast', 'ForeignExpression', 'HaloSpot', 'IterationTree',
-           'ExpressionBundle', 'AugmentedExpression', 'Increment', 'Return', 'While',
-           'ParallelIteration', 'ParallelBlock', 'Dereference']
+__all__ = ['Node', 'Block', 'Expression', 'Element', 'Callable', 'Call',
+           'Conditional', 'Iteration', 'List', 'Section', 'TimedList', 'Prodder',
+           'MetaCall', 'PointerCast', 'HaloSpot', 'IterationTree', 'Definition',
+           'ExpressionBundle', 'AugmentedExpression', 'Increment', 'Return',
+           'While', 'ParallelIteration', 'ParallelBlock', 'Dereference', 'Lambda',
+           'SyncSpot', 'Pragma', 'PragmaTransfer', 'DummyExpr', 'BlankLine',
+           'ParallelTree', 'BusyWait', 'CallableBody']
 
 # First-class IET nodes
 
@@ -32,26 +34,24 @@ class Node(Signer):
 
     __metaclass__ = abc.ABCMeta
 
-    is_Node = True
     is_Block = False
     is_Iteration = False
-    is_IterationFold = False
     is_While = False
     is_Expression = False
     is_Increment = False
-    is_ForeignExpression = False
     is_Callable = False
+    is_CallableBody = False
+    is_Conditional = False
     is_ElementalFunction = False
     is_Call = False
     is_List = False
+    is_Definition = False
     is_PointerCast = False
     is_Dereference = False
-    is_Element = False
     is_Section = False
     is_HaloSpot = False
     is_ExpressionBundle = False
-    is_ParallelIteration = False
-    is_ParallelBlock = False
+    is_SyncSpot = False
 
     _traversable = []
     """
@@ -62,10 +62,15 @@ class Node(Signer):
 
     def __new__(cls, *args, **kwargs):
         obj = super(Node, cls).__new__(cls)
-        argnames = inspect.getfullargspec(cls.__init__).args
+        argnames, _, _, defaultvalues, _, _, _ = inspect.getfullargspec(cls.__init__)
+        try:
+            defaults = dict(zip(argnames[-len(defaultvalues):], defaultvalues))
+        except TypeError:
+            # No default kwarg values
+            defaults = {}
         obj._args = {k: v for k, v in zip(argnames[1:], args)}
         obj._args.update(kwargs.items())
-        obj._args.update({k: None for k in argnames[1:] if k not in obj._args})
+        obj._args.update({k: defaults.get(k) for k in argnames[1:] if k not in obj._args})
         return obj
 
     def _rebuild(self, *args, **kwargs):
@@ -115,20 +120,25 @@ class Node(Signer):
     def __str__(self):
         return str(self.ccode)
 
-    @abc.abstractproperty
+    @property
     def functions(self):
         """All AbstractFunction objects used by this node."""
-        raise NotImplementedError()
+        return ()
 
-    @abc.abstractproperty
-    def free_symbols(self):
-        """All Symbol objects used by this node."""
-        raise NotImplementedError()
+    @property
+    def expr_symbols(self):
+        """All symbols appearing in an expression within this node."""
+        return ()
 
-    @abc.abstractproperty
+    @property
     def defines(self):
-        """All Symbol objects defined by this node."""
-        raise NotImplementedError()
+        """All Basic objects defined by this node."""
+        return ()
+
+    @property
+    def writes(self):
+        """All Basic objects modified by this node."""
+        return ()
 
     def _signature_items(self):
         return (str(self.ccode),)
@@ -185,18 +195,6 @@ class List(Node):
         return "<%s (%d, %d, %d)>" % (self.__class__.__name__, len(self.header),
                                       len(self.body), len(self.footer))
 
-    @property
-    def functions(self):
-        return ()
-
-    @property
-    def free_symbols(self):
-        return ()
-
-    @property
-    def defines(self):
-        return ()
-
 
 class Block(List):
 
@@ -217,8 +215,6 @@ class Element(Node):
     be expressed through an IET type.
     """
 
-    is_Element = True
-
     def __init__(self, element):
         assert isinstance(element, (c.Comment, c.Statement, c.Value, c.Initializer,
                                     c.Pragma, c.Line, c.Assign, c.POD))
@@ -230,64 +226,141 @@ class Element(Node):
 
 class Call(ExprStmt, Node):
 
-    """A function call."""
+    """
+    A function call.
+
+    Parameters
+    ----------
+    name : str or CallFromPointer
+        The called function.
+    arguments : list of Basic, optional
+        The objects in input to the function call.
+    retobj : Symbol or Indexed, optional
+        The object the return value of the Call is assigned to.
+    is_indirect : bool, optional
+        If True, the object represents an indirect function call. The emitted
+        code will be `name, arg1, ..., argN` rather than `name(arg1, ..., argN)`.
+        Defaults to False.
+    cast : bool, optional
+        If True, the Call return value is explicitly casted to the `retobj` type.
+        Defaults to False.
+    writes : list, optional
+        The AbstractFunctions that will be written to by the called function.
+        Explicitly tagging these AbstractFunctions is useful in the case of external
+        calls, that is whenever the compiler would be unable to retrieve that
+        information by analysis of the IET graph.
+    """
 
     is_Call = True
 
-    def __init__(self, name, arguments=None):
-        self.name = name
+    def __init__(self, name, arguments=None, retobj=None, is_indirect=False,
+                 cast=False, writes=None):
+        if isinstance(name, CallFromPointer):
+            self.base = name.base
+        else:
+            self.base = None
+        self.name = str(name)
         self.arguments = as_tuple(arguments)
+        self.retobj = retobj
+        self.is_indirect = is_indirect
+        self.cast = cast
+        self._writes = as_tuple(writes)
 
     def __repr__(self):
-        return "Call::\n\t%s(...)" % self.name
-
-    @property
-    def functions(self):
-        return tuple(i for i in self.arguments if isinstance(i, AbstractFunction))
+        ret = "" if self.retobj is None else "%s = " % self.retobj
+        return "%sCall::\n\t%s(...)" % (ret, self.name)
 
     @property
     def children(self):
-        return tuple(i for i in self.arguments if isinstance(i, Call))
+        return tuple(i for i in self.arguments if isinstance(i, (Call, Lambda)))
 
     @cached_property
-    def free_symbols(self):
-        free = set()
+    def functions(self):
+        retval = []
         for i in self.arguments:
-            if isinstance(i, numbers.Number):
-                continue
-            elif isinstance(i, AbstractFunction):
-                free.add(i)
+            if isinstance(i, (AbstractFunction, Indexed, IndexedBase, LocalObject)):
+                retval.append(i.function)
+            elif isinstance(i, Call):
+                retval.extend(i.functions)
             else:
-                free.update(i.free_symbols)
-        return free
+                try:
+                    v = i.free_symbols
+                except AttributeError:
+                    continue
+                for s in v:
+                    try:
+                        # `try-except` necessary for e.g. Macro
+                        if isinstance(s.function, (AbstractFunction, LocalObject)):
+                            retval.append(s.function)
+                    except AttributeError:
+                        continue
+        if self.base is not None:
+            retval.append(self.base.function)
+        if self.retobj is not None:
+            retval.append(self.retobj.function)
+        return tuple(filter_ordered(retval))
+
+    @cached_property
+    def expr_symbols(self):
+        retval = []
+        for i in self.arguments:
+            if isinstance(i, AbstractFunction):
+                continue
+            elif isinstance(i, (Indexed, IndexedBase, LocalObject, Symbol)):
+                retval.append(i)
+            elif isinstance(i, Call):
+                retval.extend(i.expr_symbols)
+            else:
+                try:
+                    retval.extend(i.free_symbols)
+                except AttributeError:
+                    pass
+        if self.base is not None:
+            retval.append(self.base)
+        if self.retobj is not None:
+            retval.extend(self.retobj.free_symbols)
+        return tuple(filter_ordered(retval))
 
     @property
     def defines(self):
-        return ()
+        ret = ()
+        if self.base is not None:
+            ret += (self.base,)
+        if self.retobj is not None:
+            ret += (self.retobj,)
+        return ret
+
+    @property
+    def writes(self):
+        return self._writes
 
 
 class Expression(ExprStmt, Node):
 
-    """A node encapsulating a ClusterizedEq."""
+    """
+    A node encapsulating a ClusterizedEq.
+
+    Parameters
+    ----------
+    expr : ClusterizedEq
+        The encapsulated expression.
+    pragmas : cgen.Pragma or list of cgen.Pragma, optional
+        A bag of pragmas attached to this Expression.
+    init : bool, optional
+        True if an initialization, False otherwise (default).
+    """
 
     is_Expression = True
 
-    @validate_type(('expr', ClusterizedEq))
-    def __init__(self, expr):
-        self.__expr_finalize__(expr)
-
-    def __expr_finalize__(self, expr):
-        """Finalize the Expression initialization."""
-        self._expr = expr
+    def __init__(self, expr, pragmas=None, init=False):
+        self.expr = expr
+        self.pragmas = as_tuple(pragmas)
+        self.init = init
 
     def __repr__(self):
         return "<%s::%s>" % (self.__class__.__name__,
                              filter_ordered([f.func for f in self.functions]))
 
-    @property
-    def expr(self):
-        return self._expr
-
     @property
     def dtype(self):
         return self.expr.dtype
@@ -302,10 +375,10 @@ class Expression(ExprStmt, Node):
         """The Functions read by the Expression."""
         return detect_io(self.expr, relax=True)[0]
 
-    @property
+    @cached_property
     def write(self):
         """The Function written by the Expression."""
-        return self.expr.lhs.function
+        return self.expr.lhs.base.function
 
     @cached_property
     def dimensions(self):
@@ -314,28 +387,28 @@ class Expression(ExprStmt, Node):
 
     @property
     def is_scalar(self):
-        """True if a scalar expression, False otherwise."""
-        return self.expr.lhs.is_Symbol
+        """True if the LHS is a scalar, False otherwise."""
+        return isinstance(self.expr.lhs, (AbstractSymbol, IndexedBase))
 
     @property
     def is_tensor(self):
-        """True if a tensor expression, False otherwise."""
-        return not self.is_scalar
+        """True if the LHS is an array entry, False otherwise."""
+        return self.expr.lhs.is_Indexed
 
     @property
-    def is_definition(self):
+    def is_initializable(self):
         """
-        True if it is an assignment, False otherwise
+        True if it can be an initializing assignment, False otherwise.
         """
         return ((self.is_scalar and not self.is_Increment) or
                 (self.is_tensor and isinstance(self.expr.rhs, ListInitializer)))
 
     @property
     def defines(self):
-        return (self.write,) if self.is_definition else ()
+        return (self.output.base,) if self.is_initializable else ()
 
     @property
-    def free_symbols(self):
+    def expr_symbols(self):
         return tuple(self.expr.free_symbols)
 
     @cached_property
@@ -345,6 +418,10 @@ class Expression(ExprStmt, Node):
             functions.append(self.write)
         return tuple(filter_ordered(functions))
 
+    @property
+    def writes(self):
+        return (self.write,)
+
 
 class AugmentedExpression(Expression):
 
@@ -352,8 +429,8 @@ class AugmentedExpression(Expression):
 
     is_Increment = True
 
-    def __init__(self, expr, op):
-        super(AugmentedExpression, self).__init__(expr)
+    def __init__(self, expr, op, pragmas=None):
+        super(AugmentedExpression, self).__init__(expr, pragmas=pragmas)
         self.op = op
 
 
@@ -361,8 +438,8 @@ class Increment(AugmentedExpression):
 
     """Shortcut for ``AugmentedExpression(expr, '+'), since it's so widely used."""
 
-    def __init__(self, expr):
-        super(Increment, self).__init__(expr, '+')
+    def __init__(self, expr, pragmas=None):
+        super(Increment, self).__init__(expr, '+', pragmas=pragmas)
 
 
 class Iteration(Node):
@@ -448,9 +525,17 @@ class Iteration(Node):
     def is_ParallelAtomic(self):
         return PARALLEL_IF_ATOMIC in self.properties
 
+    @property
+    def is_ParallelPrivate(self):
+        return PARALLEL_IF_PVT in self.properties
+
+    @property
+    def is_ParallelNoAtomic(self):
+        return any([self.is_Parallel, self.is_ParallelPrivate])
+
     @property
     def is_ParallelRelaxed(self):
-        return self.is_Parallel or self.is_ParallelAtomic
+        return any([self.is_Parallel, self.is_ParallelAtomic, self.is_ParallelPrivate])
 
     @property
     def is_Vectorized(self):
@@ -516,13 +601,12 @@ class Iteration(Node):
         return _max - _min + 1
 
     @property
-    def functions(self):
-        """All Functions appearing in the Iteration header."""
-        return ()
+    def dimensions(self):
+        """All Dimensions appearing in the Iteration header."""
+        return tuple(self.dim._defines) + self.uindices
 
-    @property
-    def free_symbols(self):
-        """All Symbols appearing in the Iteration header."""
+    @cached_property
+    def expr_symbols(self):
         return tuple(self.symbolic_min.free_symbols) \
             + tuple(self.symbolic_max.free_symbols) \
             + self.uindices \
@@ -531,19 +615,8 @@ class Iteration(Node):
 
     @property
     def defines(self):
-        """All Symbols defined in the Iteration header."""
         return self.dimensions
 
-    @property
-    def dimensions(self):
-        """All Dimensions appearing in the Iteration header."""
-        return tuple(self.dim._defines) + self.uindices
-
-    @property
-    def write(self):
-        """All Functions written to in this Iteration"""
-        return []
-
 
 class While(Node):
 
@@ -586,26 +659,92 @@ class Callable(Node):
     parameters : list of Basic, optional
         The objects in input to the Callable.
     prefix : list of str, optional
-        Qualifiers to prepend to the Callable signature. Defaults to ``('static',
-        'inline')``.
+        Qualifiers to prepend to the Callable signature. None by defaults.
     """
 
     is_Callable = True
 
     _traversable = ['body']
 
-    def __init__(self, name, body, retval, parameters=None, prefix=('static', 'inline')):
+    def __init__(self, name, body, retval, parameters=None, prefix=None):
         self.name = name
-        self.body = as_tuple(body)
+        if not isinstance(body, CallableBody):
+            self.body = CallableBody(body)
+        else:
+            self.body = body
         self.retval = retval
         self.prefix = as_tuple(prefix)
         self.parameters = as_tuple(parameters)
 
     def __repr__(self):
-        parameters = ",".join(['void*' if i.is_Object else dtype_to_cstr(i.dtype)
-                               for i in self.parameters])
         return "%s[%s]<%s; %s>" % (self.__class__.__name__, self.name, self.retval,
-                                   parameters)
+                                   ",".join([i._C_typename for i in self.parameters]))
+
+    @property
+    def functions(self):
+        return tuple(i.function for i in self.parameters
+                     if isinstance(i.function, AbstractFunction))
+
+    @property
+    def defines(self):
+        return self.parameters
+
+
+class CallableBody(Node):
+
+    """
+    The immediate child of a Callable.
+
+    Parameters
+    ----------
+    body : Node or list of Node
+        The actual body.
+    init : Node, optional
+        A piece of IET to perform some initialization relevant for `body`
+        (e.g., to initialize the target language runtime).
+    unpacks : list of Nodes, optional
+        Statements unpacking data from composite types.
+    allocs : list of Nodes, optional
+        Data definitions and allocations for `body`.
+    objs : list of Definitions, optional
+        Object definitions for `body`.
+    casts : list of PointerCasts, optional
+        Sequence of PointerCasts required by the `body`.
+    maps : Transfer or list of Transfer, optional
+        Data maps for `body` (a data map may e.g. trigger a data transfer from
+        host to device).
+    unmaps : Transfer or list of Transfer, optional
+        Data unmaps for `body`.
+    frees : list of Calls, optional
+        Data deallocations for `body`.
+    """
+
+    is_CallableBody = True
+
+    _traversable = ['init', 'unpacks', 'allocs', 'objs', 'casts', 'maps',
+                    'body', 'unmaps', 'frees']
+
+    def __init__(self, body, init=None, unpacks=None, allocs=None, objs=None,
+                 casts=None, maps=None, unmaps=None, frees=None):
+        # Sanity check
+        assert not isinstance(body, CallableBody), "CallableBody's cannot be nested"
+
+        self.body = as_tuple(body)
+        self.init = as_tuple(init)
+        self.unpacks = as_tuple(unpacks)
+        self.allocs = as_tuple(allocs)
+        self.objs = as_tuple(objs)
+        self.casts = as_tuple(casts)
+        self.maps = as_tuple(maps)
+        self.unmaps = as_tuple(unmaps)
+        self.frees = as_tuple(frees)
+
+    def __repr__(self):
+        return ("<CallableBody <unpacks=%d, allocs=%d, objs=%d, casts=%d, "
+                "maps=%d> <unmaps=%d, frees=%d>>" %
+                (len(self.unpacks), len(self.allocs), len(self.objs),
+                 len(self.casts), len(self.maps), len(self.unmaps),
+                 len(self.frees)))
 
 
 class Conditional(Node):
@@ -650,13 +789,9 @@ class Conditional(Node):
         return tuple(ret)
 
     @property
-    def free_symbols(self):
+    def expr_symbols(self):
         return tuple(self.condition.free_symbols)
 
-    @property
-    def defines(self):
-        return ()
-
 
 # Second level IET nodes
 
@@ -678,14 +813,21 @@ class TimedList(List):
     def __init__(self, timer, lname, body):
         self._name = lname
         self._timer = timer
-        header = [c.Statement("struct timeval start_%s, end_%s" % (lname, lname)),
-                  c.Statement("gettimeofday(&start_%s, NULL)" % lname)]
-        footer = [c.Statement("gettimeofday(&end_%s, NULL)" % lname),
-                  c.Statement(("%(gn)s->%(ln)s += " +
-                               "(double)(end_%(ln)s.tv_sec-start_%(ln)s.tv_sec)+" +
-                               "(double)(end_%(ln)s.tv_usec-start_%(ln)s.tv_usec)" +
-                               "/1000000") % {'gn': timer.name, 'ln': lname})]
-        super(TimedList, self).__init__(header, body, footer)
+
+        super().__init__(header=c.Line('START_TIMER(%s)' % lname),
+                         body=body,
+                         footer=c.Line('STOP_TIMER(%s,%s)' % (lname, timer.name)))
+
+    @classmethod
+    def _start_timer_header(cls):
+        return ('START_TIMER(S)', ('struct timeval start_ ## S , end_ ## S ; '
+                                   'gettimeofday(&start_ ## S , NULL);'))
+
+    @classmethod
+    def _stop_timer_header(cls):
+        return ('STOP_TIMER(S,T)', ('gettimeofday(&end_ ## S, NULL); T->S += (double)'
+                                    '(end_ ## S .tv_sec-start_ ## S.tv_sec)+(double)'
+                                    '(end_ ## S .tv_usec-start_ ## S .tv_usec)/1000000;'))
 
     @property
     def name(self):
@@ -695,21 +837,55 @@ class TimedList(List):
     def timer(self):
         return self._timer
 
-    @property
-    def free_symbols(self):
-        return (self.timer,)
 
-
-class PointerCast(Node):
+class Definition(ExprStmt, Node):
 
     """
-    A node encapsulating a cast of a raw C pointer to a multi-dimensional array.
+    A node encapsulating a variable definition.
+
+    If `shape` is given, then the Definition implements an array allocated
+    on the stack, otherwise it implements an uninitialized pointer.
+    """
+
+    is_Definition = True
+
+    def __init__(self, function, shape=None, qualifier=None, initvalue=None,
+                 constructor_args=None):
+        self.function = function
+        self.shape = shape
+        self.qualifier = qualifier
+        self.initvalue = initvalue
+        self.constructor_args = as_tuple(constructor_args)
+
+    def __repr__(self):
+        return "<Def(%s)>" % self.function
+
+    @property
+    def functions(self):
+        return (self.function,)
+
+    @property
+    def defines(self):
+        if self.shape is not None:
+            return (self.function.indexed,)
+        else:
+            return (self.function,)
+
+
+class PointerCast(ExprStmt, Node):
+
+    """
+    A node encapsulating a cast of a raw void pointer to a non-void,
+    potentially multi-dimensional array.
     """
 
     is_PointerCast = True
 
-    def __init__(self, function):
+    def __init__(self, function, obj=None, alignment=True, flat=None):
         self.function = function
+        self.obj = obj
+        self.alignment = alignment
+        self.flat = flat
 
     def __repr__(self):
         return "<PointerCast(%s)>" % self.function
@@ -730,108 +906,95 @@ class PointerCast(Node):
         return (self.function,)
 
     @property
-    def free_symbols(self):
-        """
-        The symbols required by the PointerCast.
-
-        This may include DiscreteFunctions as well as Dimensions.
-        """
-        if self.function.is_ArrayBasic:
-            sizes = flatten(s.free_symbols for s in self.function.symbolic_shape[1:])
-            return (self.function, ) + as_tuple(sizes)
+    def expr_symbols(self):
+        f = self.function
+        if f.is_ArrayBasic:
+            return tuple(flatten(s.free_symbols for s in f.symbolic_shape[1:]))
         else:
-            return (self.function,)
+            return ()
 
     @property
     def defines(self):
-        return ()
+        return (self.function.indexed,)
 
 
 class Dereference(ExprStmt, Node):
 
     """
-    A node encapsulating a dereference from a PointerArray to an Array.
+    A node encapsulating a dereference from a `pointer` to a `pointee`.
+    The following cases are supported:
+
+        * `pointer` is a PointerArray or TempFunction, and `pointee` is an Array.
+        * `pointer` is an ArrayObject representing a pointer to a C struct, and
+          `pointee` is a field in `pointer`.
     """
 
     is_Dereference = True
 
-    def __init__(self, array, parray):
-        self.array = array
-        self.parray = parray
+    def __init__(self, pointee, pointer, flat=None):
+        self.pointee = pointee
+        self.pointer = pointer
+        self.flat = flat
 
     def __repr__(self):
-        return "<Dereference(%s,%s)>" % (self.array, self.parray)
+        return "<Dereference(%s,%s)>" % (self.pointee, self.pointer)
 
     @property
     def functions(self):
-        return (self.array, self.parray)
+        return (self.pointee, self.pointer)
 
     @property
-    def free_symbols(self):
-        """
-        The symbols required by the PointerCast.
-
-        This may include DiscreteFunctions as well as Dimensions.
-        """
-        return ((self.array, self.parray) +
-                tuple(flatten(i.free_symbols for i in self.array.symbolic_shape[1:])) +
-                tuple(self.parray.free_symbols))
+    def expr_symbols(self):
+        return ((self.pointee.indexed.label, self.pointer.indexed.label) +
+                (self.pointer.indexify(),) +
+                tuple(flatten(i.free_symbols for i in self.pointee.symbolic_shape[1:])) +
+                tuple(self.pointer.free_symbols))
 
     @property
     def defines(self):
-        return (self.array,)
-
-
-class LocalExpression(Expression):
-
-    """
-    A node encapsulating a SymPy equation which also defines its LHS.
-    """
-
-    @property
-    def defines(self):
-        return (self.write, )
-
-
-class ForeignExpression(Expression):
-
-    """A node representing a SymPy FunctionFromPointer expression."""
-
-    is_ForeignExpression = True
-
-    @validate_type(('expr', FunctionFromPointer),
-                   ('dtype', type))
-    def __init__(self, expr, dtype, **kwargs):
-        self._dtype = dtype
-        self._is_increment = kwargs.get('is_Increment', False)
-        self.__expr_finalize__(expr)
-
-    @property
-    def dtype(self):
-        return self._dtype
-
-    @property
-    def output(self):
-        return self.expr.base
-
-    @property
-    def write(self):
-        if isinstance(self.output, (Symbol, Indexed)):
-            return self.output.function
+        if self.pointer.is_PointerArray or \
+           self.pointer.is_TempFunction or \
+           self.pointee._mem_stack:
+            return (self.pointee.indexed,)
         else:
-            return None
+            return (self.pointee,)
 
-    @property
-    def is_Increment(self):
-        return self._is_increment
 
-    @property
-    def is_scalar(self):
-        return False
+class Lambda(Node):
 
-    @property
-    def is_tensor(self):
-        return False
+    """
+    A callable C++ lambda function. Several syntaxes are possible; here we
+    implement one of the common ones:
+
+        [captures](parameters){body}
+
+    For more info about C++ lambda functions:
+
+        https://en.cppreference.com/w/cpp/language/lambda
+
+    Parameters
+    ----------
+    body : Node or list of Node
+        The lambda function body.
+    captures : list of str or expr-like, optional
+        The captures of the lambda function.
+    parameters : list of Basic or expr-like, optional
+        The objects in input to the lambda function.
+    """
+
+    _traversable = ['body']
+
+    def __init__(self, body, captures=None, parameters=None):
+        self.body = as_tuple(body)
+        self.captures = as_tuple(captures)
+        self.parameters = as_tuple(parameters)
+
+    def __repr__(self):
+        return "Lambda[%s](%s)" % (self.captures, self.parameters)
+
+    @cached_property
+    def expr_symbols(self):
+        return tuple(self.parameters)
 
 
 class Section(List):
@@ -848,12 +1011,13 @@ class Section(List):
 
     is_Section = True
 
-    def __init__(self, name, body=None):
+    def __init__(self, name, body=None, is_subsection=False):
         super(Section, self).__init__(body=body)
         self.name = name
+        self.is_subsection = is_subsection
 
     def __repr__(self):
-        return "<Section (%d)>" % len(self.body)
+        return "<Section (%s)>" % self.name
 
     @property
     def roots(self):
@@ -899,7 +1063,7 @@ class Prodder(Call):
     """
 
     def __init__(self, name, arguments=None, single_thread=False, periodic=False):
-        super(Prodder, self).__init__(name, arguments)
+        super().__init__(name, arguments)
 
         # Prodder properties
         self._single_thread = single_thread
@@ -914,13 +1078,159 @@ class Prodder(Call):
         return self._periodic
 
 
+class Pragma(Node):
+
+    """
+    One or more pragmas floating in the IET.
+    """
+
+    def __init__(self, pragmas):
+        super().__init__()
+        self.pragmas = as_tuple(pragmas)
+
+    def __repr__(self):
+        return '<Pragmas>'
+
+
+class Transfer(object):
+
+    """
+    A mixin for nodes representing data transfers.
+    """
+
+    pass
+
+
+class PragmaTransfer(Pragma, Transfer):
+
+    """
+    A data transfer between host and device expressed by means of one or more pragmas.
+    """
+
+    def __init__(self, callback, function, **kwargs):
+        super().__init__(callback(function, **kwargs))
+        self.callback = callback
+        self.function = function
+        self.kwargs = kwargs
+
+    @property
+    def functions(self):
+        return (self.function,)
+
+    @cached_property
+    def expr_symbols(self):
+        retval = [self.function.indexed]
+        for i in flatten(self.kwargs.items()):
+            try:
+                retval.extend(i.free_symbols)
+            except AttributeError:
+                pass
+        return tuple(retval)
+
+
 class ParallelIteration(Iteration):
 
     """
     Implement a parallel for-loop.
     """
 
-    is_ParallelIteration = True
+    def __init__(self, *args, **kwargs):
+        pragmas, kwargs, properties = self._make_header(**kwargs)
+        super().__init__(*args, pragmas=pragmas, properties=properties, **kwargs)
+
+    @classmethod
+    def _make_header(cls, **kwargs):
+        construct = cls._make_construct(**kwargs)
+        clauses = cls._make_clauses(**kwargs)
+        header = c.Pragma(' '.join([construct] + clauses))
+
+        # Extract the Iteration Properties
+        properties = cls._process_properties(**kwargs)
+
+        # Drop the unrecognised or unused kwargs
+        kwargs = cls._process_kwargs(**kwargs)
+
+        return (header,), kwargs, properties
+
+    @classmethod
+    def _make_construct(cls, **kwargs):
+        # To be overridden by subclasses
+        raise NotImplementedError
+
+    @classmethod
+    def _make_clauses(cls, **kwargs):
+        return []
+
+    @classmethod
+    def _process_properties(cls, **kwargs):
+        properties = as_tuple(kwargs.get('properties'))
+        properties += (COLLAPSED(kwargs.get('ncollapse', 1)),)
+
+        return properties
+
+    @classmethod
+    def _process_kwargs(cls, **kwargs):
+        kwargs.pop('pragmas', None)
+        kwargs.pop('properties', None)
+
+        # Recognised clauses
+        kwargs.pop('ncollapse', None)
+        kwargs.pop('reduction', None)
+
+        return kwargs
+
+    @cached_property
+    def collapsed(self):
+        ret = [self]
+        for i in range(self.ncollapsed - 1):
+            ret.append(ret[i].nodes[0])
+        assert all(i.is_Iteration for i in ret)
+        return tuple(ret)
+
+
+class ParallelTree(List):
+
+    """
+    This class is to group together a parallel for-loop with some setup
+    statements, for example:
+
+        .. code-block:: C
+
+          int chunk_size = ...
+          #pragma parallel for ... schedule(..., chunk_size)
+          for (int i = ...)
+          {
+            ...
+          }
+    """
+
+    _traversable = ['prefix', 'body']
+
+    def __init__(self, prefix, body, nthreads=None):
+        # Normalize and sanity-check input
+        body = as_tuple(body)
+        assert len(body) == 1 and body[0].is_Iteration
+
+        self.prefix = as_tuple(prefix)
+        self.nthreads = nthreads
+
+        super().__init__(body=body)
+
+    def __getattr__(self, name):
+        if 'body' in self.__dict__:
+            # During unpickling, `__setattr__` calls `__getattr__(..., 'body')`,
+            # which would cause infinite recursion if we didn't check whether
+            # 'body' is present or not
+            return getattr(self.body[0], name)
+        raise AttributeError
+
+    @property
+    def functions(self):
+        return as_tuple(self.nthreads)
+
+    @property
+    def root(self):
+        return self.body[0]
 
 
 class ParallelBlock(Block):
@@ -929,9 +1239,86 @@ class ParallelBlock(Block):
     A sequence of Nodes, wrapped in a parallel block {...}.
     """
 
-    is_ParallelBlock = True
+    def __init__(self, body, private=None):
+        # Normalize and sanity-check input. A bit ugly, but it makes everything
+        # much simpler to manage and reconstruct
+        body = as_tuple(body)
+        assert len(body) == 1
+        body = body[0]
+        assert body.is_List
+        if isinstance(body, ParallelTree):
+            partree = body
+        elif body.is_List:
+            assert len(body.body) == 1 and isinstance(body.body[0], ParallelTree)
+            assert len(body.footer) == 0
+            partree = body.body[0]
+            partree = partree._rebuild(prefix=(List(header=body.header,
+                                                    body=partree.prefix)))
+
+        header = self._make_header(partree.nthreads, private)
+        super().__init__(header=header, body=partree)
+
+    @classmethod
+    def _make_header(cls, nthreads, private=None):
+        return None
+
+    @property
+    def partree(self):
+        return self.body[0]
+
+    @property
+    def root(self):
+        return self.partree.root
+
+    @property
+    def nthreads(self):
+        return self.partree.nthreads
+
+    @property
+    def collapsed(self):
+        return self.partree.collapsed
 
 
+class BusyWait(While):
+
+    """
+    A while-loop implementing a busy waiting.
+    """
+
+    pass
+
+
+class SyncSpot(List):
+
+    """
+    A node representing one or more synchronization operations, e.g., WaitLock,
+    withLock, etc.
+    """
+
+    is_SyncSpot = True
+
+    def __init__(self, sync_ops, body=None):
+        super().__init__(body=body)
+        self.sync_ops = sync_ops
+
+    def __repr__(self):
+        return "<SyncSpot (%s)>" % ",".join(str(i) for i in self.sync_ops)
+
+
+class CBlankLine(List):
+
+    def __init__(self, **kwargs):
+        super().__init__(header=c.Line())
+
+    def __repr__(self):
+        return ""
+
+
+def DummyExpr(*args, init=False):
+    return Expression(DummyEq(*args), init=init)
+
+
+BlankLine = CBlankLine()
 Return = lambda i='': Element(c.Statement('return%s' % ((' %s' % i) if i else i)))
 
 
@@ -997,14 +1384,6 @@ class HaloSpot(Node):
     def functions(self):
         return tuple(self.fmapper)
 
-    @property
-    def free_symbols(self):
-        return ()
-
-    @property
-    def defines(self):
-        return ()
-
 
 # Utility classes
 
