@@ -6,20 +6,23 @@ The main Visitor class is adapted from https://github.com/coneoproject/COFFEE.
 
 from collections import OrderedDict
 from collections.abc import Iterable
-from operator import attrgetter
+from itertools import chain, groupby
 
 import cgen as c
+from sympy import IndexedBase
 
 from devito.exceptions import VisitorException
-from devito.ir.iet.nodes import Node, Iteration, Expression, Call
+from devito.ir.iet.nodes import (Node, Iteration, Expression, ExpressionBundle,
+                                 Call, Lambda, BlankLine, Section)
 from devito.ir.support.space import Backward
 from devito.symbolics import ccode, uxreplace
-from devito.tools import GenericVisitor, as_tuple, filter_sorted, flatten
-from devito.types.basic import AbstractFunction
+from devito.tools import GenericVisitor, as_tuple, filter_ordered, filter_sorted, flatten
+from devito.types.basic import AbstractFunction, Basic, IndexedData
+from devito.types import ArrayObject, Dimension, VoidPointer
 
 
 __all__ = ['FindNodes', 'FindSections', 'FindSymbols', 'MapExprStmts', 'MapNodes',
-           'IsPerfectIteration', 'XSubs', 'printAST', 'CGen', 'Transformer']
+           'IsPerfectIteration', 'printAST', 'CGen', 'Transformer', 'Uxreplace']
 
 
 class Visitor(GenericVisitor):
@@ -85,6 +88,12 @@ class PrintAST(Visitor):
         self._depth -= 1
         return self.indent + '<Callable %s>\n%s' % (o.name, body)
 
+    def visit_CallableBody(self, o):
+        self._depth += 1
+        body = [self._visit(o.init), self._visit(o.unpacks), self._visit(o.body)]
+        self._depth -= 1
+        return self.indent + "%s\n%s" % (o.__repr__(), '\n'.join([i for i in body if i]))
+
     def visit_list(self, o):
         return ('\n').join([self._visit(i) for i in o])
 
@@ -138,12 +147,6 @@ class PrintAST(Visitor):
         else:
             return self.indent + str(o)
 
-    def visit_ForeignExpression(self, o):
-        if self.verbose:
-            return self.indent + "<Expression %s>" % o.expr
-        else:
-            return self.indent + str(o)
-
     def visit_HaloSpot(self, o):
         self._depth += 1
         body = self._visit(o.children)
@@ -172,7 +175,7 @@ class CGen(Visitor):
         """Generate cgen declarations from an iterable of symbols and expressions."""
         ret = []
         for i in args:
-            if i.is_Tensor:
+            if isinstance(i, (AbstractFunction, IndexedData)):
                 ret.append(c.Value('%srestrict' % i._C_typename, i._C_name))
             elif i.is_AbstractObject or i.is_Symbol:
                 ret.append(c.Value(i._C_typename, i._C_name))
@@ -181,107 +184,210 @@ class CGen(Visitor):
         return ret
 
     def _args_call(self, args):
-        """Generate cgen function call arguments from an iterable of symbols and
-        expressions."""
+        """
+        Generate cgen function call arguments from an iterable of symbols and expressions.
+        """
         ret = []
         for i in args:
             try:
                 if isinstance(i, Call):
-                    ret.append(self.visit(i).text)
-                elif i.is_LocalObject:
-                    ret.append('&%s' % i._C_name)
-                elif i.is_ArrayBasic:
-                    ret.append("(%s)%s" % (i._C_typename, i.name))
+                    ret.append(self._visit(i, nested_call=True))
+                elif isinstance(i, Lambda):
+                    ret.append(self._visit(i))
                 else:
                     ret.append(i._C_name)
             except AttributeError:
                 ret.append(ccode(i))
         return ret
 
-    def visit_PointerCast(self, o):
-        f = o.function
-        if f.is_PointerArray:
-            rvalue = '(%s**) %s' % (f._C_typedata, f._C_name)
-            lvalue = c.Value(f._C_typedata, '**%s' % f.name)
-        else:
-            shape = ''.join("[%s]" % ccode(i) for i in o.castshape)
-            if f.is_DiscreteFunction:
-                rvalue = '(%s (*)%s) %s->%s' % (f._C_typedata, shape, f._C_name,
-                                                f._C_field_data)
-            else:
-                rvalue = '(%s (*)%s) %s' % (f._C_typedata, shape, f._C_name)
-            lvalue = c.AlignedAttribute(f._data_alignment,
-                                        c.Value(f._C_typedata,
-                                                '(*restrict %s)%s' % (f.name, shape)))
-        return c.Initializer(lvalue, rvalue)
+    def _blankline_logic(self, children):
+        """
+        Generate cgen blank lines in between logical units.
+        """
+        candidates = (ExpressionBundle, Iteration, Section)
 
-    def visit_Dereference(self, o):
-        a0, a1 = o.functions
-        shape = ''.join("[%s]" % ccode(i) for i in a0.symbolic_shape[1:])
-        rvalue = '(%s (*)%s) %s[%s]' % (a1._C_typedata, shape, a1.name, a1.dim.name)
-        lvalue = c.AlignedAttribute(a0._data_alignment,
-                                    c.Value(a0._C_typedata,
-                                            '(*restrict %s)%s' % (a0.name, shape)))
-        return c.Initializer(lvalue, rvalue)
+        processed = []
+        for child in children:
+            prev = None
+            rebuilt = []
+            for k, group in groupby(child, key=type):
+                g = list(group)
+
+                if k in (ExpressionBundle, Section) and len(g) >= 2:
+                    # Separate consecutive Sections/ExpressionBundles with BlankLine
+                    for i in g[:-1]:
+                        rebuilt.append(i)
+                        rebuilt.append(BlankLine)
+                    rebuilt.append(g[-1])
+                elif prev in candidates and k in candidates:
+                    rebuilt.append(BlankLine)
+                    rebuilt.extend(g)
+                else:
+                    rebuilt.extend(g)
+
+                prev = k
+
+            processed.append(tuple(rebuilt))
+
+        return tuple(processed)
+
+    def visit_object(self, o):
+        return o
+
+    visit_Generable = visit_object
+    visit_Collection = visit_object
 
     def visit_tuple(self, o):
         return tuple(self._visit(i) for i in o)
 
+    def visit_PointerCast(self, o):
+        f = o.function
+        if isinstance(o.obj, VoidPointer):
+            obj = o.obj.name
+        elif isinstance(o.obj, ArrayObject):
+            obj = '%s->%s' % (o.obj.name, f._C_name)
+        else:
+            obj = f._C_name
+        if f.is_PointerArray:
+            lvalue = c.Value(f._C_typedata, '**%s' % f.name)
+            rvalue = '(%s**) %s' % (f._C_typedata, obj)
+        else:
+            if o.flat is None:
+                shape = ''.join("[%s]" % ccode(i) for i in o.castshape)
+                rshape = '(*)%s' % shape
+                lvalue = c.Value(f._C_typedata, '(*restrict %s)%s' % (f.name, shape))
+            else:
+                rshape = '*'
+                lvalue = c.Value(f._C_typedata, '*%s' % o.flat)
+            if o.alignment:
+                lvalue = c.AlignedAttribute(f._data_alignment, lvalue)
+            if f.is_DiscreteFunction:
+                rvalue = '(%s %s) %s->%s' % (f._C_typedata, rshape, obj, f._C_field_data)
+            else:
+                rvalue = '(%s %s) %s' % (f._C_typedata, rshape, obj)
+        return c.Initializer(lvalue, rvalue)
+
+    def visit_Dereference(self, o):
+        a0, a1 = o.functions
+        if a1.is_PointerArray or a1.is_TempFunction:
+            if o.flat is None:
+                shape = ''.join("[%s]" % ccode(i) for i in a0.symbolic_shape[1:])
+                rvalue = '(%s (*)%s) %s[%s]' % (a1._C_typedata, shape, a1.name,
+                                                a1.dim.name)
+                lvalue = c.AlignedAttribute(
+                    a0._data_alignment,
+                    c.Value(a0._C_typedata, '(*restrict %s)%s' % (a0.name, shape))
+                )
+            else:
+                rvalue = '(%s *) %s[%s]' % (a1._C_typedata, a1.name, a1.dim.name)
+                lvalue = c.AlignedAttribute(
+                    a0._data_alignment, c.Value(a0._C_typedata, '*restrict %s' % o.flat)
+                )
+        else:
+            rvalue = '%s->%s' % (a1.name, a0._C_name)
+            lvalue = c.Value(a0._C_typename, a0._C_name)
+        return c.Initializer(lvalue, rvalue)
+
     def visit_Block(self, o):
-        body = flatten(self._visit(i) for i in o.children)
+        body = flatten(self._visit(i) for i in self._blankline_logic(o.children))
         return c.Module(o.header + (c.Block(body),) + o.footer)
 
     def visit_List(self, o):
-        body = flatten(self._visit(i) for i in o.children)
+        body = flatten(self._visit(i) for i in self._blankline_logic(o.children))
         return c.Module(o.header + (c.Collection(body),) + o.footer)
 
     def visit_Section(self, o):
-        header = c.Comment("Begin %s" % o.name)
         body = flatten(self._visit(i) for i in o.children)
-        footer = c.Comment("End %s" % o.name)
-        return c.Module([header] + body + [footer])
+        if o.is_subsection:
+            header = []
+            footer = []
+        else:
+            header = [c.Comment("Begin %s" % o.name)]
+            footer = [c.Comment("End %s" % o.name)]
+        return c.Module(header + body + footer)
 
     def visit_Element(self, o):
         return o.element
 
+    def visit_Definition(self, o):
+        f = o.function
+
+        if f.is_AbstractFunction:
+            if f.is_PointerArray:
+                v = "*"
+            else:
+                v = ""
+            if o.shape is None:
+                v = "%s*%s" % (v, f._C_name)
+            else:
+                v = "%s%s%s" % (v, f._C_name, o.shape)
+            if o.qualifier is not None:
+                v = "%s %s" % (v, o.qualifier)
+        elif f.is_LocalObject:
+            v = f._C_name
+        else:
+            assert False
+
+        if o.constructor_args:
+            v = MultilineCall(v, o.constructor_args, True)
+
+        v = c.Value(f._C_typedata, v)
+
+        if o.initvalue is not None:
+            v = c.Initializer(v, o.initvalue)
+
+        return v
+
     def visit_Expression(self, o):
-        return c.Assign(ccode(o.expr.lhs, dtype=o.dtype),
-                        ccode(o.expr.rhs, dtype=o.dtype))
+        lhs = ccode(o.expr.lhs, dtype=o.dtype)
+        rhs = ccode(o.expr.rhs, dtype=o.dtype)
+
+        if o.init:
+            code = c.Initializer(c.Value(o.expr.lhs._C_typename, lhs), rhs)
+        else:
+            code = c.Assign(lhs, rhs)
+
+        if o.pragmas:
+            code = c.Module(list(o.pragmas) + [code])
+
+        return code
 
     def visit_AugmentedExpression(self, o):
-        return c.Statement("%s %s= %s" % (ccode(o.expr.lhs, dtype=o.dtype), o.op,
-                                          ccode(o.expr.rhs, dtype=o.dtype)))
+        code = c.Statement("%s %s= %s" % (ccode(o.expr.lhs, dtype=o.dtype), o.op,
+                           ccode(o.expr.rhs, dtype=o.dtype)))
+        if o.pragmas:
+            code = c.Module(list(o.pragmas) + [code])
+        return code
 
-    def visit_LocalExpression(self, o):
-        if o.write.is_Array:
-            lhs = '%s%s' % (
-                o.expr.lhs.name,
-                ''.join(['[%s]' % d.symbolic_size for d in o.expr.lhs.dimensions])
-            )
-        else:
-            lhs = ccode(o.expr.lhs, dtype=o.dtype)
-
-        return c.Initializer(c.Value(o.expr.lhs._C_typedata, lhs),
-                             ccode(o.expr.rhs, dtype=o.dtype))
-
-    def visit_ForeignExpression(self, o):
-        return c.Statement(ccode(o.expr))
-
-    def visit_Call(self, o):
+    def visit_Call(self, o, nested_call=False):
+        retobj = o.retobj
+        cast = o.cast and retobj._C_typename
         arguments = self._args_call(o.arguments)
-        code = '%s(%s)' % (o.name, ','.join(arguments))
-        return c.Statement(code)
+        if retobj is None:
+            return MultilineCall(o.name, arguments, nested_call, o.is_indirect, cast)
+        else:
+            call = MultilineCall(o.name, arguments, True, o.is_indirect, cast)
+            if retobj.is_AbstractFunction:
+                return c.Initializer(c.Value(retobj._C_typename, retobj._C_name), call)
+            else:
+                return c.Initializer(c.Value(retobj._C_typedata, ccode(retobj)), call)
 
     def visit_Conditional(self, o):
-        then_body = c.Block(self._visit(o.then_body))
-        if o.else_body:
-            else_body = c.Block(self._visit(o.else_body))
+        try:
+            then_body, else_body = self._blankline_logic(o.children)
+        except ValueError:
+            # Some special subclasses of Conditional such as ThreadedProdder
+            # have zero children actually
+            then_body, else_body = o.then_body, o.else_body
+        then_body = c.Block(self._visit(then_body))
+        if else_body:
+            else_body = c.Block(self._visit(else_body))
             return c.If(ccode(o.condition), then_body, else_body)
         else:
             return c.If(ccode(o.condition), then_body)
 
     def visit_Iteration(self, o):
-        body = flatten(self._visit(i) for i in o.children)
+        body = flatten(self._visit(i) for i in self._blankline_logic(o.children))
 
         _min = o.limits[0]
         _max = o.limits[1]
@@ -316,11 +422,17 @@ class CGen(Visitor):
 
         return handle
 
+    def visit_Pragma(self, o):
+        if len(o.pragmas) == 1:
+            return o.pragmas[0]
+        else:
+            return c.Collection(o.pragmas)
+
     def visit_While(self, o):
         condition = ccode(o.condition)
         if o.body:
             body = flatten(self._visit(i) for i in o.children)
-            return c.While(condition, body)
+            return c.While(condition, c.Block(body))
         else:
             # Hack: cgen doesn't support body-less while-loops, i.e. `while(...);`
             return c.Statement('while(%s)' % condition)
@@ -328,9 +440,29 @@ class CGen(Visitor):
     def visit_Callable(self, o):
         body = flatten(self._visit(i) for i in o.children)
         decls = self._args_decl(o.parameters)
-        signature = c.FunctionDeclaration(c.Value(o.retval, o.name), decls)
+        prefix = ' '.join(o.prefix + (o.retval,))
+        signature = c.FunctionDeclaration(c.Value(prefix, o.name), decls)
         return c.FunctionBody(signature, c.Block(body))
 
+    def visit_CallableBody(self, o):
+        body = []
+        prev = None
+        for i in o.children:
+            v = self._visit(i)
+            if v:
+                if prev:
+                    body.append(c.Line())
+                prev = v
+                body.extend(as_tuple(v))
+        return c.Collection(body)
+
+    def visit_Lambda(self, o):
+        body = flatten(self._visit(i) for i in o.children)
+        captures = [str(i) for i in o.captures]
+        decls = [i.inline() for i in self._args_decl(o.parameters)]
+        top = c.Line('[%s](%s)' % (', '.join(captures), ', '.join(decls)))
+        return LambdaCollection([top, c.Block(body)])
+
     def visit_HaloSpot(self, o):
         body = flatten(self._visit(i) for i in o.children)
         return c.Collection(body)
@@ -342,7 +474,7 @@ class CGen(Visitor):
         body = flatten(self._visit(i) for i in o.children)
         decls = self._args_decl(o.parameters)
         signature = c.FunctionDeclaration(c.Value(o.retval, o.name), decls)
-        retval = [c.Statement("return 0")]
+        retval = [c.Line(), c.Statement("return 0")]
         kernel = c.FunctionBody(signature, c.Block(body + retval))
 
         # Elemental functions
@@ -350,17 +482,23 @@ class CGen(Visitor):
         efuncs = [blankline]
         for i in o._func_table.values():
             if i.local:
-                esigns.append(c.FunctionDeclaration(c.Value(i.root.retval, i.root.name),
+                prefix = ' '.join(i.root.prefix + (i.root.retval,))
+                esigns.append(c.FunctionDeclaration(c.Value(prefix, i.root.name),
                                                     self._args_decl(i.root.parameters)))
-                efuncs.extend([i.root.ccode, blankline])
+                efuncs.extend([self._visit(i.root), blankline])
 
         # Header files, extra definitions, ...
-        header = [c.Line(i) for i in o._headers]
-        includes = [c.Include(i, system=False) for i in o._includes]
+        header = [c.Define(*i) for i in o._headers] + [blankline]
+        includes = [c.Include(i, system=(False if i.endswith('.h') else True))
+                    for i in o._includes]
         includes += [blankline]
         cdefs = [i._C_typedecl for i in o.parameters if i._C_typedecl is not None]
+        for i in o._func_table.values():
+            if i.local:
+                cdefs.extend([j._C_typedecl for j in i.root.parameters
+                              if j._C_typedecl is not None])
         cdefs = filter_sorted(cdefs, key=lambda i: i.tpname)
-        if o._compiler.src_ext == 'cpp':
+        if o._compiler.src_ext in ('cpp', 'cu'):
             cdefs += [c.Extern('C', signature)]
         cdefs = [i for j in cdefs for i in (j, blankline)]
 
@@ -380,6 +518,9 @@ class FindSections(Visitor):
     Conditionals, Calls, ...).
     """
 
+    def visit_object(self, o, ret=None, queue=None):
+        return ret
+
     def visit_tuple(self, o, ret=None, queue=None):
         if ret is None:
             ret = self.default_retval()
@@ -437,8 +578,7 @@ class MapExprStmts(FindSections):
         ret[o] = as_tuple(queue)
         return ret
 
-    visit_Conditional = FindSections.visit_Node
-
+    visit_Conditional = FindSections.visit_Iteration
     visit_Block = FindSections.visit_Iteration
 
 
@@ -520,9 +660,14 @@ class MapNodes(Visitor):
 
 class FindSymbols(Visitor):
 
+    class Retval(list):
+        def __init__(self, *retvals):
+            elements = filter_ordered(flatten(retvals), key=id)
+            super().__init__(elements)
+
     @classmethod
     def default_retval(cls):
-        return []
+        return cls.Retval()
 
     """
     Find symbols in an Iteration/Expression tree.
@@ -531,65 +676,50 @@ class FindSymbols(Visitor):
     ----------
     mode : str, optional
         Drive the search. Accepted:
-        - ``symbolics``: Collect all AbstractFunction objects, default.
-        - ``free-symbols``: Collect all free symbols.
-        - ``indexeds``: Collect all Indexeds.
-        - ``defines``: Collect all defined (bound) objects.
+        - `symbolics`: Collect all AbstractFunction objects, default
+        - `basics`: Collect all Basic objects
+        - `dimensions`: Collect all Dimensions
+        - `indexeds`: Collect all Indexed objects
+        - `indexedbases`: Collect all IndexedBase objects
+        - `defines`: Collect all defined objects
+        - `defines-aliases`: Collect all defined objects and their aliases
     """
 
-    def _symbolics(e):
-        try:
-            return e.functions
-        except AttributeError:
-            # A SymPy expression
-            return [i for i in e.free_symbols if isinstance(i, AbstractFunction)]
-
-    def _defines(e):
-        try:
-            return as_tuple(e.defines)
-        except AttributeError:
-            return ()
-
     rules = {
-        'symbolics': _symbolics,
-        'free-symbols': lambda e: e.free_symbols,
-        'indexeds': lambda e: [i for i in e.free_symbols if i.is_Indexed],
-        'defines': _defines,
+        'symbolics': lambda n: n.functions,
+        'basics': lambda n: [i for i in n.expr_symbols if isinstance(i, Basic)],
+        'dimensions': lambda n: [i for i in n.expr_symbols if isinstance(i, Dimension)],
+        'indexeds': lambda n: [i for i in n.expr_symbols if i.is_Indexed],
+        'indexedbases': lambda n: [i for i in n.expr_symbols
+                                   if isinstance(i, IndexedBase)],
+        'defines': lambda n: as_tuple(n.defines),
+        'defines-aliases': lambda n: as_tuple(flatten(i._C_aliases for i in n.defines)),
     }
 
     def __init__(self, mode='symbolics'):
-        super(FindSymbols, self).__init__()
-        self.rule = self.rules[mode]
+        super().__init__()
+
+        modes = mode.split('|')
+        if len(modes) == 1:
+            self.rule = self.rules[mode]
+        else:
+            self.rule = lambda n: chain(*[self.rules[mode](n) for mode in modes])
+
+    def _post_visit(self, ret):
+        return sorted(ret, key=lambda i: i.name)
 
     def visit_tuple(self, o):
-        symbols = flatten([self._visit(i) for i in o])
-        return filter_sorted(symbols, key=attrgetter('name'))
+        return self.Retval(*[self._visit(i) for i in o])
 
     visit_list = visit_tuple
 
-    def visit_Iteration(self, o):
-        symbols = flatten([self._visit(i) for i in o.children])
-        symbols += self.rule(o)
-        return filter_sorted(symbols, key=attrgetter('name'))
+    def visit_Node(self, o):
+        return self.Retval(self._visit(o.children), self.rule(o))
 
-    visit_List = visit_Iteration
-
-    def visit_Conditional(self, o):
-        symbols = flatten([self._visit(i) for i in o.children])
-        symbols += self.rule(o)
-        symbols += self.rule(o.condition)
-        return filter_sorted(symbols, key=attrgetter('name'))
-
-    def visit_Expression(self, o):
-        return filter_sorted([f for f in self.rule(o)], key=attrgetter('name'))
-
-    def visit_Call(self, o):
-        symbols = self._visit(o.children)
-        symbols.extend([f for f in self.rule(o)])
-        return filter_sorted(symbols, key=attrgetter('name'))
-
-    visit_PointerCast = visit_Expression
-    visit_Dereference = visit_Expression
+    def visit_Operator(self, o):
+        ret = self._visit(o.body)
+        ret.extend(flatten(self._visit(v) for v in o._func_table.values()))
+        return self.Retval(ret, self.rule(o))
 
 
 class FindNodes(Visitor):
@@ -667,6 +797,14 @@ class IsPerfectIteration(Visitor):
         nomore = nomore or len(o.children) > 1
         return all(self._visit(i, found=found, nomore=nomore) for i in o.children)
 
+    def visit_While(self, o, **kwargs):
+        return False
+
+    def visit_HaloSpot(self, o, found=False, **kwargs):
+        if not found:
+            return False
+        return all(self._visit(i, found=found, nomore=True) for i in o.children)
+
     def visit_List(self, o, found=False, nomore=False):
         nomore = nomore or (found and (len(o.children) > 1) or o.header or o.footer)
         return all(self._visit(i, found=found, nomore=nomore) for i in o.children)
@@ -683,7 +821,7 @@ class IsPerfectIteration(Visitor):
 class Transformer(Visitor):
 
     """
-    Given an Iteration/Expression tree T and a mapper from nodes in T to
+    Given an Iteration/Expression tree T and a mapper M from nodes in T to
     a set of new nodes L, M : N --> L, build a new Iteration/Expression tree T'
     where a node ``n`` in N is replaced with ``M[n]``.
 
@@ -693,9 +831,9 @@ class Transformer(Visitor):
     "extended" by pre-pending to its body the nodes in ``M[n]``.
     """
 
-    def __init__(self, mapper={}, nested=False):
+    def __init__(self, mapper, nested=False):
         super(Transformer, self).__init__()
-        self.mapper = mapper.copy()
+        self.mapper = mapper
         self.nested = nested
 
     def visit_object(self, o, **kwargs):
@@ -738,32 +876,77 @@ class Transformer(Visitor):
         raise ValueError("Cannot apply a Transformer visitor to an Operator directly")
 
 
-class XSubs(Transformer):
+class Uxreplace(Transformer):
     """
-    Transformer that performs substitutions on Expressions
-    in a given tree, akin to SymPy's ``subs``.
+    Apply substitutions to SymPy objects wrapped in IET nodes.
+    This is the IET-equivalent of `uxreplace` in the expressions layer.
 
     Parameters
     ----------
-    mapper : dict, optional
+    mapper : dict
         The substitution rules.
-    replacer : callable, optional
-        An ad-hoc function to perform the substitution. Defaults to ``uxreplace``.
     """
 
-    def __init__(self, mapper=None, replacer=None):
-        super(XSubs, self).__init__()
-        self.replacer = replacer or (lambda i: uxreplace(i, mapper))
+    def visit_Expression(self, o):
+        return o._rebuild(expr=uxreplace(o.expr, self.mapper))
+
+    def visit_Call(self, o):
+        arguments = [uxreplace(i, self.mapper) for i in o.arguments]
+        return o._rebuild(arguments=arguments)
 
     def visit_Conditional(self, o):
-        condition = self.replacer(o.condition)
+        condition = uxreplace(o.condition, self.mapper)
         then_body = self._visit(o.then_body)
         else_body = self._visit(o.else_body)
         return o._rebuild(condition=condition, then_body=then_body, else_body=else_body)
 
-    def visit_Expression(self, o):
-        return o._rebuild(expr=self.replacer(o.expr))
+    visit_ThreadedProdder = visit_Call
 
 
+# Utils
+
 def printAST(node, verbose=True):
     return PrintAST(verbose=verbose)._visit(node)
+
+
+class LambdaCollection(c.Collection):
+    pass
+
+
+class MultilineCall(c.Generable):
+
+    def __init__(self, name, arguments, is_expr=False, is_indirect=False, cast=None):
+        self.name = name
+        self.arguments = as_tuple(arguments)
+        self.is_expr = is_expr
+        self.is_indirect = is_indirect
+        self.cast = cast
+
+    def generate(self):
+        if not self.is_indirect:
+            tip = "%s(" % self.name
+        else:
+            tip = "%s%s" % (self.name, ',' if self.arguments else '')
+        processed = []
+        for i in self.arguments:
+            if isinstance(i, (MultilineCall, LambdaCollection)):
+                lines = list(i.generate())
+                if len(lines) > 1:
+                    yield tip + ",".join(processed + [lines[0]])
+                    for line in lines[1:-1]:
+                        yield line
+                    tip = ""
+                    processed = [lines[-1]]
+                else:
+                    assert len(lines) == 1
+                    processed.append(lines[0])
+            else:
+                processed.append(str(i))
+        tip = tip + ",".join(processed)
+        if not self.is_indirect:
+            tip += ")"
+        if not self.is_expr:
+            tip += ";"
+        if self.cast:
+            tip = '(%s)%s' % (self.cast, tip)
+        yield tip
