@@ -5,10 +5,10 @@ from sympy import S
 
 from devito.ir.support.space import Backward, IterationSpace
 from devito.ir.support.vector import LabeledVector, Vector
-from devito.symbolics import retrieve_terminals, q_constant, q_affine
-from devito.tools import (EnrichedTuple, Tag, as_tuple, is_integer,
-                          filter_sorted, flatten, memoized_meth, memoized_generator)
-from devito.types import Dimension
+from devito.symbolics import retrieve_terminals, q_constant, q_affine, q_terminal
+from devito.tools import (Tag, as_tuple, is_integer, filter_sorted, flatten,
+                          memoized_meth, memoized_generator)
+from devito.types import Dimension, DimensionTuple
 
 __all__ = ['IterationInstance', 'TimedAccess', 'Scope']
 
@@ -17,6 +17,7 @@ class IndexMode(Tag):
     """Tag for access functions."""
     pass
 AFFINE = IndexMode('affine')  # noqa
+REGULAR = IndexMode('regular')
 IRREGULAR = IndexMode('irregular')
 
 
@@ -61,38 +62,40 @@ class IterationInstance(LabeledVector):
         * obj5 is irregular, as two findices appear in the same index function.
     """
 
-    def __new__(cls, indexed):
-        findices = tuple(indexed.function.dimensions)
+    def __new__(cls, access):
+        findices = tuple(access.function.dimensions)
         if len(findices) != len(set(findices)):
             raise ValueError("Illegal non-unique `findices`")
-        return super(IterationInstance, cls).__new__(cls,
-                                                     list(zip(findices, indexed.indices)))
+        return super().__new__(cls, list(zip(findices, access.indices)))
 
     def __hash__(self):
-        return super(IterationInstance, self).__hash__()
+        return super().__hash__()
 
     @cached_property
     def index_mode(self):
         retval = []
         for i, fi in zip(self, self.findices):
-            dims = {i for i in i.free_symbols if isinstance(i, Dimension)}
+            dims = {j for j in i.free_symbols if isinstance(j, Dimension)}
             if len(dims) == 0 and q_constant(i):
                 retval.append(AFFINE)
             elif len(dims) == 1:
                 candidate = dims.pop()
-                if fi in candidate._defines and q_affine(i, candidate):
-                    retval.append(AFFINE)
+                if fi._defines & candidate._defines:
+                    if q_affine(i, candidate):
+                        retval.append(AFFINE)
+                    else:
+                        retval.append(REGULAR)
                 else:
                     retval.append(IRREGULAR)
             else:
                 retval.append(IRREGULAR)
-        return tuple(retval)
+        return DimensionTuple(*retval, getters=self.findices)
 
     @cached_property
     def aindices(self):
         retval = []
         for i, fi in zip(self, self.findices):
-            dims = {i for i in i.free_symbols if isinstance(i, Dimension)}
+            dims = {j for j in i.free_symbols if isinstance(j, Dimension)}
             if len(dims) == 1:
                 retval.append(dims.pop())
             elif isinstance(i, Dimension):
@@ -101,46 +104,53 @@ class IterationInstance(LabeledVector):
                 retval.append(fi)
             else:
                 retval.append(None)
-        return EnrichedTuple(*retval, getters=self.findices)
+        return DimensionTuple(*retval, getters=self.findices)
 
     @property
     def findices(self):
         return self.labels
 
     @cached_property
-    def findices_affine(self):
-        return tuple(fi for fi, im in zip(self.findices, self.index_mode) if im is AFFINE)
+    def defined_findices_affine(self):
+        ret = set()
+        for fi, im in zip(self.findices, self.index_mode):
+            if im is AFFINE:
+                ret.update(fi._defines)
+        return ret
 
     @cached_property
-    def findices_irregular(self):
-        return tuple(fi for fi, im in zip(self.findices, self.index_mode)
-                     if im is IRREGULAR)
+    def defined_findices_irregular(self):
+        ret = set()
+        for fi, im in zip(self.findices, self.index_mode):
+            if im is IRREGULAR:
+                ret.update(fi._defines)
+        return ret
 
     def affine(self, findices):
         """
         Return True if all of the provided findices appear in self and are
         affine, False otherwise.
         """
-        return set(as_tuple(findices)).issubset(set(self.findices_affine))
+        return set(as_tuple(findices)).issubset(self.defined_findices_affine)
 
     def affine_if_present(self, findices):
         """
         Return False if any of the provided findices appears in self and
         is not affine, True otherwise.
         """
-        findices = as_tuple(findices)
-        return (set(findices) & set(self.findices)).issubset(set(self.findices_affine))
+        present_findices = set(as_tuple(findices)) & set(self.findices)
+        return present_findices.issubset(self.defined_findices_affine)
 
     def irregular(self, findices):
         """
         Return True if all of the provided findices appear in self and are
         irregular, False otherwise.
         """
-        return set(as_tuple(findices)).issubset(set(self.findices_irregular))
+        return set(as_tuple(findices)).issubset(self.defined_findices_irregular)
 
     @cached_property
     def is_regular(self):
-        return all(i is AFFINE for i in self.index_mode)
+        return all(i in (AFFINE, REGULAR) for i in self.index_mode)
 
     @property
     def is_irregular(self):
@@ -169,14 +179,14 @@ class TimedAccess(IterationInstance):
 
     _modes = ('R', 'W', 'RI', 'WI')
 
-    def __new__(cls, indexed, mode, timestamp, ispace=None):
+    def __new__(cls, access, mode, timestamp, ispace=None):
         assert mode in cls._modes
         assert is_integer(timestamp)
 
-        obj = super(TimedAccess, cls).__new__(cls, indexed)
+        obj = super().__new__(cls, access)
 
-        obj.indexed = indexed
-        obj.function = indexed.function
+        obj.access = access
+        obj.function = access.function
         obj.mode = mode
         obj.timestamp = timestamp
 
@@ -196,12 +206,12 @@ class TimedAccess(IterationInstance):
         # which might require expensive comparisons of Vector entries (i.e.,
         # SymPy expressions)
 
-        return (self.indexed is other.indexed and  # => self.function is other.function
+        return (self.access is other.access and  # => self.function is other.function
                 self.mode == other.mode and
                 self.ispace == other.ispace)
 
     def __hash__(self):
-        return super(TimedAccess, self).__hash__()
+        return super().__hash__()
 
     @property
     def name(self):
@@ -245,7 +255,7 @@ class TimedAccess(IterationInstance):
 
     @cached_property
     def is_regular(self):
-        if not super(TimedAccess, self).is_regular:
+        if not super().is_regular:
             return False
 
         # The order of the `aindices` must match the order of the iteration
@@ -265,7 +275,7 @@ class TimedAccess(IterationInstance):
             raise TypeError("Cannot compare due to mismatching `direction`")
         if self.intervals != other.intervals:
             raise TypeError("Cannot compare due to mismatching `intervals`")
-        return super(TimedAccess, self).__lt__(other)
+        return super().__lt__(other)
 
     def lex_eq(self, other):
         return self.timestamp == other.timestamp
@@ -489,7 +499,7 @@ class Dependence(object):
 
     @cached_property
     def is_increment(self):
-        return self.source.is_increment and self.sink.is_increment
+        return self.source.is_increment or self.sink.is_increment
 
     @cached_property
     def is_regular(self):
@@ -556,10 +566,9 @@ class Dependence(object):
         Return True if ``dim`` may represent a reduction dimension for
         ``self``, False otherwise.
         """
-        test0 = self.is_increment
-        test1 = self.is_regular
-        test2 = all(i not in self._defined_findices for i in dim._defines)
-        return test0 and test1 and test2
+        return (self.is_increment and
+                self.is_regular and
+                not (dim._defines & self._defined_findices))
 
     @memoized_meth
     def is_reduce_atmost(self, dim=None):
@@ -568,7 +577,8 @@ class Dependence(object):
         represent a reduction dimension for ``self`` or if `self`` is definitely
         independent of ``dim``, False otherwise.
         """
-        return self.is_reduce(dim) or self.is_indep(dim)
+        return (not (dim._defines & self._defined_findices)
+                or self.is_indep(dim))
 
     @memoized_meth
     def is_indep(self, dim=None):
@@ -580,17 +590,23 @@ class Dependence(object):
                 # Note: we cannot just return `self.distance == 0` as an irregular
                 # source/sink might mean that an array is actually accessed indirectly
                 # (e.g., A[B[i]]), thus there would be no guarantee on independence
-                return False
-            elif dim is None:
+                # The only hope at this point is that the irregularity is not along
+                # `dim` *and* that the distance along `dim` is 0 (e.g., dim=x and
+                # f[x, g[x, y]] -> f[x, h[x, y]])
+                try:
+                    return (self.source.affine(dim) and
+                            self.sink.affine(dim) and
+                            self.distance_mapper[dim] == 0)
+                except KeyError:
+                    # `dim is None` or anything not in `self._defined_findices`
+                    return False
+            if dim is None:
                 return self.distance == 0
             else:
-                # Note: below, `i in self._defined_findices` is to check whether `i`
-                # is actually (one of) the reduction dimension(s), in which case
-                # `self` would indeed be a dimension-dependent dependence
-                test0 = (not self.is_increment or
-                         any(i in self._defined_findices for i in dim._defines))
-                test1 = len(self.cause & dim._defines) == 0
-                return test0 and test1
+                # The only hope at this point is that `dim` appears in the findices
+                # with distance 0 (that is, it's not the cause of the dependence)
+                return (any(i in self._defined_findices for i in dim._defines) and
+                        len(self.cause & dim._defines) == 0)
         except TypeError:
             # Conservatively assume this is not dimension-independent
             return False
@@ -683,9 +699,10 @@ class Scope(object):
                 v.append(TimedAccess(j, mode, i, e.ispace))
 
             # Write
-            v = self.writes.setdefault(e.lhs.function, [])
-            mode = 'WI' if e.is_Increment else 'W'
-            v.append(TimedAccess(e.lhs, mode, i, e.ispace))
+            if q_terminal(e.lhs):
+                v = self.writes.setdefault(e.lhs.function, [])
+                mode = 'WI' if e.is_Increment else 'W'
+                v.append(TimedAccess(e.lhs, mode, i, e.ispace))
 
             # If an increment, we got one implicit read
             if e.is_Increment:
@@ -754,6 +771,10 @@ class Scope(object):
         groups = list(self.reads.values()) + list(self.writes.values())
         return [i for group in groups for i in group]
 
+    @cached_property
+    def indexeds(self):
+        return tuple(i.access for i in self.accesses if i.access.is_Indexed)
+
     @cached_property
     def functions(self):
         return set(self.reads) | set(self.writes)
