@@ -16,15 +16,23 @@ SEQUENTIAL = Property('sequential')
 PARALLEL = Property('parallel')
 """A fully parallel Dimension."""
 
+PARALLEL_IF_ATOMIC = Property('parallel_if_atomic')
+"""
+A parallel Dimension that requires atomic reductions. This is weaker than PARALLEL.
+"""
+
+PARALLEL_IF_PVT = Property('parallel_if_private')
+"""
+A parallel Dimension that requires all compiler-generated AbstractFunctions be
+privatized at the thread level. This is weaker than PARALLEL.
+"""
+
 PARALLEL_INDEP = Property('parallel=')
 """
 A fully parallel Dimension, where all dependences have dependence distance
-equals to 0 (i.e., the distance vector is '='). This is stronger than PARALLEL.
+equals to 0 (i.e., the distance vector is '=').
 """
 
-PARALLEL_IF_ATOMIC = Property('parallel_if_atomic')
-"""A parallel Dimension with local reductions. This is weaker than PARALLEL."""
-
 COLLAPSED = lambda i: Property('collapsed', i)
 """Collapsed Dimensions."""
 
@@ -34,6 +42,9 @@ VECTORIZED = Property('vector-dim')
 TILABLE = Property('tilable')
 """A fully parallel Dimension that would benefit from tiling (or "blocking")."""
 
+SKEWABLE = Property('skewable')
+"""A fully parallel Dimension that would benefit from wavefront/skewed tiling."""
+
 ROUNDABLE = Property('roundable')
 """
 A Dimension whose upper limit may be rounded up to a multiple of the SIMD
@@ -55,9 +66,11 @@ def normalize_properties(*args):
         return args[0]
 
     if any(SEQUENTIAL in p for p in args):
-        drop = {PARALLEL, PARALLEL_INDEP, PARALLEL_IF_ATOMIC}
+        drop = {PARALLEL, PARALLEL_INDEP, PARALLEL_IF_ATOMIC, PARALLEL_IF_PVT}
     elif any(PARALLEL_IF_ATOMIC in p for p in args):
         drop = {PARALLEL, PARALLEL_INDEP}
+    elif any(PARALLEL_IF_PVT in p for p in args):
+        drop = {PARALLEL}
     elif any(PARALLEL_INDEP not in p for p in args):
         drop = {PARALLEL_INDEP}
     else:
@@ -68,3 +81,7 @@ def normalize_properties(*args):
         properties.update(p - drop)
 
     return properties
+
+
+def relax_properties(properties):
+    return frozenset(properties - {PARALLEL_INDEP, ROUNDABLE})
