@@ -1,18 +1,23 @@
 import abc
 from collections import OrderedDict
 from functools import reduce
+from itertools import chain
 from operator import mul
 
 from cached_property import cached_property
 from sympy import Expr
 
 from devito.ir.support.vector import Vector, vmin, vmax
-from devito.tools import (PartialOrderTuple, as_list, as_tuple, filter_ordered,
-                          frozendict, toposort, is_integer)
-from devito.types import Dimension
+from devito.tools import (PartialOrderTuple, Stamp, as_list, as_tuple, filter_ordered,
+                          flatten, frozendict, is_integer, toposort)
+from devito.types import Dimension, ModuloDimension
 
-__all__ = ['NullInterval', 'Interval', 'IntervalGroup', 'IterationSpace', 'DataSpace',
-           'Forward', 'Backward', 'Any']
+__all__ = ['NullInterval', 'Interval', 'IntervalGroup', 'IterationSpace',
+           'DataSpace', 'Forward', 'Backward', 'Any']
+
+
+# The default Stamp, used by all new Intervals
+S0 = Stamp()
 
 
 class AbstractInterval(object):
@@ -26,7 +31,7 @@ class AbstractInterval(object):
     is_Null = False
     is_Defined = False
 
-    def __init__(self, dim, stamp=0):
+    def __init__(self, dim, stamp=S0):
         self.dim = dim
         self.stamp = stamp
 
@@ -80,7 +85,7 @@ class NullInterval(AbstractInterval):
     is_Null = True
 
     def __repr__(self):
-        return "%s[Null]<%d>" % (self.dim, self.stamp)
+        return "%s[Null]%s" % (self.dim, self.stamp)
 
     def __hash__(self):
         return hash(self.dim)
@@ -112,25 +117,27 @@ class Interval(AbstractInterval):
 
     An Interval defines the compact region
 
-        ``[dim.symbolic_min + lower, dim.symbolic_max + upper]``
-
-    The size of the Interval is defined as the number of points iterated over
-    through ``dim``, namely
-
-        ``(dim.symbolic_max + upper - dim.symbolic_min - lower + 1) / dim.symbolic_incr``
+        [dim.symbolic_min + lower, dim.symbolic_max + upper]
     """
 
     is_Defined = True
 
-    def __init__(self, dim, lower, upper, stamp=0):
-        assert is_integer(lower) or isinstance(lower, Expr)
-        assert is_integer(upper) or isinstance(upper, Expr)
+    def __init__(self, dim, lower, upper, stamp=S0):
         super(Interval, self).__init__(dim, stamp)
-        self.lower = lower
-        self.upper = upper
+
+        try:
+            self.lower = int(lower)
+        except TypeError:
+            assert isinstance(lower, Expr)
+            self.lower = lower
+        try:
+            self.upper = int(upper)
+        except TypeError:
+            assert isinstance(upper, Expr)
+            self.upper = upper
 
     def __repr__(self):
-        return "%s[%s,%s]<%d>" % (self.dim, self.lower, self.upper, self.stamp)
+        return "%s[%s,%s]%s" % (self.dim, self.lower, self.upper, self.stamp)
 
     def __hash__(self):
         return hash((self.dim, self.offsets))
@@ -148,9 +155,53 @@ class Interval(AbstractInterval):
 
     @cached_property
     def size(self):
-        upper_extreme = self.dim.symbolic_max + self.upper
-        lower_extreme = self.dim.symbolic_min + self.lower
-        return (upper_extreme - lower_extreme + 1) / self.dim.symbolic_incr
+        """
+        The Interval size, defined as the number of points iterated over through
+        ``self.dim``, namely
+
+            (dim.symbolic_max + upper - dim.symbolic_min - lower + 1) / dim.symbolic_incr
+
+        Notes
+        -----
+        The Interval size is typically a function of several symbols (e.g.,
+        `self.dim.symbolic_max`), and all such symbols must be mappable
+        to actual numbers at `op.apply` time (i.e., the runtime values).
+        When `self.dim` is an "unstructured Dimension", such as ModuloDimension,
+        things can get nasty since the symbolic min/max/incr can literally be
+        anything (any expression involving any Dimension/symbol/...), which makes
+        it extremely complicated to numerically compute the size. However,
+        the compiler only uses such unstructured Dimensions in well defined
+        circumstances, which we explicitly handle here. Ultimately, therefore,
+        we return a `size` that is made up of known symbols.
+        """
+        if self.dim.is_Custom and isinstance(self.dim.symbolic_min, ModuloDimension):
+            # Special case 1)
+            # Caused by the performance option `cire-rotate=True`
+            d = self.dim.symbolic_min
+            n = d.parent.symbolic_size
+
+            # Iteration 0:
+            assert is_integer(d.symbolic_min)
+            assert is_integer(d.symbolic_incr)
+            assert is_integer(self.dim.symbolic_max)
+            assert self.lower == self.upper
+            npoints = self.dim.symbolic_max - d.symbolic_min + 1
+            # Iterations [1, ..., n-1]:
+            assert d.symbolic_incr == self.dim.symbolic_max
+            npoints += 1 * (n-1)
+            npoints /= n
+        else:
+            # Typically we end up here (Dimension, SubDimension, BlockDimension)
+            upper_extreme = self.dim.symbolic_max + self.upper
+            lower_extreme = self.dim.symbolic_min + self.lower
+
+            npoints = (upper_extreme - lower_extreme + 1)
+
+        return npoints / self.dim.symbolic_incr
+
+    @cached_property
+    def min_size(self):
+        return self.upper - self.lower + 1
 
     @property
     def relaxed(self):
@@ -204,17 +255,27 @@ class Interval(AbstractInterval):
 
     def lift(self, v=None):
         if v is None:
-            v = self.stamp + 1
+            v = Stamp()
         return Interval(self.dim, self.lower, self.upper, v)
 
     def reset(self):
-        return Interval(self.dim, self.lower, self.upper, 0)
+        return Interval(self.dim, self.lower, self.upper, S0)
 
     def switch(self, d):
         return Interval(d, self.lower, self.upper, self.stamp)
 
-    def translate(self, v):
-        return Interval(self.dim, self.lower + v, self.upper + v, self.stamp)
+    def translate(self, v0=0, v1=None):
+        if v1 is None:
+            v1 = v0
+        return Interval(self.dim, self.lower + v0, self.upper + v1, self.stamp)
+
+    def promote(self, cond):
+        if cond(self.dim):
+            try:
+                return self.switch(self.dim.parent).promote(cond)
+            except AttributeError:
+                pass
+        return self
 
 
 class IntervalGroup(PartialOrderTuple):
@@ -289,8 +350,7 @@ class IntervalGroup(PartialOrderTuple):
         >>> ig1 = IntervalGroup([Interval(x, 2, -2), Interval(y, 3, -3)])
         >>> ig2 = IntervalGroup([Interval(y, 2, -2), Interval(z, 1, -1)])
 
-        >>> IntervalGroup.generate('intersection', ig0, ig1, ig2)
-        IntervalGroup[x[2,-2]<0>, y[3,-3]<0>, z[1,-1]<0>]
+        >>> ig = IntervalGroup.generate('intersection', ig0, ig1, ig2)
         """
         mapper = {}
         for ig in interval_groups:
@@ -306,10 +366,6 @@ class IntervalGroup(PartialOrderTuple):
         relations = set().union(*[ig.relations for ig in interval_groups])
         return IntervalGroup(intervals, relations=relations)
 
-    @cached_property
-    def relaxed(self):
-        return IntervalGroup.generate('union', IntervalGroup(i.relaxed for i in self))
-
     def is_compatible(self, o):
         """
         Two IntervalGroups are compatible iff they can be ordered according
@@ -356,6 +412,21 @@ class IntervalGroup(PartialOrderTuple):
         intervals = [i.subtract(mapper.get(i.dim, NullInterval(i.dim))) for i in self]
         return IntervalGroup(intervals, relations=(self.relations | o.relations))
 
+    def relaxed(self, d=None):
+        d = set(self.dimensions if d is None else as_tuple(d))
+        intervals = [i.relaxed if i.dim in d else i for i in self]
+        return IntervalGroup(intervals, relations=self.relations)
+
+    def promote(self, cond):
+        intervals = IntervalGroup([i.promote(cond) for i in self],
+                                  relations=self.relations)
+
+        # There could be duplicate Dimensions at this point, so we sum up the Intervals
+        # defined over the same Dimension to produce a well-defined IntervalGroup
+        intervals = IntervalGroup.generate('add', intervals)
+
+        return intervals
+
     def drop(self, d):
         # Dropping
         dims = set().union(*[i._defines for i in as_tuple(d)])
@@ -374,7 +445,7 @@ class IntervalGroup(PartialOrderTuple):
         return IntervalGroup([i.zero() if i.dim in d else i for i in self],
                              relations=self.relations)
 
-    def lift(self, d, v=None):
+    def lift(self, d=None, v=None):
         d = set(self.dimensions if d is None else as_tuple(d))
         return IntervalGroup([i.lift(v) if i.dim._defines & d else i for i in self],
                              relations=self.relations)
@@ -382,6 +453,18 @@ class IntervalGroup(PartialOrderTuple):
     def reset(self):
         return IntervalGroup([i.reset() for i in self], relations=self.relations)
 
+    def switch(self, d0, d1):
+        intervals = [i.switch(d1) if i.dim is d0 else i for i in self]
+
+        # Update relations too
+        relations = {tuple(d1 if i is d0 else i for i in r) for r in self.relations}
+
+        return IntervalGroup(intervals, relations=relations)
+
+    def translate(self, d, v0=0, v1=None):
+        intervals = [i.translate(v0, v1) if i.dim in as_tuple(d) else i for i in self]
+        return IntervalGroup(intervals, relations=self.relations)
+
     def index(self, key):
         if isinstance(key, Interval):
             return super(IntervalGroup, self).index(key)
@@ -405,7 +488,7 @@ class IntervalGroup(PartialOrderTuple):
         for i in self:
             if i.dim is key:
                 return i
-            if key.is_NonlinearDerived and i.dim is key.parent:
+            if key.is_NonlinearDerived and i.dim in key._defines:
                 # NonlinearDerived Dimensions cannot appear in iteration Intervals,
                 # but their parent can
                 return i
@@ -448,10 +531,10 @@ class IterationInterval(object):
     An Interval associated with an IterationDirection.
     """
 
-    def __init__(self, interval, direction, sub_iterators):
+    def __init__(self, interval, sub_iterators, direction):
         self.interval = interval
-        self.direction = direction
         self.sub_iterators = sub_iterators
+        self.direction = direction
 
     def __repr__(self):
         return "%s%s" % (self.interval, self.direction)
@@ -462,6 +545,10 @@ class IterationInterval(object):
     def __hash__(self):
         return hash((self.interval, self.direction))
 
+    @property
+    def args(self):
+        return (self.interval, self.sub_iterators, self.direction)
+
     @property
     def dim(self):
         return self.interval.dim
@@ -470,6 +557,10 @@ class IterationInterval(object):
     def offsets(self):
         return self.interval.offsets
 
+    @property
+    def size(self):
+        return self.interval.size
+
 
 class Space(object):
 
@@ -498,6 +589,13 @@ class Space(object):
     def __hash__(self):
         return hash(self.intervals)
 
+    def __len__(self):
+        return len(self.intervals)
+
+    def __iter__(self):
+        for i in self.intervals:
+            yield i
+
     @property
     def intervals(self):
         return self._intervals
@@ -513,7 +611,7 @@ class Space(object):
     @property
     def dimension_map(self):
         """
-        Map between the Space Dimensions and the size of their iterated region.
+        Map between the Space Dimensions and the size of their Interval.
         """
         return OrderedDict([(i.dim, i.size) for i in self.intervals])
 
@@ -521,20 +619,20 @@ class Space(object):
 class DataSpace(Space):
 
     """
-    Represent a data space as an enriched Space.
+    Represent a data space as a Space with additional metadata and operations.
 
     Parameters
     ----------
     intervals : tuple of Intervals
         Data space description.
-    parts : dict
+    parts : dict, optional
         A mapper from Functions to IntervalGroup, describing the individual
         components of the data space.
     """
 
-    def __init__(self, intervals, parts):
+    def __init__(self, intervals, parts=None):
         super(DataSpace, self).__init__(intervals)
-        self._parts = frozendict(parts)
+        self._parts = frozendict(parts or {})
 
     def __eq__(self, other):
         return isinstance(other, DataSpace) and\
@@ -547,27 +645,21 @@ class DataSpace(Space):
     def union(cls, *others):
         if not others:
             return DataSpace(IntervalGroup(), {})
+
         intervals = IntervalGroup.generate('union', *[i.intervals for i in others])
+
         parts = {}
         for i in others:
             for k, v in i.parts.items():
                 parts.setdefault(k, []).append(v)
         parts = {k: IntervalGroup.generate('union', *v) for k, v in parts.items()}
+
         return DataSpace(intervals, parts)
 
     @property
     def parts(self):
         return self._parts
 
-    @cached_property
-    def relaxed(self):
-        """
-        A view of the DataSpace assuming that any SubDimensions entirely span
-        their root Dimension.
-        """
-        return DataSpace(self.intervals.relaxed,
-                         {k: v.relaxed for k, v in self.parts.items()})
-
     def __getitem__(self, key):
         ret = self.intervals[key]
         if ret.is_Null:
@@ -577,42 +669,17 @@ class DataSpace(Space):
                 ret = IntervalGroup()
         return ret
 
-    def zero(self, d=None):
-        intervals = self.intervals.zero(d)
-        parts = {k: v.zero(d) for k, v in self.parts.items()}
-        return DataSpace(intervals, parts)
-
-    def lift(self, d=None):
-        intervals = self.intervals.lift(d)
-        parts = {k: v.lift(d) for k, v in self.parts.items()}
-        return DataSpace(intervals, parts)
-
     def reset(self):
         intervals = self.intervals.reset()
         parts = {k: v.reset() for k, v in self.parts.items()}
+
         return DataSpace(intervals, parts)
 
-    def project(self, cond):
-        """
-        Create a new DataSpace in which only some of the Dimensions in
-        ``self`` are retained. In particular, a dimension ``d`` in ``self``
-        is retained if:
-
-            * either ``cond(d)`` is True (``cond`` is a callable),
-            * or ``d in cond`` is True (``cond`` is an iterable)
-        """
-        if callable(cond):
-            func = cond
-        else:
-            func = lambda i: i in cond
-        intervals = [i for i in self.intervals if func(i.dim)]
-        return DataSpace(intervals, self.parts)
-
 
 class IterationSpace(Space):
 
     """
-    Represent an iteration space as an enriched Space.
+    Represent an iteration space as a Space with additional metadata and operations.
 
     Parameters
     ----------
@@ -627,7 +694,13 @@ class IterationSpace(Space):
 
     def __init__(self, intervals, sub_iterators=None, directions=None):
         super(IterationSpace, self).__init__(intervals)
-        self._sub_iterators = frozendict(sub_iterators or {})
+
+        # Normalize sub-iterators
+        sub_iterators = sub_iterators or {}
+        self._sub_iterators = frozendict([(k, tuple(filter_ordered(as_tuple(v))))
+                                          for k, v in sub_iterators.items()])
+
+        # Normalize directions
         if directions is None:
             self._directions = frozendict([(i.dim, Any) for i in self.intervals])
         else:
@@ -646,10 +719,19 @@ class IterationSpace(Space):
                 self.intervals == other.intervals and
                 self.directions == other.directions)
 
+    def __lt__(self, other):
+        """
+        A rudimentary less-then comparison between two IterationSpaces.
+        """
+        return len(self.itintervals) < len(other.itintervals)
+
     def __hash__(self):
         return hash((super(IterationSpace, self).__hash__(), self.sub_iterators,
                      self.directions))
 
+    def __getitem__(self, key):
+        return self.intervals[key]
+
     @classmethod
     def union(cls, *others):
         if not others:
@@ -684,11 +766,27 @@ class IterationSpace(Space):
 
     def augment(self, sub_iterators):
         """
-        Create a new IterationSpace with additional sub iterators.
+        Create a new IterationSpace with additional sub-iterators.
         """
-        v = {k: as_list(v) for k, v in sub_iterators.items() if k in self.intervals}
-        sub_iterators = {**self.sub_iterators, **v}
-        return IterationSpace(self.intervals, sub_iterators, self.directions)
+        items = dict(self.sub_iterators)
+        for k, v in sub_iterators.items():
+            if k not in self.intervals:
+                continue
+            items[k] = as_list(items.get(k))
+            for i in as_list(v):
+                if i not in items[k]:
+                    items[k].append(i)
+
+        return IterationSpace(self.intervals, items, self.directions)
+
+    def switch(self, d0, d1):
+        intervals = self.intervals.switch(d0, d1)
+        sub_iterators = {d1 if k is d0 else k: v
+                         for k, v in self.sub_iterators.items()}
+        directions = {d1 if k is d0 else k: v
+                      for k, v in self.directions.items()}
+
+        return IterationSpace(intervals, sub_iterators, directions)
 
     def reset(self):
         return IterationSpace(self.intervals.reset(), self.sub_iterators, self.directions)
@@ -710,12 +808,50 @@ class IterationSpace(Space):
         intervals = [i for i in self.intervals if func(i.dim)]
         sub_iterators = {k: v for k, v in self.sub_iterators.items() if func(k)}
         directions = {k: v for k, v in self.directions.items() if func(k)}
+
         return IterationSpace(intervals, sub_iterators, directions)
 
+    def zero(self, d=None):
+        intervals = self.intervals.zero(d)
+
+        return IterationSpace(intervals, self.sub_iterators, self.directions)
+
     def lift(self, d=None, v=None):
         intervals = self.intervals.lift(d, v)
+
         return IterationSpace(intervals, self.sub_iterators, self.directions)
 
+    def relaxed(self, cond):
+        """
+        Create a new IterationSpace in which certain DerivedDimensions
+        are replaced with their roots. In particular, a Dimension `d` in
+        `self` is relaxed, and therefore replaced with `d.root`, if:
+
+            * either `cond(d)` is true (`cond` is a callable),
+            * or `d in cond` is true (`cond` is an iterable)
+        """
+        if callable(cond):
+            dims = tuple(i.dim for i in self if cond(i.dim))
+        else:
+            dims = as_tuple(cond)
+
+        intervals = self.intervals.relaxed(dims)
+        sub_iterators = {d.root if d in dims else d: v
+                         for d, v in self.sub_iterators.items()}
+        directions = {d.root if d in dims else d: v
+                      for d, v in self.directions.items()}
+
+        return IterationSpace(intervals, sub_iterators, directions)
+
+    def promote(self, cond):
+        intervals = self.intervals.promote(cond)
+        sub_iterators = {i.promote(cond).dim: self.sub_iterators.get(i.dim, ())
+                         for i in self.intervals}
+        directions = {i.promote(cond).dim: self.directions.get(i.dim, ())
+                      for i in self.intervals}
+
+        return IterationSpace(intervals, sub_iterators, directions)
+
     def is_compatible(self, other):
         """
         A relaxed version of ``__eq__``, in which only non-derived dimensions
@@ -727,6 +863,13 @@ class IterationSpace(Space):
     def is_forward(self, dim):
         return self.directions[dim] is Forward
 
+    def is_sub_iterator(self, dim):
+        return dim in chain(*self.sub_iterators.values())
+
+    @property
+    def itdimensions(self):
+        return self.intervals.dimensions
+
     @property
     def relations(self):
         return self.intervals.relations
@@ -742,13 +885,13 @@ class IterationSpace(Space):
     @cached_property
     def itintervals(self):
         return tuple(IterationInterval(
-            i, self.directions[i.dim], self.sub_iterators.get(i.dim)
+            i, self.sub_iterators.get(i.dim), self.directions[i.dim]
         ) for i in self.intervals)
 
     @cached_property
     def dimensions(self):
-        sub_dims = [i.parent for v in self.sub_iterators.values() for i in v]
-        return filter_ordered(self.intervals.dimensions + sub_dims)
+        sub_dims = flatten(i._defines for v in self.sub_iterators.values() for i in v)
+        return filter_ordered(self.itdimensions + sub_dims)
 
     @cached_property
     def nonderived_directions(self):
