@@ -10,7 +10,7 @@ from cgen import Struct, Value
 
 from devito.data import LEFT, CENTER, RIGHT, Decomposition
 from devito.parameters import configuration
-from devito.tools import EnrichedTuple, as_tuple, ctypes_to_cstr, is_integer
+from devito.tools import EnrichedTuple, as_tuple, ctypes_to_cstr, filter_ordered
 from devito.types import CompositeObject, Object
 
 
@@ -22,12 +22,19 @@ try:
     mpi4py.rc(initialize=False, finalize=False)
     from mpi4py import MPI  # noqa
 
+    # Devito can be used in two main ways: as a software or as a library. In the
+    # latter case, it's likely MPI will be initialized and finalized by the
+    # overarching application. So we track who initialized MPI to avoid double
+    # finalization
+    init_by_devito = False
+
     # From the `atexit` documentation: "At normal program termination [...]
     # all functions registered are in last in, first out order.". So, MPI.Finalize
-    # will be called only at the very end, after all cloned communicators
-    # will have been freed
+    # will be called only at the very end and only if necessary, after all cloned
+    # communicators will have been freed
     def cleanup():
-        if MPI.Is_initialized():
+        global init_by_devito
+        if init_by_devito and MPI.Is_initialized() and not MPI.Is_finalized():
             MPI.Finalize()
     atexit.register(cleanup)
 except ImportError:
@@ -173,30 +180,31 @@ class Distributor(AbstractDistributor):
         MPI.COMM_WORLD.
     """
 
-    def __init__(self, shape, dimensions, input_comm=None):
+    def __init__(self, shape, dimensions, input_comm=None, topology=None):
         super(Distributor, self).__init__(shape, dimensions)
 
         if configuration['mpi']:
             # First time we enter here, we make sure MPI is initialized
             if not MPI.Is_initialized():
                 MPI.Init()
+                global init_by_devito
+                init_by_devito = True
 
+            # Note: the cloned communicator doesn't need to be explicitly freed;
+            # mpi4py takes care of that when the object gets out of scope
             self._input_comm = (input_comm or MPI.COMM_WORLD).Clone()
 
-            # Make sure the cloned communicator will be freed up upon exit
-            def cleanup():
-                if self._input_comm is not None:
-                    self._input_comm.Free()
-            atexit.register(cleanup)
-
-            # `MPI.Compute_dims` sets the dimension sizes to be as close to each other
-            # as possible, using an appropriate divisibility algorithm. Thus, in 3D:
-            # * topology[0] >= topology[1] >= topology[2]
-            # * topology[0] * topology[1] * topology[2] == self._input_comm.size
-            # However, `MPI.Compute_dims` is distro-dependent, so we have to enforce
-            # some properties through our own wrapper (e.g., OpenMPI v3 does not
-            # guarantee that 9 ranks are arranged into a 3x3 grid when shape=(9, 9))
-            self._topology = compute_dims(self._input_comm.size, len(shape))
+            if topology is None:
+                # `MPI.Compute_dims` sets the dimension sizes to be as close to each other
+                # as possible, using an appropriate divisibility algorithm. Thus, in 3D:
+                # * topology[0] >= topology[1] >= topology[2]
+                # * topology[0] * topology[1] * topology[2] == self._input_comm.size
+                # However, `MPI.Compute_dims` is distro-dependent, so we have to enforce
+                # some properties through our own wrapper (e.g., OpenMPI v3 does not
+                # guarantee that 9 ranks are arranged into a 3x3 grid when shape=(9, 9))
+                self._topology = compute_dims(self._input_comm.size, len(shape))
+            else:
+                self._topology = topology
 
             if self._input_comm is not input_comm:
                 # By default, Devito arranges processes into a cartesian topology.
@@ -300,29 +308,32 @@ class Distributor(AbstractDistributor):
 
         Parameters
         ----------
-        index : int or list of ints
-            The index, or list of indices, for which the owning MPI rank(s) is
+        index : array of ints
+            The index, or array of indices, for which the owning MPI rank(s) is
             retrieved.
         """
-        assert isinstance(index, (tuple, list))
-        if len(index) == 0:
+        assert isinstance(index, np.ndarray)
+        if index.shape[0] == 0:
             return None
-        elif is_integer(index[0]):
-            # `index` is a single point
-            indices = [index]
-        else:
-            indices = index
-        ret = []
-        for i in indices:
-            assert len(i) == self.ndim
-            found = False
-            for r, j in enumerate(self.all_ranges):
-                if all(v in j[d] for v, d in zip(i, self.dimensions)):
-                    ret.append(r)
-                    found = True
-                    break
-            assert found
-        return as_tuple(ret)
+        elif sum(index.shape) == 1:
+            return index
+
+        assert index.shape[1] == self.ndim
+
+        # Add singleton dimension at the end if only single gridpoint is passed
+        # instead of support.
+        if len(index.shape) == 2:
+            index = np.expand_dims(index, axis=2)
+
+        ret = {}
+        for r, j in enumerate(self.all_ranges):
+            mins = np.array([b[0] for b in j]).reshape(1, -1, 1)
+            maxs = np.array([b[-1] for b in j]).reshape(1, -1, 1)
+            inds = np.where(((index <= maxs) & (index >= mins)).all(axis=1))
+            if inds[0].size == 0:
+                continue
+            ret[r] = filter_ordered(inds[0])
+        return ret
 
     @property
     def neighborhood(self):
@@ -356,11 +367,11 @@ class Distributor(AbstractDistributor):
         # Set up diagonal neighbours
         for i in product([LEFT, CENTER, RIGHT], repeat=self.ndim):
             neighbor = [c + s.val for c, s in zip(self.mycoords, i)]
-            try:
-                ret[i] = self.comm.Get_cart_rank(neighbor)
-            except:
-                # Fallback for MPI ranks at the grid boundary
+
+            if any(c < 0 or c >= s for c, s in zip(neighbor, self.topology)):
                 ret[i] = MPI.PROC_NULL
+            else:
+                ret[i] = self.comm.Get_cart_rank(neighbor)
 
         return ret
 
@@ -473,6 +484,7 @@ class MPICommObject(Object):
         comm_ptr = MPI._addressof(comm)
         comm_val = self.dtype.from_address(comm_ptr)
         self.value = comm_val
+        self.comm = comm
 
     def _arg_values(self, *args, **kwargs):
         grid = kwargs.get('grid', None)
