@@ -7,7 +7,7 @@ from operator import mul
 
 from sympy import Integer
 
-from devito.data import OWNED, HALO, NOPAD, LEFT, CENTER, RIGHT, default_allocator
+from devito.data import OWNED, HALO, NOPAD, LEFT, CENTER, RIGHT
 from devito.ir.equations import DummyEq
 from devito.ir.iet import (Call, Callable, Conditional, Expression, ExpressionBundle,
                            AugmentedExpression, Iteration, List, Prodder, Return,
@@ -15,11 +15,12 @@ from devito.ir.iet import (Call, Callable, Conditional, Expression, ExpressionBu
 from devito.ir.support import AFFINE, PARALLEL
 from devito.mpi import MPI
 from devito.symbolics import (Byref, CondNe, FieldFromPointer, FieldFromComposite,
-                              IndexedPointer, Macro, subs_op_args)
+                              IndexedPointer, Macro, cast_mapper, subs_op_args)
 from devito.tools import OrderedSet, dtype_to_mpitype, dtype_to_ctype, flatten, generator
 from devito.types import Array, Dimension, Symbol, LocalObject, CompositeObject
+from devito.types.dense import AliasFunction
 
-__all__ = ['HaloExchangeBuilder']
+__all__ = ['HaloExchangeBuilder', 'mpi_registry']
 
 
 class HaloExchangeBuilder(object):
@@ -28,19 +29,10 @@ class HaloExchangeBuilder(object):
     Build IET-based routines to implement MPI halo exchange.
     """
 
-    def __new__(cls, mode, **generators):
-        if mode is True or mode == 'basic':
-            obj = object.__new__(BasicHaloExchangeBuilder)
-        elif mode == 'diag':
-            obj = object.__new__(DiagHaloExchangeBuilder)
-        elif mode == 'overlap':
-            obj = object.__new__(OverlapHaloExchangeBuilder)
-        elif mode == 'overlap2':
-            obj = object.__new__(Overlap2HaloExchangeBuilder)
-        elif mode == 'full':
-            obj = object.__new__(FullHaloExchangeBuilder)
-        else:
-            assert False, "unexpected value `mode=%s`" % mode
+    def __new__(cls, mode, sregistry, **generators):
+        obj = object.__new__(mpi_registry[mode])
+
+        obj._sregistry = sregistry
 
         # Unique name generators
         obj._gen_msgkey = generators.get('msg', generator())
@@ -56,6 +48,10 @@ class HaloExchangeBuilder(object):
 
         return obj
 
+    @property
+    def sregistry(self):
+        return self._sregistry
+
     @property
     def efuncs(self):
         return self._efuncs
@@ -98,13 +94,13 @@ class HaloExchangeBuilder(object):
         # Callable for poking the asynchronous progress engine
         key = self._gen_compkey()
         poke = self._make_poke(hs, key, msgs)
-        if poke is not None:
+        if isinstance(poke, Callable):
             self._efuncs.append(poke)
 
         # Callable for compute over the CORE region
         callpoke = self._call_poke(poke)
         compute = self._make_compute(hs, key, msgs, callpoke)
-        if compute is not None:
+        if isinstance(compute, Callable):
             self._efuncs.append(compute)
 
         # Callable for compute over the OWNED region
@@ -112,17 +108,23 @@ class HaloExchangeBuilder(object):
         region = self._regions.setdefault(hs, region)
         callcompute = self._call_compute(hs, compute, msgs)
         remainder = self._make_remainder(hs, key, callcompute, region)
-        if remainder is not None:
+        if isinstance(remainder, Callable):
             self._efuncs.append(remainder)
 
         # Now build up the HaloSpot body, with explicit Calls to the constructed Callables
-        body = [callcompute]
+        haloupdates = []
+        halowaits = []
         for i, (f, hse) in enumerate(hs.fmapper.items()):
             msg = self._msgs[(f, hse)]
             haloupdate, halowait = self._cache_halo[(f.ndim, hse)]
-            body.insert(i, self._call_haloupdate(haloupdate.name, f, hse, msg))
+            haloupdates.append(self._call_haloupdate(haloupdate.name, f, hse, msg))
             if halowait is not None:
-                body.append(self._call_halowait(halowait.name, f, hse, msg))
+                halowaits.append(self._call_halowait(halowait.name, f, hse, msg))
+        body = []
+        body.append(HaloUpdateList(body=haloupdates))
+        if callcompute is not None:
+            body.append(callcompute)
+        body.append(HaloWaitList(body=halowaits))
         if remainder is not None:
             body.append(self._call_remainder(remainder))
 
@@ -268,14 +270,19 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
 
     """
     A HaloExchangeBuilder making use of synchronous MPI routines only.
+
+    Generates:
+
+        haloupdate()
+        compute()
     """
 
     def _make_msg(self, f, hse, key):
         return
 
     def _make_all(self, f, hse, msg):
-        df = f.__class__.__base__(name='a', grid=f.grid, shape=f.shape_global,
-                                  dimensions=f.dimensions)
+        df = AliasFunction(name=self.sregistry.make_name(prefix='a'),
+                           grid=f.grid, shape=f.shape_global, dimensions=f.dimensions)
 
         if f.dimensions not in self._cache_dims:
             key = "".join(str(d) for d in f.dimensions)
@@ -304,7 +311,8 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
             if d not in hse.loc_indices:
                 buf_dims.append(Dimension(name='buf_%s' % d.root))
                 buf_indices.append(d.root)
-        buf = Array(name='buf', dimensions=buf_dims, dtype=f.dtype, padding=0)
+        buf = Array(name=self.sregistry.make_name(prefix='buf'),
+                    dimensions=buf_dims, dtype=f.dtype, padding=0)
 
         f_offsets = []
         f_indices = []
@@ -315,10 +323,10 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
 
         if swap is False:
             eq = DummyEq(buf[buf_indices], f[f_indices])
-            name = 'gather_%s' % key
+            name = 'gather%s' % key
         else:
             eq = DummyEq(f[f_indices], buf[buf_indices])
-            name = 'scatter_%s' % key
+            name = 'scatter%s' % key
 
         iet = Expression(eq)
         for i, d in reversed(list(zip(buf_indices, buf_dims))):
@@ -333,8 +341,10 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
 
         buf_dims = [Dimension(name='buf_%s' % d.root) for d in f.dimensions
                     if d not in hse.loc_indices]
-        bufg = Array(name='bufg', dimensions=buf_dims, dtype=f.dtype, padding=0)
-        bufs = Array(name='bufs', dimensions=buf_dims, dtype=f.dtype, padding=0)
+        bufg = Array(name=self.sregistry.make_name(prefix='bufg'),
+                     dimensions=buf_dims, dtype=f.dtype, padding=0)
+        bufs = Array(name=self.sregistry.make_name(prefix='bufs'),
+                     dimensions=buf_dims, dtype=f.dtype, padding=0)
 
         ofsg = [Symbol(name='og%s' % d.root) for d in f.dimensions]
         ofss = [Symbol(name='os%s' % d.root) for d in f.dimensions]
@@ -342,8 +352,8 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
         fromrank = Symbol(name='fromrank')
         torank = Symbol(name='torank')
 
-        gather = Call('gather_%s' % key, [bufg] + list(bufg.shape) + [f] + ofsg)
-        scatter = Call('scatter_%s' % key, [bufs] + list(bufs.shape) + [f] + ofss)
+        gather = Call('gather%s' % key, [bufg] + list(bufg.shape) + [f] + ofsg)
+        scatter = Call('scatter%s' % key, [bufs] + list(bufs.shape) + [f] + ofss)
 
         # The `gather` is unnecessary if sending to MPI.PROC_NULL
         gather = Conditional(CondNe(torank, Macro('MPI_PROC_NULL')), gather)
@@ -354,13 +364,13 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
         count = reduce(mul, bufs.shape, 1)
         rrecv = MPIRequestObject(name='rrecv')
         rsend = MPIRequestObject(name='rsend')
-        recv = Call('MPI_Irecv', [bufs, count, Macro(dtype_to_mpitype(f.dtype)),
-                                  fromrank, Integer(13), comm, rrecv])
-        send = Call('MPI_Isend', [bufg, count, Macro(dtype_to_mpitype(f.dtype)),
-                                  torank, Integer(13), comm, rsend])
+        recv = IrecvCall([bufs, count, Macro(dtype_to_mpitype(f.dtype)),
+                         fromrank, Integer(13), comm, Byref(rrecv)])
+        send = IsendCall([bufg, count, Macro(dtype_to_mpitype(f.dtype)),
+                         torank, Integer(13), comm, Byref(rsend)])
 
-        waitrecv = Call('MPI_Wait', [rrecv, Macro('MPI_STATUS_IGNORE')])
-        waitsend = Call('MPI_Wait', [rsend, Macro('MPI_STATUS_IGNORE')])
+        waitrecv = Call('MPI_Wait', [Byref(rrecv), Macro('MPI_STATUS_IGNORE')])
+        waitsend = Call('MPI_Wait', [Byref(rsend), Macro('MPI_STATUS_IGNORE')])
 
         iet = List(body=[recv, gather, send, waitsend, waitrecv, scatter])
         parameters = ([f] + list(bufs.shape) + ofsg + ofss + [fromrank, torank, comm])
@@ -427,7 +437,7 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
         comm = f.grid.distributor._obj_comm
         nb = f.grid.distributor._obj_neighborhood
         args = [f, comm, nb] + list(hse.loc_indices.values())
-        return Call(name, flatten(args))
+        return HaloUpdateCall(name, flatten(args))
 
     def _make_compute(self, *args):
         return
@@ -459,6 +469,11 @@ class DiagHaloExchangeBuilder(BasicHaloExchangeBuilder):
     """
     Similar to a BasicHaloExchangeBuilder, but communications to diagonal
     neighbours are performed explicitly.
+
+    Generates:
+
+        haloupdate()
+        compute()
     """
 
     def _make_haloupdate(self, f, hse, key, **kwargs):
@@ -503,6 +518,13 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
     """
     A DiagHaloExchangeBuilder making use of asynchronous MPI routines to implement
     computation-communication overlap.
+
+    Generates:
+
+        haloupdate()
+        compute_core()
+        halowait()
+        remainder()
     """
 
     def _make_msg(self, f, hse, key):
@@ -511,8 +533,8 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
         return MPIMsg('msg%d' % key, f, halos)
 
     def _make_all(self, f, hse, msg):
-        df = f.__class__.__base__(name='a', grid=f.grid, shape=f.shape_global,
-                                  dimensions=f.dimensions)
+        df = AliasFunction(name=self.sregistry.make_name(prefix='a'),
+                           grid=f.grid, shape=f.shape_global, dimensions=f.dimensions)
 
         if f.dimensions not in self._cache_dims:
             key = "".join(str(d) for d in f.dimensions)
@@ -548,17 +570,17 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
         sizes = [FieldFromPointer('%s[%d]' % (msg._C_field_sizes, i), msg)
                  for i in range(len(f._dist_dimensions))]
 
-        gather = Call('gather_%s' % key, [bufg] + sizes + [f] + ofsg)
+        gather = Call('gather%s' % key, [bufg] + sizes + [f] + ofsg)
         # The `gather` is unnecessary if sending to MPI.PROC_NULL
         gather = Conditional(CondNe(torank, Macro('MPI_PROC_NULL')), gather)
 
         count = reduce(mul, sizes, 1)
         rrecv = Byref(FieldFromPointer(msg._C_field_rrecv, msg))
         rsend = Byref(FieldFromPointer(msg._C_field_rsend, msg))
-        recv = Call('MPI_Irecv', [bufs, count, Macro(dtype_to_mpitype(f.dtype)),
-                                  fromrank, Integer(13), comm, rrecv])
-        send = Call('MPI_Isend', [bufg, count, Macro(dtype_to_mpitype(f.dtype)),
-                                  torank, Integer(13), comm, rsend])
+        recv = IrecvCall([bufs, count, Macro(dtype_to_mpitype(f.dtype)),
+                         fromrank, Integer(13), comm, rrecv])
+        send = IsendCall([bufg, count, Macro(dtype_to_mpitype(f.dtype)),
+                         torank, Integer(13), comm, rsend])
 
         iet = List(body=[recv, gather, send])
         parameters = ([f] + ofsg + [fromrank, torank, comm, msg])
@@ -605,7 +627,7 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
 
         sizes = [FieldFromPointer('%s[%d]' % (msg._C_field_sizes, i), msg)
                  for i in range(len(f._dist_dimensions))]
-        scatter = Call('scatter_%s' % key, [bufs] + sizes + [f] + ofss)
+        scatter = Call('scatter%s' % key, [bufs] + sizes + [f] + ofss)
 
         # The `scatter` must be guarded as we must not alter the halo values along
         # the domain boundary, where the sender is actually MPI.PROC_NULL
@@ -647,7 +669,7 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
 
     def _call_halowait(self, name, f, hse, msg):
         nb = f.grid.distributor._obj_neighborhood
-        return Call(name, [f] + list(hse.loc_indices.values()) + [nb, msg])
+        return HaloWaitCall(name, [f] + list(hse.loc_indices.values()) + [nb, msg])
 
     def _make_remainder(self, hs, key, callcompute, *args):
         assert callcompute.is_Call
@@ -655,7 +677,9 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
         return make_efunc('remainder%d' % key, body)
 
     def _call_remainder(self, remainder):
-        return remainder.make_call()
+        efunc = remainder.make_call()
+        call = RemainderCall(efunc.name, efunc.arguments)
+        return call
 
 
 class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
@@ -664,6 +688,13 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
     A OverlapHaloExchangeBuilder with reduced Call overhead and increased code
     readability, achieved by supplying more values via Python-land-produced
     structs, which replace explicit Call arguments.
+
+    Generates:
+
+        haloupdate()
+        compute_core()
+        halowait()
+        remainder()
     """
 
     def _make_region(self, hs, key):
@@ -675,8 +706,8 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         return MPIMsgEnriched('msg%d' % key, f, halos)
 
     def _make_all(self, f, hse, msg):
-        df = f.__class__.__base__(name='a', grid=f.grid, shape=f.shape_global,
-                                  dimensions=f.dimensions)
+        df = AliasFunction(name=self.sregistry.make_name(prefix='a'),
+                           grid=f.grid, shape=f.shape_global, dimensions=f.dimensions)
 
         if f.dimensions not in self._cache_dims:
             # Note: unlike the less smarter builders (superclasses), here we can
@@ -698,6 +729,7 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         return haloupdate, halowait
 
     def _make_haloupdate(self, f, hse, key, msg=None):
+        cast = cast_mapper[(f.dtype, '*')]
         comm = f.grid.distributor._obj_comm
 
         fixed = {d: Symbol(name="o%s" % d.root) for d in hse.loc_indices}
@@ -719,17 +751,18 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         ofsg = [fixed.get(d) or ofsg.pop(0) for d in f.dimensions]
 
         # The `gather` is unnecessary if sending to MPI.PROC_NULL
-        gather = Call('gather_%s' % key, [bufg] + sizes + [f] + ofsg)
+
+        gather = Call('gather%s' % key, [cast(bufg)] + sizes + [f] + ofsg)
         gather = Conditional(CondNe(torank, Macro('MPI_PROC_NULL')), gather)
 
         # Make Irecv/Isend
         count = reduce(mul, sizes, 1)
         rrecv = Byref(FieldFromComposite(msg._C_field_rrecv, msgi))
         rsend = Byref(FieldFromComposite(msg._C_field_rsend, msgi))
-        recv = Call('MPI_Irecv', [bufs, count, Macro(dtype_to_mpitype(f.dtype)),
-                                  fromrank, Integer(13), comm, rrecv])
-        send = Call('MPI_Isend', [bufg, count, Macro(dtype_to_mpitype(f.dtype)),
-                                  torank, Integer(13), comm, rsend])
+        recv = IrecvCall([bufs, count, Macro(dtype_to_mpitype(f.dtype)),
+                          fromrank, Integer(13), comm, rrecv])
+        send = IsendCall([bufg, count, Macro(dtype_to_mpitype(f.dtype)),
+                         torank, Integer(13), comm, rsend])
 
         # The -1 below is because an Iteration, by default, generates <=
         ncomms = Symbol(name='ncomms')
@@ -739,7 +772,8 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
 
     def _call_haloupdate(self, name, f, hse, msg):
         comm = f.grid.distributor._obj_comm
-        return Call(name, [f, comm, msg, msg.npeers] + list(hse.loc_indices.values()))
+        args = [f, comm, msg, msg.npeers] + list(hse.loc_indices.values())
+        return HaloUpdateCall(name, args)
 
     def _make_sendrecv(self, *args):
         return
@@ -748,6 +782,8 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         return
 
     def _make_halowait(self, f, hse, key, msg=None):
+        cast = cast_mapper[(f.dtype, '*')]
+
         fixed = {d: Symbol(name="o%s" % d.root) for d in hse.loc_indices}
 
         dim = Dimension(name='i')
@@ -766,7 +802,7 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
 
         # The `scatter` must be guarded as we must not alter the halo values along
         # the domain boundary, where the sender is actually MPI.PROC_NULL
-        scatter = Call('scatter_%s' % key, [bufs] + sizes + [f] + ofss)
+        scatter = Call('scatter%s' % key, [cast(bufs)] + sizes + [f] + ofss)
         scatter = Conditional(CondNe(fromrank, Macro('MPI_PROC_NULL')), scatter)
 
         rrecv = Byref(FieldFromComposite(msg._C_field_rrecv, msgi))
@@ -781,7 +817,8 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         return Callable('halowait%d' % key, iet, 'void', parameters, ('static',))
 
     def _call_halowait(self, name, f, hse, msg):
-        return Call(name, [f] + list(hse.loc_indices.values()) + [msg, msg.npeers])
+        args = [f] + list(hse.loc_indices.values()) + [msg, msg.npeers]
+        return HaloWaitCall(name, args)
 
     def _make_wait(self, *args):
         return
@@ -810,11 +847,45 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         return make_efunc('remainder%d' % key, iet)
 
 
+class Diag2HaloExchangeBuilder(Overlap2HaloExchangeBuilder):
+
+    """
+    A DiagHaloExchangeBuilder which uses preallocated buffers for comms
+    as in Overlap2HaloExchange builder.
+
+    Generates:
+
+        haloupdate()
+        halowait()
+        compute()
+    """
+
+    def _make_compute(self, hs, key, *args):
+        return
+
+    def _call_compute(self, hs, compute, *args):
+        return
+
+    def _make_remainder(self, hs, key, callcompute, region):
+        # Just a dummy value, other than a Callable, so that we enter `_call_remainder`
+        return hs.body
+
+    def _call_remainder(self, remainder):
+        return remainder
+
+
 class FullHaloExchangeBuilder(Overlap2HaloExchangeBuilder):
 
     """
-    A Overlap2HaloExchangeBuilder which generates explicit Calls to MPI_Test
+    An Overlap2HaloExchangeBuilder which generates explicit Calls to MPI_Test
     poking the MPI runtime to advance communication while computing.
+
+    Generates:
+
+        haloupdate()
+        compute_core()
+        halowait()
+        remainder()
     """
 
     def _make_compute(self, hs, key, msgs, callpoke):
@@ -858,6 +929,17 @@ class FullHaloExchangeBuilder(Overlap2HaloExchangeBuilder):
         return Prodder(poke.name, poke.parameters, single_thread=True, periodic=True)
 
 
+mpi_registry = {
+    True: BasicHaloExchangeBuilder,
+    'basic': BasicHaloExchangeBuilder,
+    'diag': DiagHaloExchangeBuilder,
+    'diag2': Diag2HaloExchangeBuilder,
+    'overlap': OverlapHaloExchangeBuilder,
+    'overlap2': Overlap2HaloExchangeBuilder,
+    'full': FullHaloExchangeBuilder
+}
+
+
 # Callable sub-hierarchy
 
 
@@ -874,7 +956,7 @@ class CopyBuffer(MPICallable):
 class SendRecv(MPICallable):
 
     def __init__(self, key, body, parameters, bufg, bufs):
-        super(SendRecv, self).__init__('sendrecv_%s' % key, body, parameters)
+        super(SendRecv, self).__init__('sendrecv%s' % key, body, parameters)
         self.bufg = bufg
         self.bufs = bufs
 
@@ -882,7 +964,49 @@ class SendRecv(MPICallable):
 class HaloUpdate(MPICallable):
 
     def __init__(self, key, body, parameters):
-        super(HaloUpdate, self).__init__('haloupdate_%s' % key, body, parameters)
+        super(HaloUpdate, self).__init__('haloupdate%s' % key, body, parameters)
+
+
+# Call sub-hierarchy
+
+class IsendCall(Call):
+
+    def __init__(self, arguments):
+        super().__init__('MPI_Isend', arguments)
+
+
+class IrecvCall(Call):
+
+    def __init__(self, arguments):
+        super().__init__('MPI_Irecv', arguments)
+
+
+class MPICall(Call):
+    pass
+
+
+class HaloUpdateCall(MPICall):
+    pass
+
+
+class HaloWaitCall(MPICall):
+    pass
+
+
+class RemainderCall(MPICall):
+    pass
+
+
+class MPIList(List):
+    pass
+
+
+class HaloUpdateList(MPIList):
+    pass
+
+
+class HaloWaitList(MPIList):
+    pass
 
 
 # Types sub-hierarchy
@@ -892,23 +1016,11 @@ class MPIStatusObject(LocalObject):
 
     dtype = type('MPI_Status', (c_void_p,), {})
 
-    def __init__(self, name):
-        self.name = name
-
-    # Pickling support
-    _pickle_args = ['name']
-
 
 class MPIRequestObject(LocalObject):
 
     dtype = type('MPI_Request', (c_void_p,), {})
 
-    def __init__(self, name):
-        self.name = name
-
-    # Pickling support
-    _pickle_args = ['name']
-
 
 class MPIMsg(CompositeObject):
 
@@ -931,15 +1043,15 @@ class MPIMsg(CompositeObject):
         (_C_field_rsend, c_mpirequest_p),
     ]
 
-    def __init__(self, name, function, halos):
-        self._function = function
+    def __init__(self, name, target, halos):
+        self._target = target
         self._halos = halos
 
-        super(MPIMsg, self).__init__(name, 'msg', self.fields)
+        super().__init__(name, 'msg', self.fields)
 
         # Required for buffer allocation/deallocation before/after jumping/returning
         # to/from C-land
-        self._allocator = default_allocator()
+        self._allocator = None
         self._memfree_args = []
 
     def __del__(self):
@@ -957,8 +1069,8 @@ class MPIMsg(CompositeObject):
         return (dtype._type_*self.npeers)()
 
     @property
-    def function(self):
-        return self._function
+    def target(self):
+        return self._target
 
     @property
     def halos(self):
@@ -968,24 +1080,31 @@ class MPIMsg(CompositeObject):
     def npeers(self):
         return len(self._halos)
 
-    def _arg_defaults(self, alias=None):
-        function = alias or self.function
+    def _arg_defaults(self, allocator, alias=None):
+        # Lazy initialization if `allocator` is necessary as the `allocator`
+        # type isn't really known until an Operator is constructed
+        self._allocator = allocator
+
+        target = alias or self.target
         for i, halo in enumerate(self.halos):
             entry = self.value[i]
+
             # Buffer size for this peer
             shape = []
             for dim, side in zip(*halo):
                 try:
-                    shape.append(getattr(function._size_owned[dim], side.name))
+                    shape.append(getattr(target._size_owned[dim], side.name))
                 except AttributeError:
                     assert side is CENTER
-                    shape.append(function._size_domain[dim])
+                    shape.append(target._size_domain[dim])
             entry.sizes = (c_int*len(shape))(*shape)
+
             # Allocate the send/recv buffers
             size = reduce(mul, shape)
-            ctype = dtype_to_ctype(function.dtype)
-            entry.bufg, bufg_memfree_args = self._allocator._alloc_C_libcall(size, ctype)
-            entry.bufs, bufs_memfree_args = self._allocator._alloc_C_libcall(size, ctype)
+            ctype = dtype_to_ctype(target.dtype)
+            entry.bufg, bufg_memfree_args = allocator._alloc_C_libcall(size, ctype)
+            entry.bufs, bufs_memfree_args = allocator._alloc_C_libcall(size, ctype)
+
             # The `memfree_args` will be used to deallocate the buffer upon returning
             # from C-land
             self._memfree_args.extend([bufg_memfree_args, bufs_memfree_args])
@@ -993,13 +1112,16 @@ class MPIMsg(CompositeObject):
         return {self.name: self.value}
 
     def _arg_values(self, args=None, **kwargs):
-        return self._arg_defaults(alias=kwargs.get(self.function.name, self.function))
+        return self._arg_defaults(
+            args.allocator,
+            alias=kwargs.get(self.target.name, self.target)
+        )
 
     def _arg_apply(self, *args, **kwargs):
         self._C_memfree()
 
     # Pickling support
-    _pickle_args = ['name', 'function', 'halos']
+    _pickle_args = ['name', 'target', 'halos']
 
 
 class MPIMsgEnriched(MPIMsg):
@@ -1016,8 +1138,8 @@ class MPIMsgEnriched(MPIMsg):
         (_C_field_to, c_int)
     ]
 
-    def _arg_defaults(self, alias=None):
-        super(MPIMsgEnriched, self)._arg_defaults(alias)
+    def _arg_defaults(self, allocator, alias=None):
+        super()._arg_defaults(allocator, alias)
 
         function = alias or self.function
         neighborhood = function.grid.distributor.neighborhood
@@ -1099,7 +1221,7 @@ class MPIRegion(CompositeObject):
     def nregions(self):
         return len(self.owned)
 
-    def _arg_values(self, args, **kwargs):
+    def _arg_values(self, args=None, **kwargs):
         values = self._arg_defaults()
         for i, (_, mapper) in enumerate(self.owned):
             entry = values[self.name][i]
