@@ -4,19 +4,16 @@ import sympy
 import numpy as np
 from cached_property import cached_property
 
-from devito.finite_differences import Evaluable
 from devito.logger import warning
 from devito.symbolics import retrieve_function_carriers, indexify, INT
 from devito.tools import powerset, flatten, prod
-from devito.types import Eq, Inc
-from devito.types.basic import Scalar
-from devito.types.dense import SubFunction
-from devito.types.dimension import ConditionalDimension, Dimension, DefaultDimension
+from devito.types import (ConditionalDimension, Dimension, DefaultDimension, Eq, Inc,
+                          Evaluable, Symbol, SubFunction)
 
 __all__ = ['LinearInterpolator', 'PrecomputedInterpolator']
 
 
-class UnevaluatedSparseOperation(Evaluable):
+class UnevaluatedSparseOperation(sympy.Expr, Evaluable):
 
     """
     Represents an Injection or an Interpolation operation performed on a
@@ -32,9 +29,13 @@ class UnevaluatedSparseOperation(Evaluable):
 
     subdomain = None
 
-    def __init__(self, interpolator, callback):
-        self.interpolator = interpolator
-        self.callback = callback
+    def __new__(cls, interpolator, callback):
+        obj = super().__new__(cls)
+
+        obj.interpolator = interpolator
+        obj.callback = callback
+
+        return obj
 
     @property
     def evaluate(self):
@@ -56,14 +57,16 @@ class Interpolation(UnevaluatedSparseOperation):
     Evaluates to a list of Eq objects.
     """
 
-    def __init__(self, expr, offset, increment, self_subs, interpolator, callback):
-        # TODO: unused now, but will be necessary to compute the adjoint
-        self.expr = expr
-        self.offset = offset
-        self.increment = increment
-        self.self_subs = self_subs
+    def __new__(cls, expr, offset, increment, self_subs, interpolator, callback):
+        obj = super().__new__(cls, interpolator, callback)
 
-        super(Interpolation, self).__init__(interpolator, callback)
+        # TODO: unused now, but will be necessary to compute the adjoint
+        obj.expr = expr
+        obj.offset = offset
+        obj.increment = increment
+        obj.self_subs = self_subs
+
+        return obj
 
     def __repr__(self):
         return "Interpolation(%s into %s)" % (repr(self.expr),
@@ -77,13 +80,15 @@ class Injection(UnevaluatedSparseOperation):
     Evaluates to a list of Eq objects.
     """
 
-    def __init__(self, field, expr, offset, interpolator, callback):
-        # TODO: unused now, but will be necessary to compute the adjoint
-        self.field = field
-        self.expr = expr
-        self.offset = offset
+    def __new__(cls, field, expr, offset, interpolator, callback):
+        obj = super().__new__(cls, interpolator, callback)
 
-        super(Injection, self).__init__(interpolator, callback)
+        # TODO: unused now, but will be necessary to compute the adjoint
+        obj.field = field
+        obj.expr = expr
+        obj.offset = offset
+
+        return obj
 
     def __repr__(self):
         return "Injection(%s into %s)" % (repr(self.expr), repr(self.field))
@@ -237,7 +242,7 @@ class LinearInterpolator(GenericInterpolator):
                     for b, v_sub in zip(self._interpolation_coeffs, idx_subs)]
 
             # Accumulate point-wise contributions into a temporary
-            rhs = Scalar(name='sum', dtype=self.sfunction.dtype)
+            rhs = Symbol(name='sum', dtype=self.sfunction.dtype)
             summands = [Eq(rhs, 0., implicit_dims=self.sfunction.dimensions)]
             summands.extend([Inc(rhs, i, implicit_dims=self.sfunction.dimensions)
                             for i in args])
@@ -302,7 +307,7 @@ class PrecomputedInterpolator(GenericInterpolator):
         gridpoints = SubFunction(name="%s_gridpoints" % self.obj.name, dtype=np.int32,
                                  dimensions=(self.obj.indices[-1], Dimension(name='d')),
                                  shape=(self._npoint, self.obj.grid.dim), space_order=0,
-                                 parent=self)
+                                 parent=self.obj)
 
         assert(gridpoints_data is not None)
         gridpoints.data[:] = gridpoints_data[:]
@@ -315,7 +320,7 @@ class PrecomputedInterpolator(GenericInterpolator):
                                            shape=(self.obj.npoint, self.obj.grid.dim,
                                                   self.r),
                                            dtype=self.obj.dtype, space_order=0,
-                                           parent=self)
+                                           parent=self.obj)
         assert(coefficients_data is not None)
         interpolation_coeffs.data[:] = coefficients_data[:]
         self.obj._interpolation_coeffs = interpolation_coeffs
@@ -371,12 +376,12 @@ class PrecomputedInterpolator(GenericInterpolator):
             _expr = indexify(expr)
             _field = indexify(field)
 
-            p, _ = self.gridpoints.indices
+            p, _ = self.obj.gridpoints.indices
             dim_subs = []
             coeffs = []
-            for i, d in enumerate(self.grid.dimensions):
+            for i, d in enumerate(self.obj.grid.dimensions):
                 rd = DefaultDimension(name="r%s" % d.name, default_value=self.r)
-                dim_subs.append((d, INT(rd + self.gridpoints[p, i])))
+                dim_subs.append((d, INT(rd + self.obj.gridpoints[p, i])))
                 coeffs.append(self.obj.interpolation_coeffs[p, i, rd])
             rhs = prod(coeffs) * _expr
             _field = _field.subs(dim_subs)
