@@ -1,31 +1,30 @@
 from collections import OrderedDict
-from functools import reduce
-from operator import attrgetter, mul
+from operator import attrgetter
 from math import ceil
 
 from cached_property import cached_property
 import ctypes
+import numpy as np
 
-from devito.archinfo import platform_registry
-from devito.compiler import compiler_registry
+from devito.arch import compiler_registry, platform_registry
+from devito.data import default_allocator
 from devito.exceptions import InvalidOperator
-from devito.finite_differences import Evaluable
 from devito.logger import info, perf, warning, is_log_enabled_for
-from devito.ir.equations import LoweredEq
+from devito.ir.equations import LoweredEq, lower_exprs, generate_implicit_exprs
 from devito.ir.clusters import ClusterGroup, clusterize
-from devito.ir.iet import Callable, MetaCall, derive_parameters, iet_build, iet_lower_dims
+from devito.ir.iet import (Callable, EntryFunction, FindSymbols, MetaCall,
+                           derive_parameters, iet_build)
 from devito.ir.stree import stree_build
-from devito.ir.equations.algorithms import lower_exprs
 from devito.operator.profiling import create_profile
 from devito.operator.registry import operator_selector
 from devito.operator.symbols import SymbolRegistry
 from devito.mpi import MPI
 from devito.parameters import configuration
-from devito.passes import Graph, NThreads, NThreadsNested, NThreadsNonaffine
-from devito.symbolics import estimate_cost, retrieve_functions
-from devito.tools import (DAG, Signer, ReducerMap, as_tuple, flatten, filter_ordered,
-                          filter_sorted, split, timed_pass, timed_region)
-from devito.types import CustomDimension, Dimension, Eq
+from devito.passes import Graph, instrument
+from devito.symbolics import estimate_cost
+from devito.tools import (DAG, Signer, ReducerMap, as_tuple, flatten, filter_sorted,
+                          split, timed_pass, timed_region)
+from devito.types import Grid, Evaluable
 
 __all__ = ['Operator']
 
@@ -133,7 +132,7 @@ class Operator(Callable):
     refer to the relevant documentation.
     """
 
-    _default_headers = ['#define _POSIX_C_SOURCE 200809L']
+    _default_headers = [('_POSIX_C_SOURCE', '200809L')]
     _default_includes = ['stdlib.h', 'math.h', 'sys/time.h']
     _default_globals = []
 
@@ -153,7 +152,7 @@ class Operator(Callable):
         kwargs = cls._normalize_kwargs(**kwargs)
 
         # Create a symbol registry
-        kwargs['sregistry'] = cls._symbol_registry()
+        kwargs['sregistry'] = SymbolRegistry()
 
         # Lower to a JIT-compilable object
         with timed_region('op-compile') as r:
@@ -169,16 +168,6 @@ class Operator(Callable):
     def _normalize_kwargs(cls, **kwargs):
         return kwargs
 
-    @classmethod
-    def _symbol_registry(cls):
-        # Special symbols an Operator might use
-        nthreads = NThreads(aliases='nthreads0')
-        nthreads_nested = NThreadsNested(aliases='nthreads1')
-        nthreads_nonaffine = NThreadsNonaffine(aliases='nthreads2')
-        threadid = CustomDimension(name='tid', symbolic_size=nthreads)
-
-        return SymbolRegistry(nthreads, nthreads_nested, nthreads_nonaffine, threadid)
-
     @classmethod
     def _build(cls, expressions, **kwargs):
         expressions = as_tuple(expressions)
@@ -219,6 +208,10 @@ class Operator(Callable):
         op._lib = None
         op._cfunction = None
 
+        # Potentially required for lazily allocated Functions
+        op._allocator = kwargs['allocator']
+        op._platform = kwargs['platform']
+
         # References to local or external routines
         op._func_table = OrderedDict()
         op._func_table.update(OrderedDict([(i, MetaCall(None, False))
@@ -232,8 +225,7 @@ class Operator(Callable):
         # Produced by the various compilation passes
         op._input = filter_sorted(flatten(e.reads + e.writes for e in expressions))
         op._output = filter_sorted(flatten(e.writes for e in expressions))
-        op._dimensions = flatten(c.dimensions for c in clusters) + byproduct.dimensions
-        op._dimensions = sorted(set(op._dimensions), key=attrgetter('name'))
+        op._dimensions = set().union(*[e.dimensions for e in expressions])
         op._dtype, op._dspace = clusters.meta
         op._profiler = profiler
 
@@ -245,48 +237,27 @@ class Operator(Callable):
 
     # Compilation -- Expression level
 
-    @classmethod
-    def _add_implicit(cls, expressions):
-        """
-        Create and add any associated implicit expressions.
-
-        Implicit expressions are those not explicitly defined by the user
-        but instead are requisites of some specified functionality.
-        """
-        processed = []
-        seen = set()
-        for e in expressions:
-            if e.subdomain:
-                try:
-                    dims = [d.root for d in e.free_symbols if isinstance(d, Dimension)]
-                    sub_dims = [d.root for d in e.subdomain.dimensions]
-                    sub_dims.append(e.subdomain.implicit_dimension)
-                    dims = [d for d in dims if d not in frozenset(sub_dims)]
-                    dims.append(e.subdomain.implicit_dimension)
-                    if e.subdomain not in seen:
-                        grid = list(retrieve_functions(e, mode='unique'))[0].grid
-                        processed.extend([i.func(*i.args, implicit_dims=dims) for i in
-                                          e.subdomain._create_implicit_exprs(grid)])
-                        seen.add(e.subdomain)
-                    dims.extend(e.subdomain.dimensions)
-                    new_e = Eq(e.lhs, e.rhs, subdomain=e.subdomain, implicit_dims=dims)
-                    processed.append(new_e)
-                except AttributeError:
-                    processed.append(e)
-            else:
-                processed.append(e)
-        return processed
-
     @classmethod
     def _initialize_state(cls, **kwargs):
         return {'optimizations': kwargs.get('mode', configuration['opt'])}
 
     @classmethod
-    def _specialize_exprs(cls, expressions):
+    def _specialize_dsl(cls, expressions, **kwargs):
         """
-        Backend hook for specialization at the Expression level.
+        Backend hook for specialization at the DSL level. The input is made of
+        expressions and other higher order objects such as Injection or
+        Interpolation; the expressions are still unevaluated at this stage,
+        meaning that they are still in tensorial form and derivatives aren't
+        expanded yet.
         """
-        return [LoweredEq(i) for i in expressions]
+        return expressions
+
+    @classmethod
+    def _specialize_exprs(cls, expressions, **kwargs):
+        """
+        Backend hook for specialization at the expression level.
+        """
+        return expressions
 
     @classmethod
     @timed_pass(name='lowering.Expressions')
@@ -295,24 +266,30 @@ class Operator(Callable):
         Expression lowering:
 
             * Form and gather any required implicit expressions;
+            * Apply rewrite rules;
             * Evaluate derivatives;
             * Flatten vectorial equations;
             * Indexify Functions;
             * Apply substitution rules;
-            * Specialize (e.g., index shifting)
+            * Shift indices for domain alignment.
         """
-        # Add in implicit expressions, e.g., induced by SubDomains
-        expressions = cls._add_implicit(expressions)
+        # Add in implicit expressions
+        expressions = generate_implicit_exprs(expressions)
 
-        # Unfold lazyiness
+        # Specialization is performed on unevaluated expressions
+        expressions = cls._specialize_dsl(expressions, **kwargs)
+
+        # Lower functional DSL
         expressions = flatten([i.evaluate for i in expressions])
-
-        # Scalarize tensor expressions
         expressions = [j for i in expressions for j in i._flatten]
 
+        # A second round of specialization is performed on nevaluated expressions
+        expressions = cls._specialize_exprs(expressions, **kwargs)
+
+        # "True" lowering (indexification, shifting, ...)
         expressions = lower_exprs(expressions, **kwargs)
 
-        processed = cls._specialize_exprs(expressions)
+        processed = [LoweredEq(i) for i in expressions]
 
         return processed
 
@@ -337,7 +314,7 @@ class Operator(Callable):
               as parallelism.
         """
         # Build a sequence of Clusters from a sequence of Eqs
-        clusters = clusterize(expressions)
+        clusters = clusterize(expressions, **kwargs)
 
         # Operation count before specialization
         init_ops = sum(estimate_cost(c.exprs) for c in clusters if c.is_dense)
@@ -397,24 +374,28 @@ class Operator(Callable):
             * Finalize (e.g., symbol definitions, array casts)
         """
         name = kwargs.get("name", "Kernel")
+        sregistry = kwargs['sregistry']
 
         # Build an IET from a ScheduleTree
         iet = iet_build(stree)
 
-        # Instrument the IET for C-level profiling
-        iet = profiler.instrument(iet)
+        # Analyze the IET Sections for C-level profiling
+        profiler.analyze(iet)
 
-        # Lower all DerivedDimensions
-        iet = iet_lower_dims(iet)
-
-        # Wrap the IET with a Callable
+        # Wrap the IET with an EntryFunction (a special Callable representing
+        # the entry point of the generated library)
         parameters = derive_parameters(iet, True)
-        iet = Callable(name, iet, 'int', parameters, ())
+        iet = EntryFunction(name, iet, 'int', parameters, ())
 
         # Lower IET to a target-specific IET
         graph = Graph(iet)
         graph = cls._specialize_iet(graph, **kwargs)
 
+        # Instrument the IET for C-level profiling
+        # Note: this is postponed until after _specialize_iet because during
+        # specialization further Sections may be introduced
+        instrument(graph, profiler=profiler, sregistry=sregistry)
+
         return graph.root, graph
 
     # Read-only properties exposed to the outside world
@@ -425,12 +406,23 @@ class Operator(Callable):
 
     @cached_property
     def dimensions(self):
-        return tuple(self._dimensions)
+        ret = set().union(*[d._defines for d in self._dimensions])
+
+        # During compilation other Dimensions may have been produced
+        dimensions = FindSymbols('dimensions').visit(self)
+        ret.update(d for d in dimensions if d.is_PerfKnob)
+
+        ret = tuple(sorted(ret, key=attrgetter('name')))
+
+        return ret
 
     @cached_property
     def input(self):
-        ret = [i for i in self._input + list(self.parameters) if i.is_Input]
-        return tuple(filter_ordered(ret))
+        return tuple(i for i in self.parameters if i.is_Input)
+
+    @cached_property
+    def temporaries(self):
+        return tuple(i for i in self.parameters if i.is_TempFunction)
 
     @cached_property
     def objects(self):
@@ -438,18 +430,25 @@ class Operator(Callable):
 
     # Arguments processing
 
-    def _prepare_arguments(self, **kwargs):
+    def _prepare_arguments(self, autotune=None, **kwargs):
         """
         Process runtime arguments passed to ``.apply()` and derive
         default values for any remaining arguments.
         """
+        # Sanity check -- all user-provided keywords must be known to the Operator
+        if not configuration['ignore-unknowns']:
+            for k, v in kwargs.items():
+                if k not in self._known_arguments:
+                    raise ValueError("Unrecognized argument %s=%s" % (k, v))
+
         overrides, defaults = split(self.input, lambda p: p.name in kwargs)
+
         # Process data-carrier overrides
-        args = ReducerMap()
+        args = kwargs['args'] = ReducerMap()
         for p in overrides:
             args.update(p._arg_values(**kwargs))
             try:
-                args = ReducerMap(args.reduce_all())
+                args.reduce_inplace()
             except ValueError:
                 raise ValueError("Override `%s` is incompatible with overrides `%s`" %
                                  (p, [i for i in overrides if i.name in args]))
@@ -465,19 +464,33 @@ class Operator(Callable):
                                      "forgot to override `%s`?" %
                                      (p, k, v, k, args[k], p))
                 args[k] = v
-        args = args.reduce_all()
+        args = kwargs['args'] = args.reduce_all()
 
-        # All DiscreteFunctions should be defined on the same Grid
-        grids = {getattr(kwargs[p.name], 'grid', None) for p in overrides}
-        grids.update({getattr(p, 'grid', None) for p in defaults})
-        grids.discard(None)
-        if len(grids) > 1 and configuration['mpi']:
-            raise ValueError("Multiple Grids found")
+        # DiscreteFunctions may be created from CartesianDiscretizations, which in
+        # turn could be Grids or SubDomains. Both may provide arguments
+        discretizations = {getattr(kwargs[p.name], 'grid', None) for p in overrides}
+        discretizations.update({getattr(p, 'grid', None) for p in defaults})
+        discretizations.discard(None)
+        for i in discretizations:
+            args.update(i._arg_values(**kwargs))
+
+        # There can only be one Grid from which DiscreteFunctions were created
+        grids = {i for i in discretizations if isinstance(i, Grid)}
+        if len(grids) > 1:
+            # We loosely tolerate multiple Grids for backwards compatibility
+            # with spacial subsampling, which should be revisited however. And
+            # With MPI it would definitely break!
+            if configuration['mpi']:
+                raise ValueError("Multiple Grids found")
         try:
             grid = grids.pop()
         except KeyError:
             grid = None
 
+        # An ArgumentsMap carries additional metadata that may be used by
+        # the subsequent phases of the arguments processing
+        args = kwargs['args'] = ArgumentsMap(args, grid, self._allocator, self._platform)
+
         # Process Dimensions
         # A topological sorting is used so that derived Dimensions are processed after
         # their parents (note that a leaf Dimension can have an arbitrary long list of
@@ -485,11 +498,11 @@ class Operator(Callable):
         dag = DAG(self.dimensions,
                   [(i, i.parent) for i in self.dimensions if i.is_Derived])
         for d in reversed(dag.topological_sort()):
-            args.update(d._arg_values(args, self._dspace[d], grid, **kwargs))
+            args.update(d._arg_values(self._dspace[d], grid, **kwargs))
 
-        # Process Objects (which may need some `args`)
+        # Process Objects
         for o in self.objects:
-            args.update(o._arg_values(args, grid=grid, **kwargs))
+            args.update(o._arg_values(grid=grid, **kwargs))
 
         # Sanity check
         for p in self.parameters:
@@ -503,28 +516,13 @@ class Operator(Callable):
         # pointers to ctypes.Struct
         for p in self.parameters:
             try:
-                args.update(kwargs.get(p.name, p)._arg_as_ctype(args, alias=p))
+                args.update(kwargs.get(p.name, p)._arg_finalize(args, alias=p))
             except AttributeError:
-                # User-provided floats/ndarray obviously do not have `_arg_as_ctype`
-                args.update(p._arg_as_ctype(args, alias=p))
-
-        # Add in the profiler argument
-        args[self._profiler.name] = self._profiler.timer.reset()
-
-        # Add in any backend-specific argument
-        args.update(kwargs.pop('backend', {}))
+                # User-provided floats/ndarray obviously do not have `_arg_finalize`
+                args.update(p._arg_finalize(args, alias=p))
 
         # Execute autotuning and adjust arguments accordingly
-        args = self._autotune(args, kwargs.pop('autotune', configuration['autotuning']))
-
-        # Check all user-provided keywords are known to the Operator
-        if not configuration['ignore-unknowns']:
-            for k, v in kwargs.items():
-                if k not in self._known_arguments:
-                    raise ValueError("Unrecognized argument %s=%s" % (k, v))
-
-        # Attach `grid` to the arguments map
-        args = ArgumentsMap(grid, **args)
+        args.update(self._autotune(args, autotune or configuration['autotuning']))
 
         return args
 
@@ -539,8 +537,16 @@ class Operator(Callable):
     @cached_property
     def _known_arguments(self):
         """The arguments that can be passed to ``apply`` when running the Operator."""
-        ret = set.union(*[set(i._arg_names) for i in self.input + self.dimensions])
-        return tuple(sorted(ret))
+        ret = set()
+        for i in self.input:
+            ret.update(i._arg_names)
+            try:
+                ret.update(i.grid._arg_names)
+            except AttributeError:
+                pass
+        for d in self.dimensions:
+            ret.update(d._arg_names)
+        return frozenset(ret)
 
     def _autotune(self, args, setup):
         """Auto-tuning to improve runtime performance."""
@@ -729,7 +735,7 @@ class Operator(Callable):
         # Rounder to 2 decimal places
         fround = lambda i: ceil(i * 100) / 100
 
-        info("Operator `%s` run in %.2f s" % (self.name,
+        info("Operator `%s` ran in %.2f s" % (self.name,
                                               fround(self._profiler.py_timers['apply'])))
 
         summary = self._profiler.summary(args, self._dtype, reduce_over='apply')
@@ -739,24 +745,24 @@ class Operator(Callable):
             return summary
 
         if summary.globals:
-            indent = " "*2
+            # Note that with MPI enabled, the global performance indicators
+            # represent "cross-rank" performance data
+            metrics = []
 
-            perf("Global performance indicators")
-
-            # With MPI enabled, the 'vanilla' entry contains "cross-rank" performance data
             v = summary.globals.get('vanilla')
             if v is not None:
-                gflopss = "%.2f GFlops/s" % fround(v.gflopss)
-                gpointss = "%.2f GPts/s" % fround(v.gpointss) if v.gpointss else None
-                metrics = ", ".join(i for i in [gflopss, gpointss] if i is not None)
-                perf("%s* Operator `%s` with OI=%.2f computed in %.2f s [%s]" %
-                     (indent, self.name, fround(v.oi), fround(v.time), metrics))
+                metrics.append("OI=%.2f" % fround(v.oi))
+                metrics.append("%.2f GFlops/s" % fround(v.gflopss))
 
             v = summary.globals.get('fdlike')
             if v is not None:
-                perf("%s* Achieved %.2f FD-GPts/s" % (indent, v.gpointss))
+                metrics.append("%.2f GPts/s" % fround(v.gpointss))
 
-            perf("Local performance indicators")
+            if metrics:
+                perf("Global performance: [%s]" % ', '.join(metrics))
+
+            perf("Local performance:")
+            indent = " "*2
         else:
             indent = ""
 
@@ -764,23 +770,23 @@ class Operator(Callable):
         # thing that will be emitted
         for k, v in summary.items():
             rank = "[rank%d]" % k.rank if k.rank is not None else ""
+            oi = "OI=%.2f" % fround(v.oi)
             gflopss = "%.2f GFlops/s" % fround(v.gflopss)
             gpointss = "%.2f GPts/s" % fround(v.gpointss) if v.gpointss else None
-            metrics = ", ".join(i for i in [gflopss, gpointss] if i is not None)
+            metrics = ", ".join(i for i in [oi, gflopss, gpointss] if i is not None)
             itershapes = [",".join(str(i) for i in its) for its in v.itershapes]
             if len(itershapes) > 1:
-                name = "%s%s<%s>" % (k.name, rank,
-                                     ",".join("<%s>" % i for i in itershapes))
-                perf("%s* %s with OI=%.2f computed in %.2f s [%s]" %
-                     (indent, name, fround(v.oi), fround(v.time), metrics))
+                itershapes = ",".join("<%s>" % i for i in itershapes)
             elif len(itershapes) == 1:
-                name = "%s%s<%s>" % (k.name, rank, itershapes[0])
-                perf("%s* %s with OI=%.2f computed in %.2f s [%s]" %
-                     (indent, name, fround(v.oi), fround(v.time), metrics))
+                itershapes = itershapes[0]
             else:
-                name = k.name
-                perf("%s* %s%s computed in %.2f s"
-                     % (indent, name, rank, fround(v.time)))
+                itershapes = ""
+            name = "%s%s<%s>" % (k.name, rank, itershapes)
+
+            perf("%s* %s ran in %.2f s [%s]" % (indent, name, fround(v.time), metrics))
+            for n, time in summary.subsections.get(k.name, {}).items():
+                perf("%s+ %s ran in %.2f s [%.2f%%]" %
+                     (indent*2, n, time, fround(time/v.time*100)))
 
         # Emit performance mode and arguments
         perf_args = {}
@@ -800,36 +806,6 @@ class Operator(Callable):
 
         return summary
 
-    # Misc properties
-
-    @cached_property
-    def _mem_summary(self):
-        """
-        The amount of data, in bytes, used by the Operator. This is provided as
-        symbolic expressions, one symbolic expression for each memory scope (external,
-        stack, heap).
-        """
-        roots = [self] + [i.root for i in self._func_table.values()]
-        functions = [i for i in derive_parameters(roots) if i.is_Function]
-
-        summary = {}
-
-        external = [i.symbolic_shape for i in functions if i._mem_external]
-        external = sum(reduce(mul, i, 1) for i in external)*self._dtype().itemsize
-        summary['external'] = external
-
-        heap = [i.symbolic_shape for i in functions if i._mem_heap]
-        heap = sum(reduce(mul, i, 1) for i in heap)*self._dtype().itemsize
-        summary['heap'] = heap
-
-        stack = [i.symbolic_shape for i in functions if i._mem_stack]
-        stack = sum(reduce(mul, i, 1) for i in stack)*self._dtype().itemsize
-        summary['stack'] = stack
-
-        summary['total'] = external + heap + stack
-
-        return summary
-
     # Pickling support
 
     def __getstate__(self):
@@ -850,6 +826,7 @@ class Operator(Callable):
             state['_args'] = None
             with open(self._lib._name, 'rb') as f:
                 state['binary'] = f.read()
+                state['soname'] = self._soname
             return state
         else:
             return self.__dict__
@@ -858,28 +835,14 @@ class Operator(Callable):
         return (None,), {}
 
     def __setstate__(self, state):
-        soname = state.pop('_soname', None)
+        soname = state.pop('soname', None)
         binary = state.pop('binary', None)
         for k, v in state.items():
             setattr(self, k, v)
-        # If the `sonames` don't match, there *might* be a hidden bug as the
-        # unpickled Operator might be generating code that differs from that
-        # generated by the pickled Operator. For example, a stupid bug that we
-        # had to fix was due to rebuilding SymPy expressions which weren't
-        # automatically getting the flag `evaluate=False`, thus producing x+2
-        # on the unpickler instead of x+1+1).  However, different `sonames`
-        # doesn't necessarily means there's a bug: if the unpickler and the
-        # pickler are two distinct processes and the unpickler runs with a
-        # different `configuration` dictionary, then the `sonames` might indeed
-        # be different, depending on which entries in `configuration` differ.
         if soname is not None:
-            if soname != self._soname:
-                warning("The pickled and unpickled Operators have different .sonames; "
-                        "this might be a bug, or simply a harmless difference in "
-                        "`configuration`. You may check they produce the same code.")
-            self._compiler.save(self._soname, binary)
-            self._lib = self._compiler.load(self._soname)
-            self._lib.name = self._soname
+            self._compiler.save(soname, binary)
+            self._lib = self._compiler.load(soname)
+            self._lib.name = soname
 
 
 # Misc helpers
@@ -887,9 +850,15 @@ class Operator(Callable):
 
 class ArgumentsMap(dict):
 
-    def __init__(self, grid, *args, **kwargs):
-        super(ArgumentsMap, self).__init__(*args, **kwargs)
+    def __init__(self, args, grid, allocator, platform):
+        super().__init__(args)
+
         self.grid = grid
+        self.allocator = allocator
+        self.platform = platform
+
+        # Compute total used memory
+        self.memused = sum(v.nbytes for v in args.values() if isinstance(v, np.ndarray))
 
     @property
     def comm(self):
@@ -951,6 +920,14 @@ def parse_kwargs(**kwargs):
     options.setdefault('mpi', configuration['mpi'])
     for k, v in configuration['opt-options'].items():
         options.setdefault(k, v)
+    # Handle deprecations
+    deprecated_options = ('cire-mincost-inv', 'cire-mincost-sops', 'cire-maxalias')
+    for i in deprecated_options:
+        try:
+            options.pop(i)
+            warning("Ignoring deprecated optimization option `%s`" % i)
+        except KeyError:
+            pass
     kwargs['options'] = options
 
     # `opt`, mode
@@ -994,9 +971,14 @@ def parse_kwargs(**kwargs):
                                                          language=kwargs['language'])
     elif any([platform, language]):
         kwargs['compiler'] =\
-            configuration['compiler'].__new_from__(platform=kwargs['platform'],
+            configuration['compiler'].__new_with__(platform=kwargs['platform'],
                                                    language=kwargs['language'])
     else:
-        kwargs['compiler'] = configuration['compiler']
+        kwargs['compiler'] = configuration['compiler'].__new_with__()
+
+    # `allocator`
+    kwargs['allocator'] = default_allocator(
+        '%s.%s' % (kwargs['compiler'].__class__.__name__, kwargs['language'])
+    )
 
     return kwargs
