@@ -4,15 +4,16 @@ Extended SymPy hierarchy.
 
 import numpy as np
 import sympy
-from sympy import Expr, Integer, Function, Symbol
+from sympy import Expr, Integer, Function, Symbol, sympify
 
 from devito.symbolics.printer import ccode
 from devito.tools import Pickable, as_tuple, is_integer
 
-__all__ = ['CondEq', 'CondNe', 'IntDiv', 'FunctionFromPointer', 'FieldFromPointer',
-           'FieldFromComposite', 'ListInitializer', 'Byref', 'IndexedPointer',
-           'DefFunction', 'Macro', 'Literal', 'INT', 'FLOAT', 'DOUBLE', 'FLOOR',
-           'cast_mapper']
+__all__ = ['CondEq', 'CondNe', 'IntDiv', 'CallFromPointer', 'FieldFromPointer',
+           'FieldFromComposite', 'ListInitializer', 'Byref', 'IndexedPointer', 'Cast',
+           'DefFunction', 'InlineIf', 'Macro', 'MacroArgument', 'Literal', 'Deref',
+           'INT', 'FLOAT', 'DOUBLE', 'VOID', 'CEIL', 'FLOOR', 'MAX', 'MIN',
+           'SizeOf', 'rfunc', 'cast_mapper']
 
 
 class CondEq(sympy.Eq):
@@ -66,6 +67,8 @@ class IntDiv(sympy.Expr):
 
     is_Atom = True
 
+    is_commutative = True
+
     def __new__(cls, lhs, rhs, params=None):
         try:
             rhs = Integer(rhs)
@@ -91,21 +94,49 @@ class IntDiv(sympy.Expr):
     __repr__ = __str__
 
 
-class FunctionFromPointer(sympy.Expr, Pickable):
+class BasicWrapperMixin(object):
 
     """
-    Symbolic representation of the C notation ``pointer->function(params)``.
+    Abstract mixin class for objects wrapping types.Basic objects.
     """
 
-    def __new__(cls, function, pointer, params=None):
+    @property
+    def base(self):
+        """
+        The wrapped object.
+        """
+        raise NotImplementedError
+
+    @property
+    def function(self):
+        """
+        The underlying function.
+        """
+        return self.base.function
+
+    @property
+    def dtype(self):
+        """
+        The wrapped object data type.
+        """
+        return self.function.dtype
+
+
+class CallFromPointer(sympy.Expr, Pickable, BasicWrapperMixin):
+
+    """
+    Symbolic representation of the C notation ``pointer->call(params)``.
+    """
+
+    def __new__(cls, call, pointer, params=None):
         args = []
         if isinstance(pointer, str):
             pointer = Symbol(pointer)
         args.append(pointer)
-        if isinstance(function, FunctionFromPointer):
-            args.append(function)
-        elif not isinstance(function, str):
-            raise ValueError("`function` must be FunctionFromPointer or str")
+        if isinstance(call, (DefFunction, CallFromPointer)):
+            args.append(call)
+        elif not isinstance(call, str):
+            raise ValueError("`call` must be CallFromPointer or str")
         _params = []
         for p in as_tuple(params):
             if isinstance(p, str):
@@ -116,50 +147,50 @@ class FunctionFromPointer(sympy.Expr, Pickable):
                 _params.append(p)
         args.extend(_params)
         obj = sympy.Expr.__new__(cls, *args)
-        obj.function = function
+        obj.call = call
         obj.pointer = pointer
         obj.params = tuple(_params)
         return obj
 
     def __str__(self):
-        return '%s->%s(%s)' % (self.pointer, self.function,
+        return '%s->%s(%s)' % (self.pointer, self.call,
                                ", ".join(str(i) for i in as_tuple(self.params)))
 
     __repr__ = __str__
 
     def _hashable_content(self):
-        return super(FunctionFromPointer, self)._hashable_content() +\
-            (self.function, self.pointer) + self.params
+        return super(CallFromPointer, self)._hashable_content() +\
+            (self.call, self.pointer) + self.params
 
     @property
     def base(self):
-        if isinstance(self.pointer, FunctionFromPointer):
-            # FunctionFromPointer may be nested
+        if isinstance(self.pointer, CallFromPointer):
+            # CallFromPointer may be nested
             return self.pointer.base
         else:
             return self.pointer
 
     # Pickling support
-    _pickle_args = ['function', 'pointer']
+    _pickle_args = ['call', 'pointer']
     _pickle_kwargs = ['params']
     __reduce_ex__ = Pickable.__reduce_ex__
 
 
-class FieldFromPointer(FunctionFromPointer, Pickable):
+class FieldFromPointer(CallFromPointer, Pickable):
 
     """
     Symbolic representation of the C notation ``pointer->field``.
     """
 
     def __new__(cls, field, pointer):
-        return FunctionFromPointer.__new__(cls, field, pointer)
+        return CallFromPointer.__new__(cls, field, pointer)
 
     def __str__(self):
         return '%s->%s' % (self.pointer, self.field)
 
     @property
     def field(self):
-        return self.function
+        return self.call
 
     # Our __new__ cannot accept the params argument
     _pickle_kwargs = []
@@ -167,7 +198,7 @@ class FieldFromPointer(FunctionFromPointer, Pickable):
     __repr__ = __str__
 
 
-class FieldFromComposite(FunctionFromPointer, Pickable):
+class FieldFromComposite(CallFromPointer, Pickable):
 
     """
     Symbolic representation of the C notation ``composite.field``,
@@ -175,14 +206,14 @@ class FieldFromComposite(FunctionFromPointer, Pickable):
     """
 
     def __new__(cls, field, composite):
-        return FunctionFromPointer.__new__(cls, field, composite)
+        return CallFromPointer.__new__(cls, field, composite)
 
     def __str__(self):
         return '%s.%s' % (self.composite, self.field)
 
     @property
     def field(self):
-        return self.function
+        return self.call
 
     @property
     def composite(self):
@@ -225,30 +256,42 @@ class ListInitializer(sympy.Expr, Pickable):
     __reduce_ex__ = Pickable.__reduce_ex__
 
 
-class Byref(sympy.Expr, Pickable):
+class UnaryOp(sympy.Expr, Pickable, BasicWrapperMixin):
 
     """
-    Symbolic representation of the C notation ``&symbol``.
+    Symbolic representation of a unary C operator.
     """
 
-    def __new__(cls, base):
-        if isinstance(base, str):
-            base = Symbol(base)
-        elif not isinstance(base, sympy.Expr):
-            raise ValueError("`base` must be sympy.Expr or str")
+    _op = ''
+
+    def __new__(cls, base, **kwargs):
+        try:
+            # If an AbstractFunction, pull the underlying Symbol
+            base = base.indexed.label
+        except AttributeError:
+            if isinstance(base, str):
+                base = Symbol(base)
+            elif not isinstance(base, sympy.Basic):
+                raise ValueError("`base` must be sympy.Basic or str")
+
         obj = sympy.Expr.__new__(cls, base)
         obj._base = base
+
         return obj
 
     @property
     def base(self):
         return self._base
 
+    @property
+    def free_symbols(self):
+        return self.base.free_symbols
+
     def __str__(self):
         if self.base.is_Symbol:
-            return "&%s" % ccode(self.base)
+            return "%s%s" % (self._op, ccode(self.base))
         else:
-            return "&(%s)" % ccode(self.base)
+            return "%s(%s)" % (self._op, ccode(self.base))
 
     __repr__ = __str__
 
@@ -257,7 +300,54 @@ class Byref(sympy.Expr, Pickable):
     __reduce_ex__ = Pickable.__reduce_ex__
 
 
-class IndexedPointer(sympy.Expr, Pickable):
+class Byref(UnaryOp):
+
+    """
+    Symbolic representation of the C notation `&expr`.
+    """
+
+    _op = '&'
+
+
+class Deref(UnaryOp):
+
+    """
+    Symbolic representation of the C notation `*expr`.
+    """
+
+    _op = '*'
+
+
+class Cast(UnaryOp):
+
+    """
+    Symbolic representation of the C notation `(type)expr`.
+    """
+
+    _base_typ = ''
+
+    def __new__(cls, base, stars=None, **kwargs):
+        obj = super().__new__(cls, base)
+        obj._stars = stars
+        return obj
+
+    @property
+    def stars(self):
+        return self._stars
+
+    @property
+    def typ(self):
+        return '%s%s' % (self._base_typ, self.stars or '')
+
+    @property
+    def _op(self):
+        return '(%s)' % self.typ
+
+    # Pickling support
+    _pickle_kwargs = ['stars']
+
+
+class IndexedPointer(sympy.Expr, Pickable, BasicWrapperMixin):
 
     """
     Symbolic representation of the C notation ``symbol[...]``
@@ -267,13 +357,19 @@ class IndexedPointer(sympy.Expr, Pickable):
     """
 
     def __new__(cls, base, index):
-        if isinstance(base, (str, sympy.IndexedBase, sympy.Symbol)):
-            return sympy.Indexed(base, index)
-        elif not isinstance(base, sympy.Basic):
-            raise ValueError("`base` must be of type sympy.Basic")
-        obj = sympy.Expr.__new__(cls, base)
+        try:
+            # If an AbstractFunction, pull the underlying Symbol
+            base = base.indexed.label
+        except AttributeError:
+            if not isinstance(base, sympy.Basic):
+                raise ValueError("`base` must be of type sympy.Basic")
+
+        index = tuple(sympify(i) for i in as_tuple(index))
+
+        obj = sympy.Expr.__new__(cls, base, *index)
         obj._base = base
-        obj._index = as_tuple(index)
+        obj._index = index
+
         return obj
 
     @property
@@ -302,6 +398,8 @@ class DefFunction(Function, Pickable):
         https://github.com/sympy/sympy/issues/4297
     """
 
+    is_Atom = True
+
     def __new__(cls, name, arguments=None):
         arguments = as_tuple(arguments)
         obj = Function.__new__(cls, name, *arguments)
@@ -317,6 +415,11 @@ class DefFunction(Function, Pickable):
     def arguments(self):
         return self._arguments
 
+    @property
+    def free_symbols(self):
+        return set().union(*[i.free_symbols for i in self.arguments
+                             if isinstance(i, Expr)])
+
     def __str__(self):
         return "%s(%s)" % (self.name, ', '.join(str(i) for i in self.arguments))
 
@@ -328,6 +431,47 @@ class DefFunction(Function, Pickable):
     __reduce_ex__ = Pickable.__reduce_ex__
 
 
+class InlineIf(sympy.Expr, Pickable):
+
+    """
+    Symbolic representation of the C notation ``(C) ? a : b``.
+    """
+
+    def __new__(cls, cond, true_expr, false_expr):
+        if not isinstance(cond, sympy.core.relational.Relational):
+            raise ValueError("`cond` must be of type sympy.relational.Relational")
+        if not isinstance(true_expr, sympy.Expr):
+            raise ValueError("`true_expr` must be of type sympy.Expr")
+        if not isinstance(false_expr, sympy.Expr):
+            raise ValueError("`false_expr` must be of type sympy.Expr")
+        obj = sympy.Expr.__new__(cls, cond, true_expr, false_expr)
+        obj._cond = cond
+        obj._true_expr = true_expr
+        obj._false_expr = false_expr
+        return obj
+
+    @property
+    def cond(self):
+        return self._cond
+
+    @property
+    def true_expr(self):
+        return self._true_expr
+
+    @property
+    def false_expr(self):
+        return self._false_expr
+
+    def __str__(self):
+        return "(%s) ? %s : %s" % (self.cond, self.true_expr, self.false_expr)
+
+    __repr__ = __str__
+
+    # Pickling support
+    _pickle_args = ['cond', 'true_expr', 'false_expr']
+    __reduce_ex__ = Pickable.__reduce_ex__
+
+
 class Macro(sympy.Symbol):
 
     """
@@ -336,6 +480,18 @@ class Macro(sympy.Symbol):
     pass
 
 
+class MacroArgument(sympy.Symbol):
+
+    """
+    Symbolic representation of a C macro.
+    """
+
+    def __str__(self):
+        return "(%s)" % self.name
+
+    __repr__ = __str__
+
+
 class Literal(sympy.Symbol):
 
     """
@@ -344,9 +500,94 @@ class Literal(sympy.Symbol):
     pass
 
 
-INT = Function('INT')
-FLOAT = Function('FLOAT')
-DOUBLE = Function('DOUBLE')
-FLOOR = Function('floor')
+# Shortcuts (mostly for retrocompatibility)
 
-cast_mapper = {np.float32: FLOAT, float: DOUBLE, np.float64: DOUBLE}
+class INT(Cast):
+    _base_typ = 'int'
+
+
+class FLOAT(Cast):
+    _base_typ = 'float'
+
+
+class DOUBLE(Cast):
+    _base_typ = 'double'
+
+
+class VOID(Cast):
+    _base_typ = 'void'
+
+
+class CastStar(object):
+
+    base = None
+
+    def __new__(cls, base=''):
+        return cls.base(base, '*')
+
+
+class INTP(CastStar):
+    base = INT
+
+
+class FLOATP(CastStar):
+    base = FLOAT
+
+
+class DOUBLEP(CastStar):
+    base = DOUBLE
+
+
+# Some other utility functions
+
+CEIL = Function('ceil')
+FLOOR = Function('floor')
+MAX = Function('MAX')
+MIN = Function('MIN')
+
+# DefFunction, unlike sympy.Function, generates e.g. `sizeof(float)`, not `sizeof(float_)`
+SizeOf = lambda *args: DefFunction('sizeof', tuple(args))
+
+
+def rfunc(func, item, *args):
+    """
+    A utility function that recursively generates 'func' nested relations.
+
+    Examples
+    ----------
+    >> rfunc(min, [a, b, c, d])
+    MIN(a, MIN(b, MIN(c, d)))
+
+    >> rfunc(max, [a, b, c, d])
+    MAX(a, MAX(b, MAX(c, d)))
+    """
+
+    assert func in rfunc_mapper
+    rf = rfunc_mapper[func]
+
+    if len(args) == 0:
+        return item
+    else:
+        return rf(item, rfunc(func, *args))
+
+
+cast_mapper = {
+    int: INT,
+    np.int32: INT,
+    np.int64: INT,
+    np.float32: FLOAT,
+    float: DOUBLE,
+    np.float64: DOUBLE,
+
+    (int, '*'): INTP,
+    (np.int32, '*'): INTP,
+    (np.int64, '*'): INTP,
+    (np.float32, '*'): FLOATP,
+    (float, '*'): DOUBLEP,
+    (np.float64, '*'): DOUBLEP
+}
+
+rfunc_mapper = {
+    min: MIN,
+    max: MAX,
+}
