@@ -5,15 +5,15 @@ Utilities to turn SymPy objects into C strings.
 import numpy as np
 
 from mpmath.libmp import prec_to_dps, to_str
-from sympy.printing.ccode import C99CodePrinter
+from sympy.printing.precedence import precedence
+from sympy.printing.c import C99CodePrinter
+
 
 __all__ = ['ccode']
 
 
 class CodePrinter(C99CodePrinter):
 
-    custom_functions = {'INT': '(int)', 'FLOAT': '(float)', 'DOUBLE': '(double)'}
-
     """
     Decorator for sympy.printing.ccode.CCodePrinter.
 
@@ -25,7 +25,6 @@ class CodePrinter(C99CodePrinter):
     def __init__(self, dtype=np.float32, settings={}):
         self.dtype = dtype
         C99CodePrinter.__init__(self, settings)
-        self.known_functions.update(self.custom_functions)
 
     def _print_Function(self, expr):
         # There exist no unknown Functions
@@ -44,10 +43,19 @@ class CodePrinter(C99CodePrinter):
         --------
         U[t,x,y,z] -> U[t][x][y][z]
         """
-        output = self._print(expr.base.label) \
+        return self._print(expr.base.label) \
             + ''.join(['[' + self._print(x) + ']' for x in expr.indices])
 
-        return output
+    def _print_FIndexed(self, expr):
+        """
+        Print an FIndexed, that is a special Indexed, as a C-like multiarguments function.
+
+        Examples
+        --------
+        U[t,x,y,z] -> U(t,x,y,z)
+        """
+        return self._print(expr.base.label) \
+            + '(%s)' % ', '.join(self._print(x) for x in expr.indices)
 
     def _print_Rational(self, expr):
         """Print a Rational as a C-like float/float division."""
@@ -61,6 +69,17 @@ class CodePrinter(C99CodePrinter):
         else:
             return '%d.0F/%d.0F' % (p, q)
 
+    def _print_Pow(self, expr):
+        # Need to override because of issue #1627
+        # E.g., (Pow(h_x, -1) AND h_x.dtype == np.float32) => 1.0F/h_x
+        try:
+            if expr.exp == -1 and self.dtype == np.float32:
+                PREC = precedence(expr)
+                return '1.0F/%s' % self.parenthesize(expr.base, PREC)
+        except AttributeError:
+            pass
+        return super()._print_Pow(expr)
+
     def _print_Mod(self, expr):
         """Print a Mod as a C-like %-based operation."""
         args = map(ccode, expr.args)
@@ -99,7 +118,9 @@ class CodePrinter(C99CodePrinter):
     def _print_Differentiable(self, expr):
         return "(" + self._print(expr._expr) + ")"
 
-    def _print_FunctionFromPointer(self, expr):
+    _print_EvalDerivative = C99CodePrinter._print_Add
+
+    def _print_CallFromPointer(self, expr):
         indices = [self._print(i) for i in expr.params]
         return "%s->%s(%s)" % (expr.pointer, expr.function, ', '.join(indices))
 
@@ -112,12 +133,17 @@ class CodePrinter(C99CodePrinter):
     def _print_ListInitializer(self, expr):
         return "{%s}" % ', '.join([self._print(i) for i in expr.params])
 
+    def _print_IndexedPointer(self, expr):
+        return "%s%s" % (expr.base, ''.join('[%s]' % self._print(i) for i in expr.index))
+
     def _print_IntDiv(self, expr):
         return expr.__str__()
 
-    _print_Byref = _print_IntDiv
-    _print_IndexedPointer = _print_IntDiv
+    _print_UnaryOp = _print_IntDiv
     _print_DefFunction = _print_IntDiv
+    _print_InlineIf = _print_IntDiv
+    _print_MacroArgument = _print_IntDiv
+    _print_IndexedData = _print_IntDiv
 
     def _print_TrigonometricFunction(self, expr):
         func_name = str(expr.func)
