@@ -9,7 +9,8 @@ from devito.tools.utils import as_tuple, filter_ordered
 from devito.tools.algorithms import toposort
 
 __all__ = ['Bunch', 'EnrichedTuple', 'ReducerMap', 'DefaultOrderedDict',
-           'OrderedSet', 'PartialOrderTuple', 'DAG', 'frozendict']
+           'OrderedSet', 'PartialOrderTuple', 'DAG', 'frozendict',
+           'UnboundedMultiTuple']
 
 
 class Bunch(object):
@@ -133,6 +134,14 @@ class ReducerMap(MultiDict):
         """Returns a dictionary with reduced/unique values for all keys."""
         return {k: self.reduce(key=k) for k in self}
 
+    def reduce_inplace(self):
+        """
+        Like `reduce_all`, but it modifies self inplace, rather than
+        returning the result as a new dict.
+        """
+        for k, v in self.reduce_all().items():
+            self[k] = v
+
 
 class DefaultOrderedDict(OrderedDict):
     # Source: http://stackoverflow.com/a/6190500/562769
@@ -472,3 +481,72 @@ class frozendict(Mapping):
                 h ^= hash((key, value))
             self._hash = h
         return self._hash
+
+
+class UnboundedMultiTuple(object):
+
+    """
+    An UnboundedMultiTuple is an ordered collection of tuples that can be
+    infinitely iterated over.
+
+    Examples
+    --------
+    >>> ub = UnboundedMultiTuple([1, 2], [3, 4])
+    >>> ub
+    UnboundedMultiTuple((1, 2), (3, 4))
+    >>> ub.iter()
+    >>> ub
+    UnboundedMultiTuple(*(1, 2), (3, 4))
+    >>> ub.next()
+    1
+    >>> ub.next()
+    2
+    >>> ub.next()
+    2
+    >>> ub.iter()
+    >>> ub.iter()  # No effect, tip has reached the last tuple
+    >>> ub.iter()  # No effect, tip has reached the last tuple
+    >>> ub
+    UnboundedMultiTuple((1, 2), *(3, 4))
+    >>> ub.next()
+    3
+    >>> ub.next()
+    4
+    >>> ub.iter()  # Reloads the last iterator
+    >>> ub.next()
+    3
+    """
+
+    def __init__(self, *items):
+        # Normalize input
+        nitems = []
+        for i in as_tuple(items):
+            if isinstance(i, Iterable):
+                nitems.append(tuple(i))
+            else:
+                raise ValueError("Expected sequence, got %s" % type(i))
+
+        self.items = tuple(nitems)
+        self.tip = -1
+        self.curiter = None
+
+    def __repr__(self):
+        items = [str(i) for i in self.items]
+        if self.curiter is not None:
+            items[self.tip] = "*%s" % items[self.tip]
+        return "%s(%s)" % (self.__class__.__name__, ", ".join(items))
+
+    def iter(self):
+        if not self.items:
+            raise ValueError("No tuples available")
+        self.tip = min(self.tip + 1, max(len(self.items) - 1, 0))
+        self.curiter = iter(self.items[self.tip])
+
+    def next(self):
+        if self.curiter is None:
+            raise StopIteration
+        try:
+            default = self.items[self.tip][-1]
+        except IndexError:
+            default = None
+        return next(self.curiter, default)
