@@ -13,7 +13,8 @@ from cgen import dtype_to_ctype as cgen_dtype_to_ctype
 __all__ = ['prod', 'as_tuple', 'is_integer', 'generator', 'grouper', 'split', 'roundm',
            'powerset', 'invert', 'flatten', 'single_or', 'filter_ordered', 'as_mapper',
            'filter_sorted', 'dtype_to_cstr', 'dtype_to_ctype', 'dtype_to_mpitype',
-           'ctypes_to_cstr', 'ctypes_pointer', 'pprint', 'sweep', 'all_equal', 'as_list']
+           'ctypes_to_cstr', 'ctypes_pointer', 'pprint', 'sweep', 'all_equal', 'as_list',
+           'indices_to_slices', 'indices_to_sections', 'transitive_closure']
 
 
 def prod(iterable, initial=1):
@@ -29,7 +30,7 @@ def as_list(item, type=None, length=None):
 
 def as_tuple(item, type=None, length=None):
     """
-    Force item to a tuple.
+    Force item to a tuple. Passes tuple subclasses through also.
 
     Partly extracted from: https://github.com/OP2/PyOP2/.
     """
@@ -38,6 +39,9 @@ def as_tuple(item, type=None, length=None):
         t = ()
     elif isinstance(item, (str, sympy.Function)):
         t = (item,)
+    elif isinstance(item, tuple):
+        # this makes tuple subclasses pass through
+        t = item
     else:
         # Convert iterable to list...
         try:
@@ -45,6 +49,7 @@ def as_tuple(item, type=None, length=None):
         # ... or create a list of a single item
         except (TypeError, NotImplementedError):
             t = (item,) * (length or 1)
+
     if length and not len(t) == length:
         raise ValueError("Tuple needs to be of length %d" % length)
     if type and not all(isinstance(i, type) for i in t):
@@ -59,7 +64,7 @@ def as_mapper(iterable, key=None, get=None):
     """
     key = key or (lambda i: i)
     get = get or (lambda i: i)
-    mapper = {}
+    mapper = OrderedDict()
     for i in iterable:
         mapper.setdefault(key(i), []).append(get(i))
     return mapper
@@ -165,7 +170,14 @@ def filter_ordered(elements, key=None):
             return unordered[np.argsort(inds)].tolist()
         except:
             return sorted(list(set(elements)), key=elements.index)
-    return [e for e in elements if not (key(e) in seen or seen.add(key(e)))]
+    else:
+        ret = []
+        for e in elements:
+            k = key(e)
+            if k not in seen:
+                ret.append(e)
+                seen.add(k)
+        return ret
 
 
 def filter_sorted(elements, key=None):
@@ -185,15 +197,14 @@ def dtype_to_cstr(dtype):
 
 def dtype_to_ctype(dtype):
     """Translate numpy.dtype into a ctypes type."""
-    return {np.int32: ctypes.c_int,
-            np.float32: ctypes.c_float,
-            np.int64: ctypes.c_int64,
-            np.float64: ctypes.c_double}[dtype]
+    return np.ctypeslib.as_ctypes_type(dtype)
 
 
 def dtype_to_mpitype(dtype):
     """Map numpy types to MPI datatypes."""
-    return {np.int32: 'MPI_INT',
+    return {np.ubyte: 'MPI_BYTE',
+            np.ushort: 'MPI_UNSIGNED_SHORT',
+            np.int32: 'MPI_INT',
             np.float32: 'MPI_FLOAT',
             np.int64: 'MPI_LONG',
             np.float64: 'MPI_DOUBLE'}[dtype]
@@ -213,17 +224,21 @@ def ctypes_to_cstr(ctype, toarray=None):
     elif issubclass(ctype, ctypes.Array):
         return '%s[%d]' % (ctypes_to_cstr(ctype._type_, toarray), ctype._length_)
     elif ctype.__name__.startswith('c_'):
+        if ctype.__name__.startswith('c_volatile_'):
+            name = 'volatile %s' % ctype.__name__[11:]
+        else:
+            name = ctype.__name__[2:]
         # A primitive datatype
         # FIXME: Is there a better way of extracting the C typename ?
         # Here, we're following the ctypes convention that each basic type has
         # either the format c_X_p or c_X, where X is the C typename, for instance
         # `int` or `float`.
-        if ctype.__name__.endswith('_p'):
-            return '%s *' % ctype.__name__[2:-2]
+        if name.endswith('_p'):
+            return '%s *' % name[:-2]
         elif toarray:
-            return '%s %s' % (ctype.__name__[2:], toarray)
+            return '%s %s' % (name, toarray)
         else:
-            return ctype.__name__[2:]
+            return name
     else:
         # A custom datatype (e.g., a typedef-ed pointer to struct)
         return ctype.__name__
@@ -254,3 +269,71 @@ def sweep(parameters, keys=None):
                     for v in sweep_values]
     for vals in product(*sweep_values):
         yield dict(zip(keys, vals))
+
+
+def indices_to_slices(inputlist):
+    """
+    Convert a flatten list of indices to a list of slices.
+
+    Extracted from:
+        https://stackoverflow.com/questions/10987777/\
+                python-converting-a-list-of-indices-to-slices
+
+    Examples
+    --------
+    >>> indices_to_slices([0,2,3,4,5,6,12,99,100,101,102,13,14,18,19,20,25])
+    [(0, 1), (2, 7), (12, 15), (18, 21), (25, 26), (99, 103)]
+    """
+    inputlist.sort()
+    pointers = np.where(np.diff(inputlist) > 1)[0]
+    pointers = zip(np.r_[0, pointers+1], np.r_[pointers, len(inputlist)-1])
+    slices = [(inputlist[i], inputlist[j]+1) for i, j in pointers]
+    return slices
+
+
+def indices_to_sections(inputlist):
+    """
+    Convert a flatten list of indices to a list of sections.
+
+    A section is a (start, size) tuple.
+
+    Examples
+    --------
+    >>> indices_to_sections([0,2,3,4,5,6,12,99,100,101,102,13,14,18,19,20,25])
+    [(0, 1), (2, 5), (12, 3), (18, 3), (25, 1), (99, 4)]
+    """
+    slices = indices_to_slices(inputlist)
+    sections = [(i, j - i) for i, j in slices]
+    return sections
+
+
+def reachable_items(R, k, visited):
+    try:
+        ans = R[k]
+        if ans != [] and ans not in visited:
+            visited.append(ans)
+            ans = reachable_items(R, ans, visited)
+        return ans
+    except:
+        return k
+
+
+def transitive_closure(R):
+    '''
+    Partially inherited from: https://www.buzzphp.com/posts/transitive-closure
+    Helps to collapse paths in a graph. In other words, helps to simplfiy a mapper's
+    keys and values when values also appears in keys.
+
+    Example
+    -------
+    mapper = {a:b, b:c, c:d}
+    mapper = transitive_closure(mapper)
+
+    mapper
+    {a:d, b:d, c:d}
+    '''
+    ans = dict()
+    for k in R.keys():
+        visited = []
+        ans[k] = reachable_items(R, k, visited)
+    return ans
