@@ -1,15 +1,20 @@
 # Need only from within Devito
 from .basic import *  # noqa
 from .array import *  # noqa
+from .lazy import *  # noqa
+from .misc import *  # noqa
 
 # Needed both within and outside Devito
 from .dimension import *  # noqa
 from .utils import *  # noqa
 from .caching import _SymbolCache, CacheManager  # noqa
-
-# Needed only outside Devito
 from .equation import *  # noqa
 from .constant import *  # noqa
+
+# Some more internal types which depend on some of the types above
+from .parallel import *  # noqa
+
+# Needed only outside Devito
 from .grid import *  # noqa
 from .dense import * # noqa
 from .relational import *  # noqa
