@@ -11,12 +11,12 @@ from cached_property import cached_property
 from cgen import Struct, Value
 
 from devito.data import default_allocator
-from devito.finite_differences import Evaluable
 from devito.symbolics import aligned_indices
-from devito.tools import (Pickable, ctypes_to_cstr, dtype_to_cstr, dtype_to_ctype,
-                          frozendict, memoized_meth)
+from devito.tools import (Pickable, as_tuple, ctypes_to_cstr, dtype_to_cstr,
+                          dtype_to_ctype, frozendict, memoized_meth, sympy_mutex)
 from devito.types.args import ArgProvider
 from devito.types.caching import Cached
+from devito.types.lazy import Evaluable
 from devito.types.utils import DimensionTuple
 
 __all__ = ['Symbol', 'Scalar', 'Indexed', 'Object', 'LocalObject', 'CompositeObject']
@@ -26,10 +26,101 @@ Size = namedtuple('Size', 'left right')
 Offset = namedtuple('Offset', 'left right')
 
 
-class Basic(object):
+class CodeSymbol(object):
 
     """
-    Three relevant types inherit from this class:
+    Abstract base class for objects representing symbols in the generated code.
+    """
+
+    @abc.abstractmethod
+    def __init__(self, *args, **kwargs):
+        return
+
+    @abc.abstractproperty
+    def _C_name(self):
+        """
+        The name of the object in the generated code.
+
+        Returns
+        -------
+        str
+        """
+        return
+
+    @abc.abstractproperty
+    def _C_typename(self):
+        """
+        The type of the object in the generated code.
+
+        Returns
+        -------
+        str
+        """
+        return
+
+    @abc.abstractproperty
+    def _C_typedata(self):
+        """
+        The type of the data values in the generated code.
+
+        Returns
+        -------
+        str
+        """
+        return
+
+    @abc.abstractproperty
+    def _C_ctype(self):
+        """
+        The type of the object as a ctypes object, necessary for jumping
+        from Python-land to generated-code-land.
+
+        Returns
+        -------
+        ctypes type
+        """
+        return
+
+    @property
+    def _C_typedecl(self):
+        """
+        The type declaration of the object in the generated code.
+
+        Returns
+        -------
+        cgen.Struct or None
+            None if the type of the object can be expressed with a basic type,
+            such as float or int, otherwise a cgen.Struct representing a C struct.
+        """
+        return
+
+    @property
+    def _C_symbol(self):
+        """
+        The entry symbol. This may or may not coincide with the symbol used
+        to construct symbolic expressions.
+
+        Returns
+        -------
+        Basic
+        """
+        return self
+
+    @property
+    def _C_aliases(self):
+        """
+        The symbols aliasing `self`, thus pointing to the same object in the
+        generated code.
+        """
+        return (self,)
+
+
+class Basic(CodeSymbol):
+
+    """
+    Abstract base class for objects to construct symbolic expressions.
+
+    Four relevant types inherit from this class:
 
         * AbstractSymbol: represents a scalar; may carry data; may be used
                           to build equations.
@@ -68,6 +159,7 @@ class Basic(object):
     is_ArrayBasic = False
     is_Array = False
     is_PointerArray = False
+    is_ObjectArray = False
     is_Object = False
     is_LocalObject = False
 
@@ -82,6 +174,7 @@ class Basic(object):
     is_DiscreteFunction = False
     is_Function = False
     is_TimeFunction = False
+    is_TempFunction = False
     is_SparseTimeFunction = False
     is_SparseFunction = False
     is_PrecomputedSparseFunction = False
@@ -96,73 +189,10 @@ class Basic(object):
 
     # Basic symbolic object properties
     is_Scalar = False
-    is_Tensor = False
 
     # Some other properties
     is_PerfKnob = False  # Does it impact the Operator performance?
 
-    @abc.abstractmethod
-    def __init__(self, *args, **kwargs):
-        return
-
-    @abc.abstractproperty
-    def _C_name(self):
-        """
-        The C-level name of the object.
-
-        Returns
-        -------
-        str
-        """
-        return
-
-    @abc.abstractproperty
-    def _C_typename(self):
-        """
-        The C-level type of the object.
-
-        Returns
-        -------
-        str
-        """
-        return
-
-    @abc.abstractproperty
-    def _C_typedata(self):
-        """
-        The C-level type of the data values.
-
-        Returns
-        -------
-        str
-        """
-        return
-
-    @abc.abstractproperty
-    def _C_ctype(self):
-        """
-        The C-level type of the object, as a ctypes object, suitable for type
-        checking when calling functions via ctypes.
-
-        Returns
-        -------
-        ctypes type
-        """
-        return
-
-    @property
-    def _C_typedecl(self):
-        """
-        The C-level struct declaration representing the object.
-
-        Returns
-        -------
-        cgen.Struct or None
-            None if the object C type can be expressed with a basic C type,
-            such as float or int.
-        """
-        return
-
 
 class AbstractSymbol(sympy.Symbol, Basic, Pickable, Evaluable):
 
@@ -205,6 +235,10 @@ class AbstractSymbol(sympy.Symbol, Basic, Pickable, Evaluable):
     def _filter_assumptions(cls, **kwargs):
         """Extract sympy.Symbol-specific kwargs."""
         assumptions = {}
+        # pop predefined assumptions
+        for key in ('real', 'imaginary', 'commutative'):
+            kwargs.pop(key, None)
+        # extract sympy.Symbol-specific kwargs
         for i in list(kwargs):
             if i in _assume_rules.defined_facts:
                 assumptions[i] = kwargs.pop(i)
@@ -293,9 +327,7 @@ class AbstractSymbol(sympy.Symbol, Basic, Pickable, Evaluable):
         return '%s%s' % ('const ' if self.is_const else '',
                          dtype_to_cstr(self.dtype))
 
-    @property
-    def _C_typedata(self):
-        return dtype_to_cstr(self.dtype)
+    _C_typedata = _C_typename
 
     @property
     def _C_ctype(self):
@@ -320,6 +352,11 @@ class AbstractSymbol(sympy.Symbol, Basic, Pickable, Evaluable):
     _pickle_kwargs = ['name', 'dtype', 'is_const']
     __reduce_ex__ = Pickable.__reduce_ex__
 
+    def __getnewargs_ex__(self):
+        args, kwargs = Pickable.__getnewargs_ex__(self)
+        kwargs.update(self.assumptions0)
+        return args, kwargs
+
 
 class Symbol(AbstractSymbol, Cached):
 
@@ -355,22 +392,23 @@ class Symbol(AbstractSymbol, Cached):
         return frozendict(key)
 
     def __new__(cls, *args, **kwargs):
-        key = cls._cache_key(*args, **kwargs)
+        assumptions, kwargs = cls._filter_assumptions(**kwargs)
+        key = cls._cache_key(*args, **{**assumptions, **kwargs})
         obj = cls._cache_get(key)
 
         if obj is not None:
             return obj
 
         # Not in cache. Create a new Symbol via sympy.Symbol
-        name = kwargs.get('name') or args[0]
-        assumptions, kwargs = cls._filter_assumptions(**kwargs)
+        args = list(args)
+        name = kwargs.pop('name', None) or args.pop(0)
 
         # Note: use __xnew__ to bypass sympy caching
         newobj = sympy.Symbol.__xnew__(cls, name, **assumptions)
 
         # Initialization
         newobj._dtype = cls.__dtype_setup__(**kwargs)
-        newobj.__init_finalize__(*args, **kwargs)
+        newobj.__init_finalize__(name, *args, **kwargs)
 
         # Store new instance in symbol cache
         Cached.__init__(newobj, key)
@@ -451,8 +489,32 @@ class Scalar(Symbol, ArgProvider):
     def __dtype_setup__(cls, **kwargs):
         return kwargs.get('dtype', np.float32)
 
+    @property
+    def default_value(self):
+        return None
+
+    @property
+    def _arg_names(self):
+        return (self.name,)
+
+    def _arg_defaults(self, **kwargs):
+        if self.default_value is None:
+            # It is possible that the Scalar value is provided indirectly
+            # through a wrapper object (e.g., a Dimension spacing `h_x` gets its
+            # value via a Grid object)
+            return {}
+        else:
+            return {self.name: self.default_value}
+
+    def _arg_values(self, **kwargs):
+        if self.name in kwargs:
+            return {self.name: kwargs.pop(self.name)}
+        else:
+            return self._arg_defaults()
+
+
+class AbstractTensor(sympy.ImmutableDenseMatrix, Basic, Pickable, Evaluable):
 
-class AbstractTensor(sympy.ImmutableDenseMatrix, Basic, Cached, Pickable, Evaluable):
     """
     Base class for vector and tensor valued functions. It inherits from and
     mimicks the behavior of a sympy.ImmutableDenseMatrix.
@@ -472,12 +534,13 @@ class AbstractTensor(sympy.ImmutableDenseMatrix, Basic, Cached, Pickable, Evalua
 
     There are four relevant AbstractTensor sub-types: ::
 
-        * TensorFunction: A space-varying tensor valued function
-        * VectorFunction: A space-varying vector valued function
-        * TensorTimeFunction: A time-space-varying tensor valued function
-        * VectorTimeFunction: A time-space-varying vector valued function
+        * TensorFunction: A space-varying tensor valued function.
+        * VectorFunction: A space-varying vector valued function.
+        * TensorTimeFunction: A time-space-varying tensor valued function.
+        * VectorTimeFunction: A time-space-varying vector valued function.
     """
-    # Sympy attributes
+
+    # SymPy attributes
     is_MatrixLike = True
     is_Matrix = True
 
@@ -487,78 +550,123 @@ class AbstractTensor(sympy.ImmutableDenseMatrix, Basic, Cached, Pickable, Evalua
     is_VectorValued = False
 
     @classmethod
-    def _cache_key(cls, *args, **kwargs):
-        return cls
+    def _new(cls, *args, **kwargs):
+        if args:
+            try:
+                # Constructor if input is (rows, cols, lambda)
+                newobj = super(AbstractTensor, cls)._new(*args)
+            except ValueError:
+                # Constructor if input is list of list as (row, cols, list_of_list)
+                # doesn't work as it expects a flattened.
+                newobj = super(AbstractTensor, cls)._new(args[2])
 
-    def __new__(cls, *args, **kwargs):
-        options = kwargs.get('options', {})
-
-        key = cls._cache_key(*args, **kwargs)
-        obj = cls._cache_get(key)
-
-        if obj is not None:
-            newobj = sympy.Matrix.__new__(cls, *args, **options)
-            newobj.__init_cached__(key)
-            return newobj
-
-        name = kwargs.get('name')
-        indices, _ = cls.__indices_setup__(**kwargs)
-
-        # Create new, unique type instance from cls and the symbol name
-        newcls = type(name, (cls,), dict(cls.__dict__))
-
-        # Create the new Function object and invoke __init__
-        comps = cls.__subfunc_setup__(*args, **kwargs)
-        newobj = sympy.ImmutableDenseMatrix.__new__(newcls, comps)
-        # Initialization. The following attributes must be available
-        newobj._indices = indices
-        newobj._name = name
-        newobj._dtype = cls.__dtype_setup__(**kwargs)
-        newobj.__init_finalize__(*args, **kwargs)
-
-        # Store new instance in symbol cache
-        Cached.__init__(newobj, newcls)
+            # Filter grid and dimensions
+            grid, dimensions = newobj._infer_dims()
+            if grid is None and dimensions is None:
+                return sympy.ImmutableDenseMatrix(*args)
+            # Initialized with constructed object
+            newobj.__init_finalize__(newobj.rows, newobj.cols, newobj.flat(),
+                                     grid=grid, dimensions=dimensions)
+        else:
+            # Initialize components and create new Matrix from standard
+            # Devito inputs
+            comps = cls.__subfunc_setup__(*args, **kwargs)
+            newobj = super(AbstractTensor, cls)._new(comps)
+            newobj.__init_finalize__(*args, **kwargs)
 
         return newobj
 
-    __hash__ = Cached.__hash__
+    @classmethod
+    def _fromrep(cls, rep):
+        """
+        This the new constructor mechanism for matrices in sympy 1.9.
+        Standard new object go through `_new` but arithmetic operations directly use
+        the representation based one.
+        This class method is only accessible from an existing AbstractTensor
+        that contains a grid or dimensions.
+        """
+        newobj = super(AbstractTensor, cls)._fromrep(rep)
+        grid, dimensions = newobj._infer_dims()
+        try:
+            # This is needed when `_fromrep` is called directly in 1.9
+            # for example with mul.
+            newobj.__init_finalize__(newobj.rows, newobj.cols, newobj.flat(),
+                                     grid=grid, dimensions=dimensions)
+        except TypeError:
+            # We can end up here when `_fromrep` is called through the default _new
+            # when input `comps` don't have grid or dimensions. For example
+            # `test_non_devito_tens` in `test_tensor.py`.
+            pass
+        return newobj
+
+    def _infer_dims(self):
+        grids = {getattr(c, 'grid', None) for c in self.flat()} - {None}
+        dimensions = {d for c in self.flat()
+                      for d in getattr(c, 'dimensions', ())} - {None}
+        # If none of the components are devito objects, returns a sympy Matrix
+        if len(grids) == 0 and len(dimensions) == 0:
+            return None, None
+        elif len(grids) > 0:
+            dimensions = None
+            assert len(grids) == 1
+            grid = grids.pop()
+        else:
+            grid = None
+            dimensions = tuple(dimensions)
+
+        return grid, dimensions
+
+    def flat(self):
+        try:
+            return super().flat()
+        except AttributeError:
+            return self._mat
 
     def __init_finalize__(self, *args, **kwargs):
         pass
 
-    @classmethod
-    def __dtype_setup__(cls, **kwargs):
-        """Extract the object data type from ``kwargs``."""
-        return None
+    __hash__ = sympy.ImmutableDenseMatrix.__hash__
+
+    def doit(self, **hint):
+        return self
+
+    def _eval_matrix_mul(self, other):
+        """
+        Copy paste from sympy to avoid explicit call to sympy.Add
+        TODO: fix inside sympy
+        """
+        other_len = other.rows*other.cols
+        new_len = self.rows*other.cols
+        new_mat = [self.zero]*new_len
+
+        # If we multiply an n x 0 with a 0 x m, the
+        # expected behavior is to produce an n x m matrix of zeros
+        if self.cols != 0 and other.rows != 0:
+            self_cols = self.cols
+            mat = self.flat()
+            try:
+                other_mat = other.flat()
+            except AttributeError:
+                other_mat = other._mat
+            for i in range(new_len):
+                row, col = i // other.cols, i % other.cols
+                row_indices = range(self_cols*row, self_cols*(row+1))
+                col_indices = range(col, other_len, other.cols)
+                vec = [mat[a]*other_mat[b] for a, b in zip(row_indices, col_indices)]
+                new_mat[i] = sum(vec)
+
+        # Get new class and return product
+        newcls = self.classof_prod(other, new_mat)
+        return newcls._new(self.rows, other.cols, new_mat, copy=False)
 
     @classmethod
     def __subfunc_setup__(cls, *args, **kwargs):
         """Setup each component of the tensor as a Devito type."""
         return []
 
-    @classmethod
-    def __indices_setup__(cls, *args, **kwargs):
-        """Extract the object indices from ``kwargs``."""
-        return (), ()
-
-    @property
-    def name(self):
-        return self._name
-
-    @property
-    def dtype(self):
-        return self._dtype
-
-    @classmethod
-    def _new2(cls, *args, **kwargs):
-        """Bypass sympy `_new` that hard codes `Matrix.__new__` to call our own."""
-        return cls.__new__(cls, *args, **kwargs)
-
-    def applyfunc(self, f):
-        return self._new2(self.rows, self.cols, [f(x) for x in self])
-
 
 class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
+
     """
     Base class for tensor symbols, cached by both SymPy and Devito. It inherits
     from and mimicks the behaviour of a sympy.Function.
@@ -600,8 +708,13 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
         * PrecomputedSparseTimeFunction: A SparseTimeFunction that uses a custom
                                          interpolation scheme, instead of linear
                                          interpolators.
+
     """
 
+    # SymPy attributes, explicitly say these are not Matrices
+    is_MatrixLike = False
+    is_Matrix = False
+
     is_AbstractFunction = True
 
     # SymPy default assumptions
@@ -613,23 +726,35 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
     def _cache_key(cls, *args, **kwargs):
         return cls, args
 
-    def __new__(cls, *args, **kwargs):
-        options = kwargs.get('options', {'evaluate': False})
-
+    @classmethod
+    def _cache_get(cls, *args, **kwargs):
         # Is the object already in cache (e.g., f(x), f(x+1)) ?
         key = cls._cache_key(*args, **kwargs)
-        obj = cls._cache_get(key)
+        obj = super()._cache_get(key)
         if obj is not None:
             return obj
 
         # Does the base object exist at least (e.g. f(x))?
-        obj = cls._cache_get(cls)
+        obj = super()._cache_get(cls)
         if obj is not None:
-            newobj = sympy.Function.__new__(cls, *args, **options)
-            newobj.__init_cached__(cls)
+            options = kwargs.get('options', {'evaluate': False})
+            with sympy_mutex:
+                newobj = sympy.Function.__new__(cls, *args, **options)
+            newobj.__init_cached__(obj)
             Cached.__init__(newobj, key)
             return newobj
 
+        # Not in cache
+        return None
+
+    def __new__(cls, *args, **kwargs):
+        # Only perform a construction if the object isn't in cache
+        obj = cls._cache_get(*args, **kwargs)
+        if obj is not None:
+            return obj
+
+        options = kwargs.get('options', {'evaluate': False})
+
         # Preprocess arguments
         args, kwargs = cls.__args_setup__(*args, **kwargs)
 
@@ -641,7 +766,8 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
         newcls = type(name, (cls,), dict(cls.__dict__))
 
         # Create the new Function object and invoke __init__
-        newobj = sympy.Function.__new__(newcls, *indices, **options)
+        with sympy_mutex:
+            newobj = sympy.Function.__new__(newcls, *indices, **options)
 
         # Initialization. The following attributes must be available
         # when executing __init_finalize__
@@ -724,7 +850,7 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
     @property
     def indices(self):
         """The indices (aka dimensions) of the object."""
-        return self.args
+        return DimensionTuple(*self.args, getters=self.dimensions)
 
     @property
     def indices_ref(self):
@@ -814,7 +940,7 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
         halo = [sympy.Add(*i, evaluate=False) for i in self._size_halo]
         padding = [sympy.Add(*i, evaluate=False) for i in self._size_padding]
         domain = [i.symbolic_size for i in self.dimensions]
-        ret = tuple(sympy.Add(i, j, k, evaluate=False)
+        ret = tuple(sympy.Add(i, j, k)
                     for i, j, k in zip(domain, halo, padding))
         return DimensionTuple(*ret, getters=self.dimensions)
 
@@ -823,6 +949,45 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
         """The wrapped IndexedData object."""
         return IndexedData(self.name, shape=self.shape, function=self.function)
 
+    @property
+    def _mem_local(self):
+        """
+        True if the associated data is allocated in the underlying platform's
+        local memory space, False otherwise.
+
+        The local memory space is:
+
+            * the host DRAM if platform=CPU
+            * the device DRAM if platform=GPU
+
+        Defaults to False because AbstractFunctions are normally allocated in
+        the host DRAM regardless of the underlying platform, and then
+        dynamically mapped to device DRAM (directly from the generated code) if
+        the underlying platform turns out to be a device.
+        """
+        return False
+
+    @property
+    def _mem_mapped(self):
+        """
+        True if the associated data is allocated in the underlying platform's
+        local memory space and subsequently mapped to the underlying platform's
+        remote memory space, False otherwise.
+
+        The local memory space is:
+
+            * the host DRAM if platform=CPU
+            * the device DRAM if platform=GPU
+
+        The remote memory space is:
+
+            * the host DRAM if platform=GPU
+            * the device DRAM if platform=CPU
+
+        Defaults to True, thus relaxing `_mem_local`.
+        """
+        return not self._mem_local
+
     @property
     def _mem_external(self):
         """
@@ -853,7 +1018,11 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
         Note that this would need to be combined with self.dtype to give the actual
         size in bytes.
         """
-        return reduce(mul, self.shape)
+        return reduce(mul, self.shape_allocated)
+
+    @property
+    def nbytes(self):
+        return self.size*self._dtype().itemsize
 
     @property
     def halo(self):
@@ -875,6 +1044,14 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
     def _C_typedata(self):
         return dtype_to_cstr(self.dtype)
 
+    @cached_property
+    def _C_symbol(self):
+        return BoundSymbol(name=self._C_name, dtype=self.dtype, function=self.function)
+
+    def _make_pointer(self):
+        """Generate a symbolic pointer to self."""
+        raise NotImplementedError
+
     @cached_property
     def _size_domain(self):
         """Number of points in the domain region."""
@@ -960,7 +1137,7 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
         """
         return default_allocator().guaranteed_alignment
 
-    def indexify(self, indices=None, lshift=False, subs=None):
+    def indexify(self, indices=None, subs=None):
         """Create a types.Indexed from the current object."""
         if indices is not None:
             return Indexed(self.indexed, *indices)
@@ -969,11 +1146,9 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
         subs = subs or {}
         subs = [{**{d.spacing: 1, -d.spacing: -1}, **subs} for d in self.dimensions]
 
-        # Add halo shift
-        shift = self._size_nodomain.left if lshift else tuple([0]*len(self.dimensions))
         # Indices after substitutions
-        indices = [sympy.sympify((a - o + f).xreplace(s)) for a, o, f, s in
-                   zip(self.args, self.origin, shift, subs)]
+        indices = [sympy.sympify((a - o).xreplace(s)) for a, o, s in
+                   zip(self.args, self.origin, subs)]
         indices = [i.xreplace({k: sympy.Integer(k) for k in i.atoms(sympy.Float)})
                    for i in indices]
         return self.indexed[indices]
@@ -991,20 +1166,31 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
         return self.__class__.__base__
 
 
-# Objects belonging to the Devito API not involving data, such as data structures
-# that need to be passed to external libraries
-
-
 class AbstractObject(Basic, sympy.Basic, Pickable):
 
     """
-    Symbol representing a generic pointer object.
+    Base class for pointers to objects with derived type.
+
+    The hierarchy is structured as follows
+
+                         AbstractObject
+                                |
+                 ---------------------------------
+                 |                               |
+              Object                       LocalObject
+                 |
+          CompositeObject
+
+    Warnings
+    --------
+    AbstractObjects are created and managed directly by Devito.
     """
 
     is_AbstractObject = True
 
     def __new__(cls, *args, **kwargs):
-        obj = sympy.Basic.__new__(cls)
+        with sympy_mutex:
+            obj = sympy.Basic.__new__(cls)
         obj.__init__(*args, **kwargs)
         return obj
 
@@ -1032,6 +1218,10 @@ class AbstractObject(Basic, sympy.Basic, Pickable):
     def _C_typename(self):
         return ctypes_to_cstr(self.dtype)
 
+    @property
+    def _C_typedata(self):
+        return self._C_typename
+
     @property
     def _C_ctype(self):
         return self.dtype
@@ -1048,7 +1238,7 @@ class AbstractObject(Basic, sympy.Basic, Pickable):
 class Object(AbstractObject, ArgProvider):
 
     """
-    Symbol representing a generic pointer object, provided by an outer scope.
+    Pointer to object with derived type, provided by an outer scope.
     """
 
     is_Object = True
@@ -1067,14 +1257,12 @@ class Object(AbstractObject, ArgProvider):
         else:
             return {self.name: self.value}
 
-    def _arg_values(self, args=None, **kwargs):
+    def _arg_values(self, **kwargs):
         """
         Produce runtime values for this Object after evaluating user input.
 
         Parameters
         ----------
-        args : dict, optional
-            Known argument values.
         **kwargs
             Dictionary of user-provided argument overrides.
         """
@@ -1087,8 +1275,8 @@ class Object(AbstractObject, ArgProvider):
 class CompositeObject(Object):
 
     """
-    Symbol representing a pointer to a composite type (e.g., a C struct),
-    provided by an outer scope.
+    Pointer to object with composite type (e.g., a C struct), provided
+    by an outer scope.
     """
 
     _dtype_cache = {}
@@ -1134,18 +1322,31 @@ class CompositeObject(Object):
 class LocalObject(AbstractObject):
 
     """
-    Symbol representing a generic pointer object, defined in the local scope.
+    Pointer to object with derived type, defined in the local scope.
     """
 
     is_LocalObject = True
 
+    dtype = None
+    """
+    LocalObjects encode their dtype as a class attribute.
+    """
+
+    def __init__(self, name, constructor_args=None, **kwargs):
+        self.name = name
+        self.constructor_args = as_tuple(constructor_args)
+
+    # Pickling support
+    _pickle_args = ['name']
+    _pickle_kwargs = ['constructor_args']
+
 
 # Extended SymPy hierarchy follows, for essentially two reasons:
 # - To keep track of `function`
 # - To override SymPy caching behaviour
 
 
-class IndexedData(sympy.IndexedBase, Pickable):
+class IndexedData(sympy.IndexedBase, Pickable, CodeSymbol):
 
     """
     Wrapper class that inserts a pointer to the symbolic data object.
@@ -1154,8 +1355,9 @@ class IndexedData(sympy.IndexedBase, Pickable):
     def __new__(cls, label, shape=None, function=None):
         # Make sure `label` is a devito.Symbol, not a sympy.Symbol
         if isinstance(label, str):
-            label = Symbol(name=label, dtype=function.dtype)
-        obj = sympy.IndexedBase.__new__(cls, label, shape)
+            label = Symbol(name=label, dtype=None)
+        with sympy_mutex:
+            obj = sympy.IndexedBase.__new__(cls, label, shape)
         obj.function = function
         return obj
 
@@ -1169,11 +1371,60 @@ class IndexedData(sympy.IndexedBase, Pickable):
         indexed = super(IndexedData, self).__getitem__(indices, **kwargs)
         return Indexed(*indexed.args)
 
+    @property
+    def _C_name(self):
+        return self.name
+
+    @cached_property
+    def _C_typename(self):
+        return ctypes_to_cstr(self._C_ctype)
+
+    @cached_property
+    def _C_typedata(self):
+        return dtype_to_cstr(self.dtype)
+
+    @cached_property
+    def _C_ctype(self):
+        return POINTER(dtype_to_ctype(self.dtype))
+
+    @property
+    def base(self):
+        return self
+
+    @property
+    def dtype(self):
+        return self.function.dtype
+
     # Pickling support
     _pickle_kwargs = ['label', 'shape', 'function']
     __reduce_ex__ = Pickable.__reduce_ex__
 
 
+class BoundSymbol(AbstractSymbol):
+
+    """
+    Wrapper class for Symbols that are bound to a symbolic data object.
+
+    Notes
+    -----
+    By deliberately inheriting from AbstractSymbol, a BoundSymbol won't be
+    in the devito cache. This will avoid cycling references in the cache
+    (e.g., an entry for a Function `u(x)` and an entry for `u._C_symbol` with
+    the latter's key including `u(x)`). This is totally fine. The BoundSymbol
+    is tied to a specific Function; once the Function gets out of scope, the
+    BoundSymbol will also become a garbage collector candidate.
+    """
+
+    def __new__(cls, *args, function=None, **kwargs):
+        obj = AbstractSymbol.__new__(cls, *args, **kwargs)
+        obj._function = function
+        return obj
+
+    @property
+    def function(self):
+        return self._function
+
+
 class Indexed(sympy.Indexed):
 
     # The two type flags have changed in upstream sympy as of version 1.1,
@@ -1186,9 +1437,17 @@ class Indexed(sympy.Indexed):
 
     is_Dimension = False
 
+    @memoized_meth
+    def __str__(self):
+        return super().__str__()
+
     def _hashable_content(self):
         return super(Indexed, self)._hashable_content() + (self.base.function,)
 
+    @cached_property
+    def indices(self):
+        return DimensionTuple(*super().indices, getters=self.function.dimensions)
+
     @property
     def function(self):
         return self.base.function
@@ -1199,7 +1458,7 @@ class Indexed(sympy.Indexed):
 
     @property
     def name(self):
-        return self.function.name
+        return self.base.name
 
     @property
     def origin(self):
@@ -1208,4 +1467,36 @@ class Indexed(sympy.Indexed):
     @cached_property
     def free_symbols(self):
         # Make it cached, since it's relatively expensive and called often
-        return super(Indexed, self).free_symbols
+        ret = super(Indexed, self).free_symbols
+        # Get rid of the IndexedBase label this Indexed stems from
+        # as in Devito we can't have it floating around in Eq's
+        ret.discard(self.base.label)
+        return ret
+
+    def compare(self, other):
+        """
+        Override `sympy.Basic.compare` to honor Devito's canonical ordering
+        of arguments.
+        In SymPy:
+
+            f[x+1] < f[x+2] < ... < f[x+9] < f[x]
+
+        While in Devito we pretend
+
+            f[x] < f[x+1] < f[x+2] < ... < f[x+9]
+
+        That is the arguments need to be ordered monothonically based on the indices
+        so that the symbolic trees of two derivative expressions can be compared
+        argument-wise.
+        """
+        if (self.__class__ != other.__class__) or (self.function is not other.function):
+            return super().compare(other)
+        for l, r in zip(self.indices, other.indices):
+            try:
+                c = int(sympy.sign(l - r))
+            except TypeError:
+                # E.g., `l=x+1` and `r=y` or `r=sqrt(x)`
+                c = l.compare(r)
+            if c:
+                return c
+        return 0
