@@ -27,7 +27,7 @@ from devito.types.caching import CacheManager
 from devito.types.basic import AbstractFunction, Size
 from devito.types.utils import Buffer, DimensionTuple, NODE, CELL
 
-__all__ = ['Function', 'TimeFunction']
+__all__ = ['Function', 'TimeFunction', 'SubFunction', 'TempFunction']
 
 
 RegionMeta = namedtuple('RegionMeta', 'offset size')
@@ -52,7 +52,11 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
 
     is_Input = True
     is_DiscreteFunction = True
-    is_Tensor = True
+
+    _DataType = Data
+    """
+    The type of the underlying data object.
+    """
 
     def __init_finalize__(self, *args, **kwargs):
         # A `Distributor` to handle domain decomposition (only relevant for MPI)
@@ -76,7 +80,7 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         # Data-related properties and data initialization
         self._data = None
         self._first_touch = kwargs.get('first_touch', configuration['first-touch'])
-        self._allocator = kwargs.get('allocator', default_allocator())
+        self._allocator = kwargs.get('allocator') or default_allocator()
         initializer = kwargs.get('initializer')
         if initializer is None or callable(initializer):
             # Initialization postponed until the first access to .data
@@ -102,9 +106,10 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         # is that they are indeed the same exact object
         return self is other
 
-    _subs = Differentiable._subs
+    def __hash__(self):
+        return id(self)
 
-    __hash__ = AbstractFunction.__hash__  # Required since we're overriding __eq__
+    _subs = Differentiable._subs
 
     def _allocate_memory(func):
         """Allocate memory as a Data."""
@@ -117,9 +122,10 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
                 CacheManager.clear(force=False)
 
                 # Allocate the actual data object
-                self._data = Data(self.shape_allocated, self.dtype,
-                                  modulo=self._mask_modulo, allocator=self._allocator,
-                                  distributor=self._distributor)
+                self._data = self._DataType(self.shape_allocated, self.dtype,
+                                            modulo=self._mask_modulo,
+                                            allocator=self._allocator,
+                                            distributor=self._distributor)
 
                 # Initialize data
                 if self._first_touch:
@@ -277,8 +283,9 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         -----
         In an MPI context, this is the *local* with_halo region shape.
         """
-        return tuple(j + i + k for i, (j, k) in zip(self._shape_with_inhalo,
-                                                    self._padding))
+        return DimensionTuple(*[j + i + k for i, (j, k) in zip(self._shape_with_inhalo,
+                                                               self._padding)],
+                              getters=self.dimensions)
 
     @cached_property
     def shape_global(self):
@@ -299,6 +306,15 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
             retval.append(size.glb if size is not None else s)
         return tuple(retval)
 
+    @property
+    def size_global(self):
+        """
+        The global number of elements this object is expected to store in memory.
+        Note that this would need to be combined with self.dtype to give the actual
+        size in bytes.
+        """
+        return reduce(mul, self.shape_global)
+
     _offset_inhalo = AbstractFunction._offset_halo
     _size_inhalo = AbstractFunction._size_halo
 
@@ -329,7 +345,12 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
                 if not self._distributor.is_boundary_rank:
                     warning(warning_msg)
                 else:
-                    for i, j, k, l in zip(left, right, self._distributor.mycoords,
+                    left_dist = [i for i, d in zip(left, self.dimensions) if d
+                                 in self._distributor.dimensions]
+                    right_dist = [i for i, d in zip(right, self.dimensions) if d
+                                  in self._distributor.dimensions]
+                    for i, j, k, l in zip(left_dist, right_dist,
+                                          self._distributor.mycoords,
                                           self._distributor.topology):
                         if l > 1 and ((j > 0 and k == 0) or (i > 0 and k == l-1)):
                             warning(warning_msg)
@@ -408,6 +429,30 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         """
         return self.data_domain
 
+    def data_gather(self, start=None, stop=None, step=1, rank=0):
+        """
+        Gather distributed `Data` attached to a `Function` onto a single rank.
+
+        Parameters
+        ----------
+        rank : int
+            The rank onto which the data will be gathered.
+        step : int or tuple of ints
+            The `slice` step in each dimension.
+        start : int or tuple of ints
+            The `slice` start in each dimension.
+        stop : int or tuple of ints
+            The final point of the `slice` to include.
+
+        Notes
+        -----
+        Alias to ``self.data._gather``.
+
+        Note that gathering data from large simulations onto a single rank may
+        result in memory blow-up and hence should use this method judiciously.
+        """
+        return self.data._gather(start=start, stop=stop, step=step, rank=rank)
+
     @property
     @_allocate_memory
     def data_domain(self):
@@ -676,7 +721,7 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
     @memoized_meth
     def _C_get_field(self, region, dim, side=None):
         """Symbolic representation of a given data region."""
-        ffp = lambda f, i: FieldFromPointer("%s[%d]" % (f, i), self._C_name)
+        ffp = lambda f, i: FieldFromPointer("%s[%d]" % (f, i), self._C_symbol)
         if region is DOMAIN:
             offset = ffp(self._C_field_owned_ofs, self._C_make_index(dim, LEFT))
             size = ffp(self._C_field_domain_size, self._C_make_index(dim))
@@ -753,9 +798,6 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         """Tuple of argument names introduced by this function."""
         return (self.name,)
 
-    # NOTE: This _arg_defaults, unlike many others, cannot be memoized
-    # since it leads to halo's not being updated when required. See
-    # test_mpi: test_haloupdate_multi_op
     def _arg_defaults(self, alias=None):
         """
         A map of default argument values defined by this symbol.
@@ -772,10 +814,6 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         for i, s in zip(key.dimensions, self.shape):
             args.update(i._arg_defaults(_min=0, size=s))
 
-        # Add MPI-related data structures
-        if self.grid is not None:
-            args.update(self.grid._arg_defaults())
-
         return args
 
     def _arg_values(self, **kwargs):
@@ -802,9 +840,6 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
                 for i, s in zip(self.dimensions, new.shape):
                     size = s - sum(self._size_nodomain[i])
                     values.update(i._arg_defaults(size=size))
-                # Add MPI-related data structures
-                if self.grid is not None:
-                    values.update(self.grid._arg_defaults())
         else:
             values = self._arg_defaults(alias=self).reduce_all()
 
@@ -834,9 +869,9 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         for i, s in zip(self.dimensions, key.shape):
             i._arg_check(args, s, intervals[i])
 
-    def _arg_as_ctype(self, args, alias=None):
+    def _arg_finalize(self, args, alias=None):
         key = alias or self
-        return ReducerMap({key.name: self._C_make_dataobj(args[key.name])})
+        return {key.name: self._C_make_dataobj(args[key.name])}
 
     # Pickling support
     _pickle_kwargs = AbstractFunction._pickle_kwargs +\
@@ -958,15 +993,19 @@ class Function(DiscreteFunction):
         else:
             raise TypeError("`space_order` must be int or 3-tuple of ints")
 
-        # Dynamically add derivative short-cuts
-        self._fd = generate_fd_shortcuts(self)
-
+        self._fd = self.__fd_setup__()
         # Flag whether it is a parameter or a variable.
         # Used at operator evaluation to evaluate the Function at the
         # variable location (i.e. if the variable is staggered in x the
         # parameter has to be computed at x + hx/2)
         self._is_parameter = kwargs.get('parameter', False)
 
+    def __fd_setup__(self):
+        """
+        Dynamically add derivative short-cuts.
+        """
+        return generate_fd_shortcuts(self.dimensions, self.space_order)
+
     @cached_property
     def _fd_priority(self):
         return 1 if self.staggered in [NODE, None] else 2
@@ -1281,6 +1320,13 @@ class TimeFunction(Function):
 
         self.save = kwargs.get('save')
 
+    def __fd_setup__(self):
+        """
+        Dynamically add derivative short-cuts.
+        """
+        return generate_fd_shortcuts(self.dimensions, self.space_order,
+                                     to=self.time_order)
+
     @classmethod
     def __indices_setup__(cls, **kwargs):
         dimensions = kwargs.get('dimensions')
@@ -1302,7 +1348,8 @@ class TimeFunction(Function):
     def __shape_setup__(cls, **kwargs):
         grid = kwargs.get('grid')
         save = kwargs.get('save') or None  # Force to None if 0/False/None/...
-        shape = kwargs.get('shape')
+        dimensions = kwargs.get('dimensions')
+        shape = kwargs.get('shape', kwargs.get('shape_global'))
         time_order = kwargs.get('time_order', 1)
 
         if grid is None:
@@ -1321,11 +1368,19 @@ class TimeFunction(Function):
                 shape.insert(cls._time_position, save)
             else:
                 raise TypeError("`save` can be None, int or Buffer, not %s" % type(save))
+        elif dimensions is None:
+            raise TypeError("`dimensions` required if both `grid` and "
+                            "`shape` are provided")
+        else:
+            shape = super(TimeFunction, cls).__shape_setup__(
+                grid=grid, shape=shape, dimensions=dimensions
+            )
+
         return tuple(shape)
 
     @cached_property
     def _fd_priority(self):
-        return super(TimeFunction, self)._fd_priority + .1
+        return 2.1 if self.staggered in [NODE, None] else 2.2
 
     @property
     def time_order(self):
@@ -1404,3 +1459,189 @@ class SubFunction(Function):
         return self._parent
 
     _pickle_kwargs = Function._pickle_kwargs + ['parent']
+
+
+class TempFunction(DiscreteFunction):
+
+    """
+    Tensor symbol used to store an intermediate sub-expression extracted from
+    one or more symbolic equations.
+
+    Users should not instantiate this class directly. TempFunctions may be created
+    by Devito to store intermediate sub-expressions ("temporary values") when the
+    user supplies the `cire-ftemps` option to an Operator.
+
+    Unlike other DiscreteFunction types, TempFunctions do not carry data directly.
+    However, they can generate Functions to override the TempFunction at Operator
+    application time (see the Examples section below).
+
+    TempFunctions are useful if the user wants to retain control over the allocation
+    and deletion of temporary storage (by default, instead, Devito uses Arrays, which
+    are allocated and deallocated upon entering and exiting C-land, respectively).
+
+    Examples
+    --------
+    The `make` method makes the TempFunction create a new Function. For more info,
+    refer to TempFunction.make.__doc__.
+
+      .. code-block:: python
+
+        op = Operator(...)
+        cfuncs = [i for i in op.input if i.is_TempFunction]
+        kwargs = {i.name: i.make(grid.shape) for i in cfuncs}
+        op.apply(..., **kwargs)
+    """
+
+    is_TempFunction = True
+
+    def __init_finalize__(self, *args, **kwargs):
+        super().__init_finalize__(*args, **kwargs)
+
+        self._pointer_dim = kwargs.get('pointer_dim')
+
+    @classmethod
+    def __indices_setup__(cls, **kwargs):
+        pointer_dim = kwargs.get('pointer_dim')
+        dimensions = as_tuple(kwargs['dimensions'])
+        if pointer_dim not in dimensions:
+            # This is a bit hacky but it does work around duplicate dimensions when
+            # it gets to pickling
+            dimensions = as_tuple(pointer_dim) + dimensions
+
+        # Sanity check
+        assert not any(d.is_NonlinearDerived for d in dimensions)
+
+        return dimensions, dimensions
+
+    def __halo_setup__(self, **kwargs):
+        pointer_dim = kwargs.get('pointer_dim')
+        dimensions = as_tuple(kwargs['dimensions'])
+        halo = as_tuple(kwargs.get('halo'))
+        if halo is None:
+            halo = tuple((0, 0) for _ in dimensions)
+        if pointer_dim is not None and pointer_dim not in dimensions:
+            halo = ((0, 0),) + as_tuple(halo)
+        return halo
+
+    @property
+    def data(self):
+        # Any attempt at allocating data by the user should fail miserably
+        raise TypeError("TempFunction cannot allocate data")
+
+    data_domain = data
+    data_with_halo = data
+    data_ro_domain = data
+    data_ro_with_halo = data
+
+    @property
+    def pointer_dim(self):
+        return self._pointer_dim
+
+    @property
+    def dim(self):
+        return self.pointer_dim
+
+    @property
+    def shape(self):
+        domain = [i.symbolic_size for i in self.dimensions]
+        return DimensionTuple(*domain, getters=self.dimensions)
+
+    @property
+    def shape_with_halo(self):
+        domain = self.shape
+        halo = [sympy.Add(*i, evaluate=False) for i in self._size_halo]
+        ret = tuple(sum(i) for i in zip(domain, halo))
+        return DimensionTuple(*ret, getters=self.dimensions)
+
+    shape_allocated = DiscreteFunction.symbolic_shape
+
+    def make(self, shape=None, initializer=None, allocator=None, **kwargs):
+        """
+        Create a Function which can be used to override this TempFunction
+        in a call to `op.apply(...)`.
+
+        Parameters
+        ----------
+        shape : tuple of ints, optional
+            Shape of the domain region in grid points.
+        initializer : callable or any object exposing the buffer interface, optional
+            Data initializer. If a callable is provided, data is allocated lazily.
+        allocator : MemoryAllocator, optional
+            Controller for memory allocation. To be used, for example, when one wants
+            to take advantage of the memory hierarchy in a NUMA architecture. Refer to
+            `default_allocator.__doc__` for more information.
+        **kwargs
+            Mapper of Operator overrides. Used to automatically derive the shape
+            if not explicitly provided.
+        """
+        if shape is None:
+            if len(kwargs) == 0:
+                raise ValueError("Either `shape` or `kwargs` (Operator overrides) "
+                                 "must be provided.")
+            shape = []
+            for n, i in enumerate(self.shape):
+                v = i.subs(kwargs)
+                if not v.is_Integer:
+                    raise ValueError("Couldn't resolve `shape[%d]=%s` with the given "
+                                     "kwargs (obtained: `%s`)" % (n, i, v))
+                shape.append(int(v))
+            shape = tuple(shape)
+        elif len(shape) != self.ndim:
+            raise ValueError("`shape` must contain %d integers, not %d"
+                             % (self.ndim, len(shape)))
+        elif not all(is_integer(i) for i in shape):
+            raise ValueError("`shape` must contain integers (got `%s`)" % str(shape))
+
+        return Function(name=self.name, dtype=self.dtype, dimensions=self.dimensions,
+                        shape=shape, halo=self.halo, initializer=initializer,
+                        allocator=allocator)
+
+    def _make_pointer(self, dim):
+        return TempFunction(name='p%s' % self.name, dtype=self.dtype, pointer_dim=dim,
+                            dimensions=self.dimensions, halo=self.halo)
+
+    def _arg_defaults(self, alias=None):
+        raise RuntimeError("TempFunction does not have default arguments ")
+
+    def _arg_values(self, **kwargs):
+        if self.name in kwargs:
+            new = kwargs.pop(self.name)
+            if isinstance(new, DiscreteFunction):
+                # Set new values and re-derive defaults
+                return new._arg_defaults().reduce_all()
+            else:
+                raise InvalidArgument("Illegal runtime value for `%s`" % self.name)
+        else:
+            raise InvalidArgument("TempFunction `%s` lacks override" % self.name)
+
+    # Pickling support
+    _pickle_kwargs = DiscreteFunction._pickle_kwargs + ['dimensions', 'pointer_dim']
+
+
+class AliasFunction(DiscreteFunction):
+
+    """
+    Tensor symbol that "aliases" another DiscreteFunction. Aliasing here means that
+    the AliasFunction logically represents another object. This is most commonly used
+    when we have a generic routine `foo(af, ...)` that we need to apply to multiple
+    DiscreteFunctions; here `af` is an AliasFunction, used in the body of `foo`.
+
+    Like a TempFunction, an AliasFunction does not carry data.
+    """
+
+    __indices_setup__ = Function.__indices_setup__
+    __shape_setup__ = Function.__shape_setup__
+
+    @property
+    def _mem_mapped(self):
+        return False
+
+    @property
+    def data(self):
+        # Any attempt at allocating data by the user should fail miserably
+        raise TypeError("AliasFunction cannot allocate data")
+
+    data_domain = data
+    data_with_halo = data
+    data_ro_domain = data
+    data_ro_with_halo = data
