@@ -7,13 +7,13 @@ from cached_property import cached_property
 from devito.data import LEFT, RIGHT
 from devito.exceptions import InvalidArgument
 from devito.logger import debug
-from devito.tools import Pickable, dtype_to_cstr, is_integer, memoized_meth
+from devito.tools import Pickable, dtype_to_cstr, is_integer
 from devito.types.args import ArgProvider
 from devito.types.basic import Symbol, DataSymbol, Scalar
 
 __all__ = ['Dimension', 'SpaceDimension', 'TimeDimension', 'DefaultDimension',
            'CustomDimension', 'SteppingDimension', 'SubDimension', 'ConditionalDimension',
-           'dimensions', 'ModuloDimension', 'IncrDimension', 'ShiftedDimension']
+           'dimensions', 'ModuloDimension', 'IncrDimension', 'BlockDimension']
 
 
 Thickness = namedtuple('Thickness', 'left right')
@@ -36,13 +36,11 @@ class Dimension(ArgProvider):
                                            |
                               ---------------------------
                               |                         |
-                       BasicDimension            DefaultDimension
+                       DerivedDimension            DefaultDimension
                               |
-                      DerivedDimension
-                              |
-            ---------------------------------------
-            |                 |                   |
-      SteppingDimension   SubDimension   ConditionalDimension
+                   ---------------------
+                   |                   |
+              SubDimension   ConditionalDimension
 
     Parameters
     ----------
@@ -101,9 +99,10 @@ class Dimension(ArgProvider):
     is_Sub = False
     is_Conditional = False
     is_Stepping = False
+    is_SubIterator = False
     is_Modulo = False
     is_Incr = False
-    is_Shifted = False
+    is_Block = False
 
     _C_typename = 'const %s' % dtype_to_cstr(np.int32)
     _C_typedata = _C_typename
@@ -122,6 +121,15 @@ class Dimension(ArgProvider):
         else:
             return BasicDimension.__new__(cls, *args, **kwargs)
 
+    @classmethod
+    def class_key(cls):
+        """
+        Overrides sympy.Symbol.class_key such that Dimensions always
+        preceed other symbols when printed (e.g. x + h_x, not h_x + x).
+        """
+        a, b, c = super().class_key()
+        return a, b - 1, c
+
     @classmethod
     def __dtype_setup__(cls, **kwargs):
         # Unlike other Symbols, Dimensions can only be integers
@@ -190,14 +198,15 @@ class Dimension(ArgProvider):
 
     @cached_property
     def _defines_symbols(self):
-        return frozenset({self.symbolic_min, self.symbolic_max, self.symbolic_size})
+        candidates = [self.symbolic_min, self.symbolic_max, self.symbolic_size,
+                      self.symbolic_incr]
+        return frozenset(i for i in candidates if not i.is_Number)
 
     @property
     def _arg_names(self):
         """Tuple of argument names introduced by the Dimension."""
         return (self.name, self.size_name, self.max_name, self.min_name)
 
-    @memoized_meth
     def _arg_defaults(self, _min=None, size=None, alias=None):
         """
         A map of default argument values defined by the Dimension.
@@ -213,10 +222,11 @@ class Dimension(ArgProvider):
             self's if None.
         """
         dim = alias or self
-        return {dim.min_name: _min or 0, dim.size_name: size,
+        return {dim.min_name: _min or 0,
+                dim.size_name: size,
                 dim.max_name: size if size is None else size-1}
 
-    def _arg_values(self, args, interval, grid, **kwargs):
+    def _arg_values(self, interval, grid, args=None, **kwargs):
         """
         Produce a map of argument values after evaluating user input. If no user
         input is provided, get a known value in ``args`` and adjust it so that no
@@ -226,12 +236,10 @@ class Dimension(ArgProvider):
 
         Parameters
         ----------
-        args : dict
-            Known argument values.
         interval : Interval
             Description of the Dimension data space.
         grid : Grid
-            Only relevant in case of MPI execution; if ``self`` is a distributed
+            Used for spacing overriding and MPI execution; if ``self`` is a distributed
             Dimension, then ``grid`` is used to translate user input into rank-local
             indices.
         **kwargs
@@ -302,6 +310,7 @@ class Dimension(ArgProvider):
     _pickle_args = ['name']
     _pickle_kwargs = ['spacing']
     __reduce_ex__ = Pickable.__reduce_ex__
+    __getnewargs_ex__ = Pickable.__getnewargs_ex__
 
 
 class BasicDimension(Dimension, Symbol):
@@ -413,9 +422,6 @@ class DerivedDimension(BasicDimension):
 
     is_Derived = True
 
-    _keymap = {}
-    """Used to create unique Dimension names based on seen kwargs."""
-
     def __init_finalize__(self, name, parent):
         assert isinstance(parent, Dimension)
         self._parent = parent
@@ -423,14 +429,6 @@ class DerivedDimension(BasicDimension):
         self.is_Time = parent.is_Time
         self.is_Space = parent.is_Space
 
-    @classmethod
-    def _gensuffix(cls, key):
-        return cls._keymap.setdefault(key, len(cls._keymap))
-
-    @classmethod
-    def _genname(cls, prefix, key):
-        return "%s%d" % (prefix, cls._gensuffix(key))
-
     @property
     def parent(self):
         return self._parent
@@ -460,6 +458,11 @@ class DerivedDimension(BasicDimension):
     _pickle_kwargs = []
 
 
+# ***
+# The Dimensions below are exposed in the user API. They can only be created by
+# the user
+
+
 class SubDimension(DerivedDimension):
 
     """
@@ -625,30 +628,48 @@ class SubDimension(DerivedDimension):
                 symbolic_thickness
             )
 
+    def overlap(self, other):
+        return (isinstance(other, SubDimension) and
+                self.root is other.root and
+                self._offset_left.extreme is other._offset_left.extreme and
+                self._offset_right.extreme is other._offset_right.extreme)
+
+    @property
+    def _arg_names(self):
+        return tuple(k.name for k, _ in self.thickness) + self.parent._arg_names
+
     def _arg_defaults(self, grid=None, **kwargs):
+        return {}
+
+    def _arg_values(self, interval, grid, **kwargs):
+        # Allow override of thickness values to disable BCs
+        # However, arguments from the user are considered global
+        # So overriding the thickness to a nonzero value should not cause
+        # boundaries to exist between ranks where they did not before
+        requested_ltkn, requested_rtkn = (
+            kwargs.get(k.name, v) for k, v in self.thickness
+        )
+
         if grid is not None and grid.is_distributed(self.root):
             # Get local thickness
             if self.local:
                 # dimension is of type ``left``/right`` - compute the 'offset'
                 # and then add 1 to get the appropriate thickness
-                ltkn = grid.distributor.glb_to_loc(self.root,
-                                                   self.thickness.left[1]-1, LEFT)
-                rtkn = grid.distributor.glb_to_loc(self.root,
-                                                   self.thickness.right[1]-1, RIGHT)
+                ltkn = grid.distributor.glb_to_loc(self.root, requested_ltkn-1, LEFT)
+                rtkn = grid.distributor.glb_to_loc(self.root, requested_rtkn-1, RIGHT)
                 ltkn = ltkn+1 if ltkn is not None else 0
                 rtkn = rtkn+1 if rtkn is not None else 0
             else:
                 # dimension is of type ``middle``
-                ltkn = grid.distributor.glb_to_loc(self.root, self.thickness.left[1],
+                ltkn = grid.distributor.glb_to_loc(self.root, requested_ltkn,
                                                    LEFT) or 0
-                rtkn = grid.distributor.glb_to_loc(self.root, self.thickness.right[1],
+                rtkn = grid.distributor.glb_to_loc(self.root, requested_rtkn,
                                                    RIGHT) or 0
-            return {i.name: v for i, v in zip(self._thickness_map, (ltkn, rtkn))}
         else:
-            return {k.name: v for k, v in self.thickness}
+            ltkn = requested_ltkn
+            rtkn = requested_rtkn
 
-    def _arg_values(self, args, interval, grid, **kwargs):
-        return self._arg_defaults(grid=grid, **kwargs)
+        return {i.name: v for i, v in zip(self._thickness_map, (ltkn, rtkn))}
 
     # Pickling support
     _pickle_args = DerivedDimension._pickle_args +\
@@ -677,7 +698,7 @@ class ConditionalDimension(DerivedDimension):
         Dimension. When it evaluates to True, the if-branch is executed. If None
         (default), ``factor`` must be provided.
     indirect : bool, optional
-        If True, use ``condition``, rather than the parent Dimension, to
+        If True, use `self`, rather than the parent Dimension, to
         index into arrays. A typical use case is when arrays are accessed
         indirectly via the ``condition`` expression. Defaults to False.
 
@@ -707,9 +728,9 @@ class ConditionalDimension(DerivedDimension):
         }
 
     Another typical use case is when one needs to constrain the execution of
-    loop iterations to make sure certain conditions are honoured. The following
-    artificial example employs indirect array accesses and uses ConditionalDimension
-    to guard against out-of-bounds accesses.
+    loop iterations so that certain conditions are honoured. The following
+    artificial example uses ConditionalDimension to guard against out-of-bounds
+    accesses in indirectly accessed arrays.
 
     >>> from sympy import And
     >>> ci = ConditionalDimension(name='ci', parent=i,
@@ -759,17 +780,418 @@ class ConditionalDimension(DerivedDimension):
     def index(self):
         return self if self.indirect is True else self.parent
 
-    @property
+    @cached_property
     def free_symbols(self):
-        retval = super(ConditionalDimension, self).free_symbols
+        retval = set(super(ConditionalDimension, self).free_symbols)
         if self.condition is not None:
             retval |= self.condition.free_symbols
+        try:
+            retval |= self.factor.free_symbols
+        except AttributeError:
+            pass
         return retval
 
     # Pickling support
     _pickle_kwargs = DerivedDimension._pickle_kwargs + ['factor', 'condition', 'indirect']
 
 
+# ***
+# The Dimensions below are for internal use only. They are created by the compiler
+# during the construction of an Operator
+
+
+class ModuloDimension(DerivedDimension):
+
+    """
+    Dimension symbol representing a non-contiguous sub-region of a given
+    ``parent`` Dimension, which cyclically produces a finite range of values,
+    such as ``0, 1, 2, 0, 1, 2, 0, ...``.
+
+    When ``modulo=None``, the ModuloDimension degenerates and keeps generating
+    the same number such as ``2, 2, 2, 2, 2, ...`` (the actual value depends
+    on ``incr``).
+
+    Parameters
+    ----------
+    name : str
+        Name of the dimension.
+    parent : Dimension
+        The Dimension from which the ModuloDimension is derived.
+    offset : expr-like, optional
+        Min value offset. Defaults to 0.
+    modulo : int, optional
+        The divisor value.
+    incr : expr-like, optional
+        The iterator increment value. Defaults to ``offset % modulo``.
+    origin : expr-like, optional
+        The expression -- typically a function of the parent Dimension -- the
+        ModuloDimension represents.
+
+    Notes
+    -----
+    This type should not be instantiated directly in user code; if in need for
+    modulo buffered iteration, use SteppingDimension instead.
+
+    About `origin` -- the ModuloDimensions `t0, t1, t2, ...` are generated by the
+    compiler to implement modulo iteration along a TimeDimension. For example,
+    `t0`'s `origin` may be `t + 0` (where `t` is a SteppingDimension), `t1`'s
+    `origin` will then be `t + 1` and so on.
+    """
+
+    is_NonlinearDerived = True
+    is_Modulo = True
+    is_SubIterator = True
+
+    def __init_finalize__(self, name, parent,
+                          offset=None, modulo=None, incr=None, origin=None):
+        super().__init_finalize__(name, parent)
+
+        # Sanity check
+        assert modulo is not None or incr is not None
+
+        self._offset = offset or 0
+        self._modulo = modulo
+        self._incr = incr
+        self._origin = origin
+
+    @property
+    def offset(self):
+        return self._offset
+
+    @property
+    def modulo(self):
+        return self._modulo
+
+    @property
+    def incr(self):
+        return self._incr
+
+    @property
+    def origin(self):
+        return self._origin
+
+    @cached_property
+    def symbolic_size(self):
+        try:
+            return sympy.Number(self.modulo)
+        except (TypeError, ValueError):
+            pass
+        try:
+            return sympy.Number(self.incr)
+        except (TypeError, ValueError):
+            return self.incr
+
+    @cached_property
+    def symbolic_min(self):
+        if self.modulo is not None:
+            return self.offset % self.modulo
+        # Make sure we return a symbolic object as this point `offset` may well
+        # be a pure Python number
+        try:
+            return sympy.Number(self.offset)
+        except (TypeError, ValueError):
+            return self.offset
+
+    @cached_property
+    def symbolic_incr(self):
+        if self._incr is not None:
+            incr = self._incr
+        else:
+            incr = self.offset
+        if self.modulo is not None:
+            incr = incr % self.modulo
+        # Make sure we return a symbolic object as this point `incr` may well
+        # be a pure Python number
+        try:
+            return sympy.Number(incr)
+        except (TypeError, ValueError):
+            return incr
+
+    def _arg_defaults(self, **kwargs):
+        """
+        A ModuloDimension provides no arguments, so this method returns an empty dict.
+        """
+        return {}
+
+    def _arg_values(self, *args, **kwargs):
+        """
+        A ModuloDimension provides no arguments, so there are no argument values
+        to be derived.
+        """
+        return {}
+
+    # Override SymPy arithmetic operators to exploit properties of modular arithmetic
+
+    def __add__(self, other):
+        # Exploit compatibility with addition:
+        # `a1 ≡ b1 (mod n) and a2 ≡ b2 (mod n)` => `a1 + a2 ≡ b1 + b2 (mod n)`
+        try:
+            if self.modulo == other.modulo:
+                return self.origin + other.origin
+        except (AttributeError, TypeError):
+            pass
+        return super().__add__(other)
+
+    def __sub__(self, other):
+        # Exploit compatibility with subtraction:
+        # `a1 ≡ b1 (mod n) and a2 ≡ b2 (mod n)` => `a1 – a2 ≡ b1 – b2 (mod n)`
+        try:
+            if self.modulo == other.modulo:
+                return self.origin - other.origin
+        except (AttributeError, TypeError):
+            pass
+        return super().__sub__(other)
+
+    # Pickling support
+    _pickle_kwargs = ['offset', 'modulo', 'incr', 'origin']
+
+
+class AbstractIncrDimension(DerivedDimension):
+
+    """
+    Dimension symbol representing a non-contiguous sub-region of a given
+    ``parent`` Dimension, with one point every ``step`` points. Thus, if
+    ``step == k``, the dimension represents the sequence ``min, min + k,
+    min + 2*k, ...``.
+
+    Parameters
+    ----------
+    name : str
+        Name of the dimension.
+    parent : Dimension
+        The Dimension from which the IncrDimension is derived.
+    _min : expr-like
+        The minimum point of the Dimension.
+    _max : expr-like
+        The maximum point of the Dimension.
+    step : expr-like, optional
+        The distance between two consecutive points. Defaults to the
+        symbolic size.
+    size : expr-like, optional
+        The symbolic size of the Dimension. Defaults to `_max-_min+1`.
+
+    Notes
+    -----
+    This type should not be instantiated directly in user code.
+    """
+
+    is_Incr = True
+
+    def __init_finalize__(self, name, parent, _min, _max, step=None, size=None):
+        super().__init_finalize__(name, parent)
+        self._min = _min
+        self._max = _max
+        self._step = step
+        self._size = size
+
+    @property
+    def size(self):
+        return self._size
+
+    @property
+    def _depth(self):
+        """
+        The depth of `self` in the hierarchy of IncrDimensions.
+        """
+        return len([i for i in self._defines if i.is_Incr])
+
+    @cached_property
+    def step(self):
+        if self._step is not None:
+            return self._step
+        else:
+            return Scalar(name=self.size_name, dtype=np.int32, is_const=True)
+
+    @cached_property
+    def symbolic_size(self):
+        if self.size is not None:
+            # Make sure we return a symbolic object as the provided size might
+            # be for example a pure int
+            try:
+                return sympy.Number(self.size)
+            except (TypeError, ValueError):
+                return self._size
+        else:
+            # The size must be given as a function of the parent's symbols
+            return self.symbolic_max - self.symbolic_min + 1
+
+    @cached_property
+    def symbolic_min(self):
+        # Make sure we return a symbolic object as the provided min might
+        # be for example a pure int
+        try:
+            return sympy.Number(self._min)
+        except (TypeError, ValueError):
+            return self._min
+
+    @cached_property
+    def symbolic_max(self):
+        # Make sure we return a symbolic object as the provided max might
+        # be for example a pure int
+        try:
+            return sympy.Number(self._max)
+        except (TypeError, ValueError):
+            return self._max
+
+    @cached_property
+    def symbolic_incr(self):
+        try:
+            return sympy.Number(self.step)
+        except (TypeError, ValueError):
+            return self.step
+
+    # Pickling support
+    _pickle_args = ['name', 'parent', 'symbolic_min', 'symbolic_max']
+    _pickle_kwargs = ['step', 'size']
+
+
+class IncrDimension(AbstractIncrDimension):
+
+    """
+    A concrete implementation of an AbstractIncrDimension.
+
+    Notes
+    -----
+    This type should not be instantiated directly in user code.
+    """
+
+    is_SubIterator = True
+
+
+class CustomDimension(BasicDimension):
+
+    """
+    Dimension defining an iteration space with known size. Unlike a DefaultDimension,
+    a CustomDimension provides more freedom -- the symbolic_{min,max,size} of the
+    # Dimension can be set at will.
+
+    Notes
+    -----
+    This type should not be instantiated directly in user code.
+    """
+
+    is_Custom = True
+
+    def __init_finalize__(self, name, symbolic_min=None, symbolic_max=None,
+                          symbolic_size=None, parent=None):
+        self._symbolic_min = symbolic_min
+        self._symbolic_max = symbolic_max
+        self._symbolic_size = symbolic_size
+        self._parent = parent
+        super().__init_finalize__(name)
+
+    @property
+    def is_Derived(self):
+        return self._parent is not None
+
+    @property
+    def parent(self):
+        return self._parent
+
+    @property
+    def root(self):
+        if self.is_Derived:
+            return self.parent.root
+        else:
+            return self
+
+    @property
+    def spacing(self):
+        if self.is_Derived:
+            return self.parent.spacing
+        else:
+            return self._spacing
+
+    @cached_property
+    def _defines(self):
+        ret = frozenset({self})
+        if self.is_Derived:
+            ret |= self.parent._defines
+        return ret
+
+    @property
+    def symbolic_min(self):
+        try:
+            return sympy.Number(self._symbolic_min)
+        except (TypeError, ValueError):
+            pass
+        if self._symbolic_min is None:
+            return super().symbolic_min
+        else:
+            return self._symbolic_min
+
+    @property
+    def symbolic_max(self):
+        try:
+            return sympy.Number(self._symbolic_max)
+        except (TypeError, ValueError):
+            pass
+        if self._symbolic_max is None:
+            return super().symbolic_max
+        else:
+            return self._symbolic_max
+
+    @property
+    def symbolic_size(self):
+        try:
+            return sympy.Number(self._symbolic_size)
+        except (TypeError, ValueError):
+            pass
+        if self._symbolic_size is None:
+            return super().symbolic_size
+        else:
+            return self._symbolic_size
+
+    def _arg_defaults(self, **kwargs):
+        return {}
+
+    def _arg_values(self, *args, **kwargs):
+        return {}
+
+    def _arg_check(self, *args):
+        """A CustomDimension performs no runtime checks."""
+        return
+
+    # Pickling support
+    _pickle_kwargs = ['symbolic_min', 'symbolic_max', 'symbolic_size', 'parent']
+
+
+class DynamicDimensionMixin(object):
+
+    """
+    A mixin to create Dimensions producing non-const Symbols.
+    """
+
+    @cached_property
+    def symbolic_size(self):
+        return Scalar(name=self.size_name, dtype=np.int32)
+
+    @cached_property
+    def symbolic_min(self):
+        return Scalar(name=self.min_name, dtype=np.int32)
+
+    @cached_property
+    def symbolic_max(self):
+        return Scalar(name=self.max_name, dtype=np.int32)
+
+
+class DynamicDimension(DynamicDimensionMixin, BasicDimension):
+    pass
+
+
+class DynamicSubDimension(DynamicDimensionMixin, SubDimension):
+
+    @classmethod
+    def _symbolic_thickness(cls, name):
+        return (Scalar(name="%s_ltkn" % name, dtype=np.int32, nonnegative=True),
+                Scalar(name="%s_rtkn" % name, dtype=np.int32, nonnegative=True))
+
+
+# ***
+# The Dimensions below are created by Devito and may eventually be
+# accessed in user code to e.g. construct or manipulate Eqs
+
+
 class SteppingDimension(DerivedDimension):
 
     """
@@ -790,6 +1212,7 @@ class SteppingDimension(DerivedDimension):
 
     is_NonlinearDerived = True
     is_Stepping = True
+    is_SubIterator = True
 
     @property
     def symbolic_min(self):
@@ -839,159 +1262,15 @@ class SteppingDimension(DerivedDimension):
         return values
 
 
-# The Dimensions below are for internal use only
-
-
-class ModuloDimension(DerivedDimension):
+class BlockDimension(AbstractIncrDimension):
 
     """
-    Dimension symbol representing a non-contiguous sub-region of a given
-    ``parent`` Dimension, which cyclically produces a finite range of values,
-    such as ``0, 1, 2, 0, 1, 2, 0, ...``.
-
-    Parameters
-    ----------
-    parent : Dimension
-        The Dimension from which the ModuloDimension is derived.
-    offset : int
-        The offset from the parent dimension
-    modulo : int
-        The divisor value.
-    name : str, optional
-        To force a different Dimension name.
-
-    Notes
-    -----
-    This type should not be instantiated directly in user code; if in need for
-    modulo buffered iteration, use SteppingDimension instead.
+    Dimension symbol for lowering TILABLE Dimensions.
     """
 
-    is_Modulo = True
-
-    def __new__(cls, parent, offset, modulo, name=None):
-        if name is None:
-            name = cls._genname(parent.name, (offset, modulo))
-        return super().__new__(cls, parent, offset, modulo, name=name)
-
-    def __init_finalize__(self, parent, offset, modulo, name=None):
-        super().__init_finalize__(name, parent)
-        self._offset = offset
-        self._modulo = modulo
-
-    @property
-    def offset(self):
-        return self._offset
-
-    @property
-    def modulo(self):
-        return self._modulo
-
-    @property
-    def origin(self):
-        return self.parent + self.offset
-
-    @cached_property
-    def symbolic_min(self):
-        return (self.root + self.offset) % self.modulo
-
-    symbolic_incr = symbolic_min
-
-    def _arg_defaults(self, **kwargs):
-        """
-        A ModuloDimension provides no arguments, so this method returns an empty dict.
-        """
-        return {}
-
-    def _arg_values(self, *args, **kwargs):
-        """
-        A ModuloDimension provides no arguments, so there are no argument values
-        to be derived.
-        """
-        return {}
-
-    # Pickling support
-    _pickle_args = ['parent', 'offset', 'modulo']
-    _pickle_kwargs = ['name']
-
-
-class IncrDimension(DerivedDimension):
-
-    """
-    Dimension symbol representing a non-contiguous sub-region of a given
-    ``parent`` Dimension, with one point every ``step`` points. Thus, if
-    ``step == k``, the dimension represents the sequence ``min, min + k,
-    min + 2*k, ...``.
-
-    Parameters
-    ----------
-    parent : Dimension
-        The Dimension from which the IncrDimension is derived.
-    _min : expr-like
-        The minimum point of the Dimension.
-    _max : expr-like
-        The maximum point of the Dimension.
-    step : int, optional
-        The distance between two consecutive points. Defaults to the
-        symbolic size.
-    name : str, optional
-        To override the default Dimension name.
-
-    Notes
-    -----
-    This type should not be instantiated directly in user code.
-    """
-
-    is_Incr = True
+    is_Block = True
     is_PerfKnob = True
 
-    def __new__(cls, parent, _min, _max, step=None, name=None):
-        if name is None:
-            name = cls._genname(parent.name, (_min, _max, step))
-        return super().__new__(cls, parent, _min, _max, step=step, name=name)
-
-    def __init_finalize__(self, parent, _min, _max, step=None, name=None):
-        super().__init_finalize__(name, parent)
-        self._min = _min
-        self._max = _max
-        self._step = step
-
-    @cached_property
-    def step(self):
-        if self._step is not None:
-            return self._step
-        else:
-            return Scalar(name=self.size_name, dtype=np.int32, is_const=True)
-
-    @cached_property
-    def symbolic_size(self):
-        # The size must be given as a function of the parent's symbols
-        return self.symbolic_max - self.symbolic_min + 1
-
-    @cached_property
-    def symbolic_min(self):
-        # Make sure we return a symbolic object as the provided min might
-        # be for example a pure int
-        try:
-            return sympy.Number(self._min)
-        except (TypeError, ValueError):
-            return self._min
-
-    @cached_property
-    def symbolic_max(self):
-        # Make sure we return a symbolic object as the provided max might
-        # be for example a pure int
-        try:
-            return sympy.Number(self._max)
-        except (TypeError, ValueError):
-            return self._max
-
-    @cached_property
-    def symbolic_incr(self):
-        try:
-            return sympy.Number(self.step)
-        except (TypeError, ValueError):
-            return self.step
-
     @cached_property
     def _arg_names(self):
         try:
@@ -1009,16 +1288,17 @@ class IncrDimension(DerivedDimension):
             # `step` not a Symbol
             return {}
 
-    def _arg_values(self, args, interval, grid, **kwargs):
+    def _arg_values(self, interval, grid, args=None, **kwargs):
         try:
             name = self.step.name
         except AttributeError:
             # `step` not a Symbol
             return {}
+
         if name in kwargs:
             return {name: kwargs.pop(name)}
-        elif isinstance(self.parent, IncrDimension):
-            # `self` is an IncrDimension within an outer IncrDimension, but
+        elif isinstance(self.parent, BlockDimension):
+            # `self` is a BlockDimension within an outer BlockDimension, but
             # no value supplied -> the sub-block will span the entire block
             return {name: args[self.parent.step.name]}
         else:
@@ -1029,7 +1309,7 @@ class IncrDimension(DerivedDimension):
                 # Avoid OOB (will end up here only in case of tiny iteration spaces)
                 return {name: 1}
 
-    def _arg_check(self, args, interval):
+    def _arg_check(self, args, *_args):
         try:
             name = self.step.name
         except AttributeError:
@@ -1037,8 +1317,8 @@ class IncrDimension(DerivedDimension):
             return
 
         value = args[name]
-        if isinstance(self.parent, IncrDimension):
-            # sub-IncrDimensions must be perfect divisors of their parent
+        if isinstance(self.parent, BlockDimension):
+            # sub-BlockDimensions must be perfect divisors of their parent
             parent_value = args[self.parent.step.name]
             if parent_value % value > 0:
                 raise InvalidArgument("Illegal block size `%s=%d`: sub-block sizes "
@@ -1055,55 +1335,6 @@ class IncrDimension(DerivedDimension):
                                       "iteration range and it will cause an OOB access"
                                       % (name, value))
 
-    # Pickling support
-    _pickle_args = ['parent', 'symbolic_min', 'symbolic_max']
-    _pickle_kwargs = ['step', 'name']
-
-
-class ShiftedDimension(IncrDimension):
-
-    """
-    A special unit-step IncrDimension with min=0 and max=parent.symbolic_size-1.
-    """
-
-    is_Shifted = True
-
-    def __new__(cls, d, name):
-        return super().__new__(cls, d, 0, d.symbolic_size - 1, step=1, name=name)
-
-    _pickle_args = ['parent', 'name']
-    _pickle_kwargs = []
-
-
-class CustomDimension(BasicDimension):
-
-    """
-    Dimension defining an iteration space with custom (known) size.
-
-    Notes
-    -----
-    This could be extended in the future for more customization (e.g., min point,
-    max point, etc.).
-    """
-
-    is_Custom = True
-
-    def __init_finalize__(self, name, symbolic_size=None):
-        self._symbolic_size = symbolic_size
-
-    @property
-    def symbolic_size(self):
-        return self._symbolic_size
-
-    def _arg_defaults(self):
-        return {}
-
-    def _arg_values(self, *args, **kwargs):
-        return {}
-
-    # Pickling support
-    _pickle_kwargs = ['symbolic_size']
-
 
 def dimensions(names):
     assert type(names) == str
