@@ -1,15 +1,14 @@
 """User API to specify equations."""
 
 import sympy
-from sympy.solvers.solveset import linear_coeffs
 
 from cached_property import cached_property
 
-from devito.finite_differences import Evaluable, default_rules
-from devito.logger import warning
+from devito.finite_differences import default_rules
 from devito.tools import as_tuple
+from devito.types.lazy import Evaluable
 
-__all__ = ['Eq', 'Inc', 'solve']
+__all__ = ['Eq', 'Inc']
 
 
 class Eq(sympy.Eq, Evaluable):
@@ -49,7 +48,7 @@ class Eq(sympy.Eq, Evaluable):
 
     Any SymPy expressions may be used in the right-hand side.
 
-    >>> from sympy import sin
+    >>> from devito import sin
     >>> Eq(f, sin(f.dx)**2)
     Eq(f(x, y), sin(Derivative(f(x, y), x))**2)
 
@@ -90,6 +89,7 @@ class Eq(sympy.Eq, Evaluable):
         eq = self.func(lhs, rhs, subdomain=self.subdomain,
                        coefficients=self.substitutions,
                        implicit_dims=self._implicit_dims)
+
         if eq._uses_symbolic_coefficients:
             # NOTE: As Coefficients.py is expanded we will not want
             # all rules to be expunged during this procress.
@@ -148,10 +148,17 @@ class Eq(sympy.Eq, Evaluable):
         else:
             TypeError('Failed to retrieve symbolic functions')
 
+    @property
+    def func(self):
+        return lambda *args, **kwargs:\
+            self.__class__(*args,
+                           subdomain=kwargs.pop('subdomain', self._subdomain),
+                           coefficients=kwargs.pop('coefficients', self._substitutions),
+                           implicit_dims=kwargs.pop('implicit_dims', self._implicit_dims),
+                           **kwargs)
+
     def xreplace(self, rules):
-        return self.func(self.lhs.xreplace(rules), rhs=self.rhs.xreplace(rules),
-                         subdomain=self._subdomain, coefficients=self._substitutions,
-                         implicit_dims=self._implicit_dims)
+        return self.func(self.lhs.xreplace(rules), self.rhs.xreplace(rules))
 
     def __str__(self):
         return "%s(%s, %s)" % (self.__class__.__name__, self.lhs, self.rhs)
@@ -208,41 +215,3 @@ class Inc(Eq):
         return "Inc(%s, %s)" % (self.lhs, self.rhs)
 
     __repr__ = __str__
-
-
-def solve(eq, target, **kwargs):
-    """
-    Algebraically rearrange an Eq w.r.t. a given symbol.
-
-    This is a wrapper around ``sympy.solve``.
-
-    Parameters
-    ----------
-    eq : expr-like
-        The equation to be rearranged.
-    target : symbol
-        The symbol w.r.t. which the equation is rearranged. May be a `Function`
-        or any other symbolic object.
-    **kwargs
-        Symbolic optimizations applied while rearranging the equation. For more
-        information. refer to ``sympy.solve.__doc__``.
-    """
-    if isinstance(eq, Eq):
-        eq = eq.lhs - eq.rhs if eq.rhs != 0 else eq.lhs
-    sols = []
-    for e, t in zip(as_tuple(eq), as_tuple(target)):
-        # Try first linear solver
-        try:
-            cc = linear_coeffs(e.evaluate, t)
-            sols.append(-cc[1]/cc[0])
-        except ValueError:
-            warning("Equation is not affine w.r.t the target, falling back to standard"
-                    "sympy.solve that may be slow")
-            kwargs['rational'] = False  # Avoid float indices
-            kwargs['simplify'] = False  # Do not attempt premature optimisation
-            sols.append(sympy.solve(e.evaluate, t, **kwargs)[0])
-    # We need to rebuild the vector/tensor as sympy.solve outputs a tuple of solutions
-    if len(sols) > 1:
-        return target.new_from_mat(sols)
-    else:
-        return sols[0]
