@@ -1,14 +1,16 @@
+from abc import ABC
 from collections import namedtuple
+from math import floor
 
 import numpy as np
 from sympy import prod
-from math import floor
+from cached_property import cached_property
 
 from devito.data import LEFT, RIGHT
-from devito.mpi import Distributor
-from devito.tools import ReducerMap, as_tuple, memoized_meth
+from devito.mpi import Distributor, MPI
+from devito.tools import ReducerMap, as_tuple
 from devito.types.args import ArgProvider
-from devito.types.constant import Constant
+from devito.types.basic import Scalar
 from devito.types.dense import Function
 from devito.types.dimension import (Dimension, SpaceDimension, TimeDimension,
                                     SteppingDimension, SubDimension)
@@ -20,7 +22,42 @@ __all__ = ['Grid', 'SubDomain', 'SubDomainSet']
 GlobalLocal = namedtuple('GlobalLocal', 'glb loc')
 
 
-class Grid(ArgProvider):
+class CartesianDiscretization(ABC):
+
+    """
+    Abstract base class for objects representing discretizations of n-dimensional
+    physical domains by congruent parallelotopes (e.g., "tiles" or "bricks").
+    """
+
+    def __init__(self, shape=None, dimensions=None, dtype=None):
+        self._shape = as_tuple(shape)
+        self._dimensions = as_tuple(dimensions)
+        self._dtype = dtype
+
+    @property
+    def shape(self):
+        """Shape of the physical domain."""
+        return self._shape
+
+    @property
+    def dimensions(self):
+        """Spatial dimensions of the computational domain."""
+        return self._dimensions
+
+    @property
+    def dim(self):
+        """Problem dimension, or number of spatial dimensions."""
+        return len(self.dimensions)
+
+    @property
+    def dtype(self):
+        """
+        Data type inherited by all Functions defined on this CartesianDiscretization.
+        """
+        return self._dtype
+
+
+class Grid(CartesianDiscretization, ArgProvider):
 
     """
     A cartesian grid that encapsulates a computational domain over which
@@ -102,46 +139,60 @@ class Grid(ArgProvider):
 
     def __init__(self, shape, extent=None, origin=None, dimensions=None,
                  time_dimension=None, dtype=np.float32, subdomains=None,
-                 comm=None):
-        self._shape = as_tuple(shape)
-        self._extent = as_tuple(extent or tuple(1. for _ in self.shape))
-        self._dtype = dtype
+                 comm=None, topology=None):
+        shape = as_tuple(shape)
 
+        # Create or pull the SpaceDimensions
         if dimensions is None:
-            # Create the spatial dimensions and constant spacing symbols
-            assert(self.dim <= 3)
-            dim_names = self._default_dimensions[:self.dim]
-            dim_spacing = tuple(self._const(name='h_%s' % n, value=v, dtype=self.dtype)
-                                for n, v in zip(dim_names, self.spacing))
-            self._dimensions = tuple(SpaceDimension(name=n, spacing=s)
-                                     for n, s in zip(dim_names, dim_spacing))
+            ndim = len(shape)
+            assert ndim <= 3
+            dim_names = self._default_dimensions[:ndim]
+            dim_spacing = tuple(Scalar(name='h_%s' % n, dtype=dtype, is_const=True)
+                                for n in dim_names)
+            dimensions = tuple(SpaceDimension(name=n, spacing=s)
+                               for n, s in zip(dim_names, dim_spacing))
         else:
-            self._dimensions = dimensions
+            for d in dimensions:
+                if not d.is_Space:
+                    raise ValueError("Cannot create Grid with Dimension `%s` "
+                                     "since it's not a SpaceDimension" % d)
+                if d.is_Derived and not d.is_Conditional:
+                    raise ValueError("Cannot create Grid with derived Dimension `%s` "
+                                     "of type `%s`" % (d, type(d)))
+            dimensions = dimensions
 
-        self._distributor = Distributor(self.shape, self.dimensions, comm)
+        super().__init__(shape, dimensions, dtype)
+
+        # Create a Distributor, used internally to implement domain decomposition
+        # by all Functions defined on this Grid
+        self._distributor = Distributor(shape, dimensions, comm, topology)
+
+        # The physical extent
+        self._extent = as_tuple(extent or tuple(1. for _ in self.shape))
 
         # Initialize SubDomains
         subdomains = tuple(i for i in (Domain(), Interior(), *as_tuple(subdomains)))
         for counter, i in enumerate(subdomains):
-            i.__subdomain_finalize__(self.dimensions, self.shape,
-                                     distributor=self._distributor, counter=counter)
+            i.__subdomain_finalize__(self, counter=counter)
         self._subdomains = subdomains
 
-        origin = as_tuple(origin or tuple(0. for _ in self.shape))
-        self._origin = tuple(self._const(name='o_%s' % d.name, value=v, dtype=self.dtype)
-                             for d, v in zip(self.dimensions, origin))
+        self._origin = as_tuple(origin or tuple(0. for _ in self.shape))
+        self._origin_symbols = tuple(Scalar(name='o_%s' % d.name, dtype=dtype,
+                                            is_const=True)
+                                     for d in self.dimensions)
 
         # Sanity check
         assert (self.dim == len(self.origin) == len(self.extent) == len(self.spacing))
 
         # Store or create default symbols for time and stepping dimensions
         if time_dimension is None:
-            spacing = self._const(name='dt', dtype=self.dtype)
-            self._time_dim = self._make_time_dim(spacing)
-            self._stepping_dim = self._make_stepping_dim(self.time_dim, name='t')
+            spacing = Scalar(name='dt', dtype=dtype, is_const=True)
+            self._time_dim = TimeDimension(name='time', spacing=spacing)
+            self._stepping_dim = SteppingDimension(name='t', parent=self.time_dim)
         elif isinstance(time_dimension, TimeDimension):
             self._time_dim = time_dimension
-            self._stepping_dim = self._make_stepping_dim(self.time_dim)
+            self._stepping_dim = SteppingDimension(name='%s_s' % self.time_dim.name,
+                                                   parent=self.time_dim)
         else:
             raise ValueError("`time_dimension` must be None or of type TimeDimension")
 
@@ -155,25 +206,27 @@ class Grid(ArgProvider):
         """Physical extent of the domain in m."""
         return self._extent
 
-    @property
-    def dtype(self):
-        """Data type inherited by all Functions defined on this Grid."""
-        return self._dtype
-
     @property
     def origin(self):
         """Physical coordinates of the domain origin."""
         return self._origin
 
     @property
-    def dimensions(self):
-        """Spatial dimensions of the computational domain."""
-        return self._dimensions
+    def origin_symbols(self):
+        """Symbols representing the grid origin in each SpaceDimension."""
+        return self._origin_symbols
 
     @property
-    def dim(self):
-        """Problem dimension, or number of spatial dimensions."""
-        return len(self.shape)
+    def origin_map(self):
+        """Map between origin symbols and their values."""
+        return dict(zip(self.origin_symbols, self.origin))
+
+    @property
+    def origin_offset(self):
+        """Offset of the local (per-process) origin from the domain origin."""
+        grid_origin = [min(i) for i in self.distributor.glb_numb]
+        assert len(grid_origin) == len(self.spacing)
+        return tuple(i*h for i, h in zip(grid_origin, self.spacing))
 
     @property
     def time_dim(self):
@@ -206,27 +259,29 @@ class Grid(ArgProvider):
         spacing = (np.array(self.extent) / (np.array(self.shape) - 1)).astype(self.dtype)
         return as_tuple(spacing)
 
-    @property
+    @cached_property
     def spacing_symbols(self):
-        """Symbols representing the grid spacing in each SpaceDimension"""
-        return as_tuple(d.spacing for d in self.dimensions)
+        """Symbols representing the grid spacing in each SpaceDimension."""
+        return as_tuple(d.root.spacing for d in self.dimensions)
 
-    @property
+    @cached_property
     def spacing_map(self):
         """Map between spacing symbols and their values for each SpaceDimension."""
-        return dict(zip(self.spacing_symbols, self.spacing))
+        mapper = {}
+        for d, s in zip(self.dimensions, self.spacing):
+            if d.is_Conditional:
+                # Special case subsampling: `Grid.dimensions` -> (xb, yb, zb)`
+                # where `xb, yb, zb` are ConditionalDimensions whose parents
+                # are SpaceDimensions
+                mapper[d.root.spacing] = s/self.dtype(d.factor)
+            elif d.is_Space:
+                # Typical case: `Grid.dimensions` -> (x, y, z)` where `x, y, z` are
+                # the SpaceDimensions
+                mapper[d.spacing] = s
+            else:
+                assert False
 
-    @property
-    def origin_offset(self):
-        """Offset of the local (per-process) origin from the domain origin."""
-        grid_origin = [min(i) for i in self.distributor.glb_numb]
-        assert len(grid_origin) == len(self.spacing)
-        return tuple(i*h for i, h in zip(grid_origin, self.spacing))
-
-    @property
-    def shape(self):
-        """Shape of the physical domain."""
-        return self._shape
+        return mapper
 
     @property
     def shape_local(self):
@@ -241,41 +296,49 @@ class Grid(ArgProvider):
 
     @property
     def distributor(self):
-        """The Distributor used for domain decomposition."""
+        """The Distributor used for MPI-decomposing the CartesianDiscretization."""
         return self._distributor
 
     @property
     def comm(self):
-        """The MPI communicator used for domain decomposition."""
+        """The MPI communicator inherited from the distributor."""
         return self._distributor.comm
 
     def is_distributed(self, dim):
-        """True if ``dim`` is a distributed Dimension, False otherwise."""
+        """
+        True if `dim` is a distributed Dimension for this CartesianDiscretization,
+        False otherwise.
+        """
         return any(dim is d for d in self.distributor.dimensions)
 
-    @property
-    def _const(self):
-        """The type to be used to create constant symbols."""
-        return Constant
+    @cached_property
+    def _arg_names(self):
+        ret = []
+        ret.append(self.time_dim.spacing.name)
+        ret.extend([i.name for i in self.origin_map])
+        for i in self.spacing_map:
+            try:
+                ret.append(i.name)
+            except AttributeError:
+                # E.g., {n*h_x: v} (the case of ConditionalDimension)
+                ret.extend([a.name for a in i.free_symbols])
+        return tuple(ret)
 
-    def _make_stepping_dim(self, time_dim, name=None):
-        """Create a SteppingDimension for this Grid."""
-        if name is None:
-            name = '%s_s' % time_dim.name
-        return SteppingDimension(name=name, parent=time_dim)
-
-    def _make_time_dim(self, spacing):
-        """Create a TimeDimension for this Grid."""
-        return TimeDimension(name='time', spacing=spacing)
-
-    @memoized_meth
     def _arg_defaults(self):
         """A map of default argument values defined by this Grid."""
         args = ReducerMap()
 
+        # Dimensions size
         for k, v in self.dimension_map.items():
             args.update(k._arg_defaults(_min=0, size=v.loc))
 
+        # Dimensions spacing
+        args.update({k.name: v for k, v in self.spacing_map.items()})
+
+        # Grid origin
+        args.update({k.name: v for k, v in self.origin_map.items()})
+
+        # MPI-related objects
         if self.distributor.is_parallel:
             distributor = self.distributor
             args[distributor._obj_comm.name] = distributor._obj_comm.value
@@ -283,6 +346,14 @@ class Grid(ArgProvider):
 
         return args
 
+    def _arg_values(self, **kwargs):
+        values = dict(self._arg_defaults())
+
+        # Override spacing and origin if necessary
+        values.update({i: kwargs[i] for i in self._arg_names if i in kwargs})
+
+        return values
+
     def __getstate__(self):
         state = self.__dict__.copy()
         # A Distributor wraps an MPI communicator, which can't and shouldn't be pickled
@@ -292,10 +363,75 @@ class Grid(ArgProvider):
     def __setstate__(self, state):
         for k, v in state.items():
             setattr(self, k, v)
-        self._distributor = Distributor(self.shape, self.dimensions)
+        self._distributor = Distributor(self.shape, self.dimensions, MPI.COMM_SELF)
 
 
-class SubDomain(object):
+class AbstractSubDomain(CartesianDiscretization):
+
+    """
+    Abstract base class for subdomains.
+    """
+
+    name = None
+    """A unique name for the SubDomain."""
+
+    def __init__(self, *args, **kwargs):
+        if self.name is None:
+            self.name = self.__class__.__name__
+
+        # All other attributes get initialized upon `__subdomain_finalize__`
+        super().__init__()
+
+    def __subdomain_finalize__(self, grid, **kwargs):
+        """
+        Finalize the subdomain initialization.
+
+        Notes
+        -----
+        Must be overridden by subclasses.
+        """
+        raise NotImplementedError
+
+    def __eq__(self, other):
+        if not isinstance(other, AbstractSubDomain):
+            return False
+        return (self.name == other.name and
+                self.dimensions == other.dimensions and
+                self.shape == other.shape and
+                self.dtype == other.dtype)
+
+    def __hash__(self):
+        return hash((self.name, self.dimensions, self.shape, self.dtype))
+
+    def __str__(self):
+        return "%s[%s%s]" % (self.__class__.__name__, self.name, self.dimensions)
+
+    __repr__ = __str__
+
+    def union(self, other):
+        """
+        Return the union of two subdomains as a new subdomain.
+        """
+        raise NotImplementedError
+
+    def intersection(self, other):
+        """
+        Return the intersection of two subdomains as a new subdomain.
+        """
+        raise NotImplementedError
+
+    def difference(self, other):
+        """
+        Return the difference of two subdomains as a new subdomain.
+        """
+        raise NotImplementedError
+
+    @property
+    def dimension_map(self):
+        return {d.root: d for d in self.dimensions}
+
+
+class SubDomain(AbstractSubDomain):
 
     """
     Base class to define Grid subdomains.
@@ -341,21 +477,13 @@ class SubDomain(object):
     Interior : An example of preset Subdomain.
     """
 
-    name = None
-    """A unique name for the SubDomain."""
-
-    def __init__(self):
-        if self.name is None:
-            raise ValueError("SubDomain requires a `name`")
-        self._dimensions = None
-
-    def __subdomain_finalize__(self, dimensions, shape, **kwargs):
+    def __subdomain_finalize__(self, grid, **kwargs):
         # Create the SubDomain's SubDimensions
         sub_dimensions = []
         sdshape = []
         counter = kwargs.get('counter', 0) - 1
-        for k, v, s in zip(self.define(dimensions).keys(),
-                           self.define(dimensions).values(), shape):
+        for k, v, s in zip(self.define(grid.dimensions).keys(),
+                           self.define(grid.dimensions).values(), grid.shape):
             if isinstance(v, Dimension):
                 sub_dimensions.append(v)
                 sdshape.append(s)
@@ -391,35 +519,10 @@ class SubDomain(object):
                         sdshape.append(thickness)
                     else:
                         raise ValueError("Expected sides 'left|right', not `%s`" % side)
-        self._dimensions = tuple(sub_dimensions)
 
-        # Compute the SubDomain shape
         self._shape = tuple(sdshape)
-
-    def __eq__(self, other):
-        if not isinstance(other, SubDomain):
-            return False
-        return self.name == other.name and self.dimensions == other.dimensions
-
-    def __hash__(self):
-        return hash((self.name, self.dimensions))
-
-    def __str__(self):
-        return "SubDomain %s[%s]" % (self.name, self.dimensions)
-
-    __repr__ = __str__
-
-    @property
-    def dimensions(self):
-        return self._dimensions
-
-    @property
-    def dimension_map(self):
-        return {d.root: d for d in self.dimensions}
-
-    @property
-    def shape(self):
-        return self._shape
+        self._dimensions = tuple(sub_dimensions)
+        self._dtype = grid.dtype
 
     def define(self, dimensions):
         """
@@ -433,31 +536,86 @@ class SubDomain(object):
         raise NotImplementedError
 
 
-class Domain(SubDomain):
+class MultiSubDimension(SubDimension):
 
     """
-    The entire computational domain (== boundary + interior).
+    A special SubDimension for graceful lowering of MultiSubDomains.
     """
 
-    name = 'domain'
+    def __init_finalize__(self, name, parent):
+        lst, rst = self._symbolic_thickness(name)
+        left = parent.symbolic_min + lst
+        right = parent.symbolic_max - rst
 
-    def define(self, dimensions):
-        return dict(zip(dimensions, dimensions))
+        super().__init_finalize__(name, parent, left, right, ((lst, 0), (rst, 0)), False)
 
 
-class Interior(SubDomain):
+class MultiSubDomain(AbstractSubDomain):
 
     """
-    The interior of the computational domain (i.e., boundaries are excluded).
+    Abstract base class for types representing groups of SubDomains.
     """
 
-    name = 'interior'
+    @classmethod
+    def _bounds_glb_to_loc(cls, dec, m, M):
+        """
+        Translate a SubDomain global bounds, that is thicknesses, into local bounds.
+        """
 
-    def define(self, dimensions):
-        return {d: ('middle', 1, 1) for d in dimensions}
+        # There are infinite ways to set `bounds_m` and `bounds_M` to set the size
+        # of an MPI-decomposed local subdomain to 0 iterations, meaning that after
+        # domain decomposition and for the given thicknesses there are 0 iterations
+        # left to execute. However, due to issue #1766, we only have one choice, or
+        # SubDomainSets might break on GPUs
+        NOITER = (dec.loc_abs_max, 1)
+
+        bounds_m = np.zeros(m.shape, dtype=m.dtype)
+        bounds_M = np.zeros(m.shape, dtype=m.dtype)
+        for j in range(m.size):
+            lmin = dec.glb_min + m[j]
+            lmax = dec.glb_max - M[j]
+
+            # Check if the subdomain doesn't intersect with the decomposition
+            if lmin < dec.loc_abs_min and lmax < dec.loc_abs_min:
+                bounds_m[j], bounds_M[j] = NOITER
+                continue
+            if lmin > dec.loc_abs_max and lmax > dec.loc_abs_max:
+                bounds_m[j], bounds_M[j] = NOITER
+                continue
+
+            if lmin < dec.loc_abs_min:
+                bounds_m[j] = 0
+            elif lmin > dec.loc_abs_max:
+                bounds_m[j], bounds_M[j] = NOITER
+                continue
+            else:
+                bounds_m[j] = dec.index_glb_to_loc(m[j], LEFT)
+
+            if lmax < dec.loc_abs_min:
+                bounds_m[j], bounds_M[j] = NOITER
+                continue
+            elif lmax >= dec.loc_abs_max:
+                bounds_M[j] = 0
+            else:
+                bounds_M[j] = dec.index_glb_to_loc(M[j], RIGHT)
+
+        return bounds_m, bounds_M
+
+    def _create_implicit_exprs(self, grid):
+        """
+        Must be overridden by subclasses.
+        """
+        raise NotImplementedError
+
+    @property
+    def implicit_dimensions(self):
+        """
+        The Dimensions implicitly defined by the MultiSubDomain.
+        """
+        return ()
 
 
-class SubDomainSet(SubDomain):
+class SubDomainSet(MultiSubDomain):
     """
     Class to define a set of N (a positive integer) subdomains.
 
@@ -535,22 +693,22 @@ class SubDomainSet(SubDomain):
     """The implicit dimension of the SubDomainSet."""
 
     def __init__(self, **kwargs):
-        super(SubDomainSet, self).__init__()
+        super().__init__()
+
         if self.implicit_dimension is None:
-            n = Dimension(name='n')
-            self.implicit_dimension = n
+            self.implicit_dimension = Dimension(name='n')
+
         self._n_domains = kwargs.get('N', 1)
         self._global_bounds = kwargs.get('bounds', None)
-        self._implicit_exprs = None
 
-    def __subdomain_finalize__(self, dimensions, shape, distributor=None, **kwargs):
-        # Create the SubDomain's SubDimensions
-        sub_dimensions = []
-        for d in dimensions:
-            sub_dimensions.append(SubDimension.middle
-                                  ('%si_%s' % (d.name, self.implicit_dimension.name),
-                                   d, 0, 0))
-        self._dimensions = tuple(sub_dimensions)
+    def __subdomain_finalize__(self, grid, **kwargs):
+        self._dtype = grid.dtype
+
+        # Create the SubDomainSet SubDimensions
+        self._dimensions = tuple(
+            MultiSubDimension('%si_%s' % (d.name, self.implicit_dimension.name), d)
+            for d in grid.dimensions
+        )
 
         # Compute the SubDomainSet shapes
         global_bounds = []
@@ -564,66 +722,23 @@ class SubDomainSet(SubDomain):
         shapes = []
         for i in range(self._n_domains):
             dshape = []
-            for s, m, M in zip(shape, d_m, d_M):
+            for s, m, M in zip(grid.shape, d_m, d_M):
                 assert(m.size == M.size)
                 dshape.append(s-m[i]-M[i])
             shapes.append(as_tuple(dshape))
         self._shape = as_tuple(shapes)
 
-        if distributor and distributor.is_parallel:
+        if grid.distributor and grid.distributor.is_parallel:
             # Now create local bounds based on distributor
             processed = []
-            for dim, d, m, M in zip(dimensions, distributor.decomposition, d_m, d_M):
-                bounds_m = np.zeros(m.shape, dtype=m.dtype)
-                bounds_M = np.zeros(m.shape, dtype=m.dtype)
-                for j in range(m.size):
-                    lmin = d.glb_min + m[j]
-                    lmax = d.glb_max - M[j]
-
-                    # Check if the subdomain doesn't intersect with the decomposition
-                    if lmin < d.loc_abs_min and lmax < d.loc_abs_min:
-                        bounds_m[j] = d.loc_abs_max
-                        bounds_M[j] = d.loc_abs_max
-                        continue
-                    if lmin > d.loc_abs_max and lmax > d.loc_abs_max:
-                        bounds_m[j] = d.loc_abs_max
-                        bounds_M[j] = d.loc_abs_max
-                        continue
-
-                    if lmin < d.loc_abs_min:
-                        bounds_m[j] = 0
-                    elif lmin > d.loc_abs_max:
-                        bounds_m[j] = d.loc_abs_max
-                        bounds_M[j] = d.loc_abs_max
-                        continue
-                    else:
-                        bounds_m[j] = d.index_glb_to_loc(m[j], LEFT)
-
-                    if lmax < d.loc_abs_min:
-                        bounds_m[j] = d.loc_abs_max
-                        bounds_M[j] = d.loc_abs_max
-                        continue
-                    elif lmax >= d.loc_abs_max:
-                        bounds_M[j] = 0
-                    else:
-                        bounds_M[j] = d.index_glb_to_loc(M[j], RIGHT)
-
-                processed.append(bounds_m)
-                processed.append(bounds_M)
+            for dec, m, M in zip(grid.distributor.decomposition, d_m, d_M):
+                processed.extend(self._bounds_glb_to_loc(dec, m, M))
             self._local_bounds = as_tuple(processed)
         else:
             # Not distributed and hence local and global bounds are
             # equivalent.
             self._local_bounds = self._global_bounds
 
-    @property
-    def n_domains(self):
-        return self._n_domains
-
-    @property
-    def bounds(self):
-        return self._local_bounds
-
     def _create_implicit_exprs(self, grid):
         if not len(self._local_bounds) == 2*len(self.dimensions):
             raise ValueError("Left and right bounds must be supplied for each dimension")
@@ -649,3 +764,42 @@ class SubDomainSet(SubDomain):
                 func.data[:] = self._local_bounds[j]
             dat.append(Eq(d.thickness[j % 2][0], func[i_dim]))
         return as_tuple(dat)
+
+    @property
+    def n_domains(self):
+        return self._n_domains
+
+    @property
+    def bounds(self):
+        return self._local_bounds
+
+    @property
+    def implicit_dimensions(self):
+        return (self.implicit_dimension,)
+
+
+# Preset SubDomains
+
+
+class Domain(SubDomain):
+
+    """
+    The entire computational domain (== boundary + interior).
+    """
+
+    name = 'domain'
+
+    def define(self, dimensions):
+        return dict(zip(dimensions, dimensions))
+
+
+class Interior(SubDomain):
+
+    """
+    The interior of the computational domain (i.e., boundaries are excluded).
+    """
+
+    name = 'interior'
+
+    def define(self, dimensions):
+        return {d: ('middle', 1, 1) for d in dimensions}
