@@ -1,6 +1,4 @@
-from sympy import sign
-
-from devito import Eq, Operator, Function, TimeFunction, Inc, solve
+from devito import Eq, Operator, Function, TimeFunction, Inc, solve, sign
 from devito.symbolics import retrieve_functions, INT
 from examples.seismic import PointSource, Receiver
 
@@ -135,6 +133,7 @@ def ForwardOperator(model, geometry, space_order=4,
 
     # Create interpolation expression for receivers
     rec_term = rec.interpolate(expr=u)
+
     # Substitute spacing terms to reduce flops
     return Operator(eqn + src_term + rec_term, subs=model.spacing_map,
                     name='Forward', **kwargs)
@@ -214,9 +213,9 @@ def GradientOperator(model, geometry, space_order=4, save=True,
     eqn = iso_stencil(v, model, kernel, forward=False)
 
     if kernel == 'OT2':
-        gradient_update = Inc(grad, - u.dt2 * v)
+        gradient_update = Inc(grad, - u * v.dt2)
     elif kernel == 'OT4':
-        gradient_update = Inc(grad, - (u.dt2 + s**2 / 12.0 * u.biharmonic(m**(-2))) * v)
+        gradient_update = Inc(grad, - u * v.dt2 - s**2 / 12.0 * u.biharmonic(m**(-2)) * v)
     # Add expression for receiver injection
     receivers = rec.inject(field=v.backward, expr=rec * s**2 / m)
 
