@@ -50,8 +50,10 @@ def ForwardOperator(model, geometry, space_order=4, save=False, **kwargs):
     """
 
     v = VectorTimeFunction(name='v', grid=model.grid,
+                           save=geometry.nt if save else None,
                            space_order=space_order, time_order=1)
     tau = TensorTimeFunction(name='tau', grid=model.grid,
+                             save=geometry.nt if save else None,
                              space_order=space_order, time_order=1)
 
     lam, mu, b = model.lam, model.mu, model.b
@@ -63,6 +65,7 @@ def ForwardOperator(model, geometry, space_order=4, save=False, **kwargs):
              model.damp * dt * mu * (grad(v.forward) + grad(v.forward).T))
 
     srcrec = src_rec(v, tau, model, geometry)
-    op = Operator([u_v] + [u_t] + srcrec, subs=model.spacing_map, name="ForwardElastic")
+    op = Operator([u_v] + [u_t] + srcrec, subs=model.spacing_map, name="ForwardElastic",
+                  **kwargs)
     # Substitute spacing terms to reduce flops
     return op
